@@ -21,11 +21,11 @@
 #include "HasTaskSet.h"
 #include "MemoryRange.h"
 #include "Monkeypatcher.h"
-#include "PropertyTable.h"
 #include "TaskishUid.h"
 #include "TraceStream.h"
 #include "core.h"
 #include "kernel_abi.h"
+#include "log.h"
 #include "remote_code_ptr.h"
 #include "util.h"
 
@@ -33,6 +33,7 @@ namespace rr {
 
 class AutoRemoteSyscalls;
 class MonitoredSharedMemory;
+class RecordSession;
 class RecordTask;
 class Session;
 class Task;
@@ -258,7 +259,7 @@ public:
             std::shared_ptr<MonitoredSharedMemory>&& monitored = nullptr);
     ~Mapping();
     Mapping(const Mapping&);
-    Mapping() = default;
+    Mapping() : local_addr(nullptr), flags(0) {}
     const Mapping& operator=(const Mapping& other) {
       this->~Mapping();
       new (this) Mapping(other);
@@ -358,6 +359,12 @@ public:
    */
   const std::string& exe_image() const { return exe; }
 
+  const std::string& interp_name() const { return interp_name_; }
+  void set_interp_name(std::string name) { interp_name_ = name; }
+
+  remote_ptr<void> interp_base() const { return interp_base_; }
+  void set_interp_base(remote_ptr<void> base) { interp_base_ = base; }
+
   /**
    * Assuming the last retired instruction has raised a SIGTRAP
    * and might be a breakpoint trap instruction, return the type
@@ -372,6 +379,11 @@ public:
    */
   BreakpointType get_breakpoint_type_at_addr(remote_code_ptr addr);
 
+  /**
+   * Check if the user has placed a hardware EXEC watchpoint at addr.
+   */
+  bool is_exec_watchpoint(remote_code_ptr addr);
+
   /**
    * Returns true when the breakpoint at |addr| is in private
    * non-writeable memory. When this returns true, the breakpoint can't be
@@ -669,17 +681,12 @@ public:
    * use.
    */
   static remote_ptr<void> rr_page_start() { return RR_PAGE_ADDR; }
-  /**
-   * This might not be the length of an actual system page, but we allocate
-   * at least this much space.
-   */
-  static uint32_t rr_page_size() { return 4096; }
   static remote_ptr<void> rr_page_end() {
-    return rr_page_start() + rr_page_size();
+    return rr_page_start() + PRELOAD_LIBRARY_PAGE_SIZE;
   }
 
   static remote_ptr<void> preload_thread_locals_start() {
-    return rr_page_start() + rr_page_size();
+    return rr_page_start() + PRELOAD_LIBRARY_PAGE_SIZE;
   }
   static uint32_t preload_thread_locals_size() {
     return PRELOAD_THREAD_LOCALS_SIZE;
@@ -775,12 +782,11 @@ public:
 
   static uint32_t chaos_mode_min_stack_size() { return 8 * 1024 * 1024; }
 
+  /* Returns null if we should not override the memory allocation */
   remote_ptr<void> chaos_mode_find_free_memory(RecordTask* t, size_t len, remote_ptr<void> hint);
   remote_ptr<void> find_free_memory(
       size_t len, remote_ptr<void> after = remote_ptr<void>());
 
-  PropertyTable& properties() { return properties_; }
-
   /**
    * The return value indicates whether we (re)created the preload_thread_locals
    * area.
@@ -850,7 +856,7 @@ public:
    */
   void fd_tables_changed();
 
-  static MemoryRange get_global_exclusion_range();
+  static MemoryRange get_global_exclusion_range(const RecordSession* session);
 
 private:
   struct Breakpoint;
@@ -1075,13 +1081,15 @@ private:
     bool changed;
   };
 
-  PropertyTable properties_;
-
   // All breakpoints set in this VM.
   BreakpointMap breakpoints;
   /* Path of the real executable image this address space was
    * exec()'d with. */
   std::string exe;
+  /* Path of the intepreter, if any, of exe. */
+  std::string interp_name_;
+  /* Base address of the interpeter (might be null!) */
+  remote_ptr<void> interp_base_;
   /* Pid of first task for this address space */
   pid_t leader_tid_;
   /* Serial number of first task for this address space */
