@@ -87,7 +87,9 @@ AutoRemoteSyscalls::AutoRemoteSyscalls(Task* t,
       new_tid_(-1),
       scratch_mem_was_mapped(false),
       use_singlestep_path(false),
-      enable_mem_params_(enable_mem_params) {
+      enable_mem_params_(enable_mem_params),
+      restore_sigmask(false),
+      need_sigpending_renable(false) {
   if (initial_at_seccomp) {
     // This should only ever happen during recording - we don't use the
     // seccomp traps during replay.
@@ -110,6 +112,39 @@ AutoRemoteSyscalls::AutoRemoteSyscalls(Task* t,
   if (enable_mem_params == ENABLE_MEMORY_PARAMS) {
     maybe_fix_stack_pointer();
   }
+  if (t->status().is_syscall() && t->regs().syscall_may_restart()) {
+    // VERY rare corner case alert: It is possible for the following sequence
+    // of events to occur:
+    //
+    // 1. Thread A is in a blocking may-restart syscall and gets interrupted by a tg-targeted signal
+    // 2. Thread B dequeues the signal
+    // 3. Thread A is in the syscall-exit-stop with TIF_SIGPENDING set (with registers indicating syscall restart)
+    // 4. We get here to perform an AutoRemoteSyscall
+    // 5. During AutoRemoteSyscall, TIF_SIGPENDING gets cleared on return to userspace
+    // 6. We finish the AutoRemoteSyscall and re-apply the registers.
+    // 7. ... As a result, the kernel does not check whether it needs to perform the
+    ///   syscall-restart register adjustment because TIF_SIGPENDING is not set.
+    // 8. The -ERESTART error code leaks to userspace.
+    //
+    // Arguably this is a kernel bug, but it's not clear how the behavior should be changed.
+    //
+    // To work around this, we forcibly re-enable TIF_SIGPENDING when cleaning up
+    // AutoRemoteSyscall (see below).
+    need_sigpending_renable = true;
+  }
+  if (t->session().is_recording()) {
+    RecordTask *rt = static_cast<RecordTask*>(t);
+    if (rt->schedule_frozen) {
+      // If we're explicitly controlling the schedule, make sure not to accidentally run
+      // any signals that we were not meant to be able to see.
+      restore_sigmask = true;
+      sigmask_to_restore = rt->get_sigmask();
+      sig_set_t all_blocked;
+      memset(&all_blocked, 0xff, sizeof(all_blocked));
+      // Ignore the process dying here - we'll notice later.
+      (void)rt->set_sigmask(all_blocked);
+    }
+  }
 }
 
 void AutoRemoteSyscalls::setup_path(bool enable_singlestep_path) {
@@ -186,10 +221,12 @@ void AutoRemoteSyscalls::maybe_fix_stack_pointer() {
   if (found_stack.start().is_null()) {
     AutoRemoteSyscalls remote(t, DISABLE_MEMORY_PARAMS);
     found_stack =
-        MemoryRange(remote.infallible_mmap_syscall(
+        MemoryRange(remote.infallible_mmap_syscall_if_alive(
                         remote_ptr<void>(), 4096, PROT_READ | PROT_WRITE,
                         MAP_PRIVATE | MAP_ANONYMOUS, -1, 0),
                     4096);
+    ASSERT(t, !found_stack.start().is_null())
+      << "Tracee unexpectedly died here";
     scratch_mem_was_mapped = true;
   }
 
@@ -215,13 +252,35 @@ void AutoRemoteSyscalls::restore_state_to(Task* t) {
   auto regs = initial_regs;
   regs.set_ip(initial_ip);
   regs.set_sp(initial_sp);
-  // Restore stomped registers.
-  t->set_regs(regs);
+  if (t->arch() == aarch64 && regs.syscall_may_restart()) {
+    // On AArch64, the kernel restarts aborted syscalls using an internal `orig_x0`.
+    // This gets overwritten everytime we make a syscall so we need to restore it
+    // if we are at a syscall that may restart.
+    // The kernel `orig_x0` isn't accessible from ptrace AFAICT but fortunately
+    // it does **NOT** get reset on syscall exit so we can actually set it's value
+    // just by making a dummy syscall with the correct x0 value.
+    auto restart_res = regs.syscall_result();
+    regs.set_ip(t->vm()->traced_syscall_ip());
+    // This can be any side-effect-free syscall that doesn't care about arg1.
+    // The kernel sets its `orig_x0` no matter whether the syscall actually needs it.
+    regs.set_syscallno(rr::ARM64Arch::getpid);
+    regs.set_arg1(regs.orig_arg1());
+    t->set_regs(regs);
+    if (t->enter_syscall(true)) {
+      t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
+    }
+    regs.set_ip(initial_ip);
+    regs.set_syscallno(regs.original_syscallno());
+    regs.set_syscall_result(restart_res);
+  }
   // If we were sitting at a seccomp trap, try to get back there by resuming
   // here. Since the original register contents caused a seccomp trap,
   // re-running the syscall with the same registers should put us right back
   // to this same seccomp trap.
   if (initial_at_seccomp && t->ptrace_event() != PTRACE_EVENT_SECCOMP) {
+    regs.set_ip(initial_ip.decrement_by_syscall_insn_length(t->arch()));
+    regs.set_syscallno(regs.original_syscallno());
+    t->set_regs(regs);
     RecordTask* rt = static_cast<RecordTask*>(t);
     while (true) {
       rt->resume_execution(RESUME_CONT, RESUME_WAIT, RESUME_NO_TICKS);
@@ -230,8 +289,22 @@ void AutoRemoteSyscalls::restore_state_to(Task* t) {
       rt->stash_sig();
     }
     ASSERT(rt, rt->ptrace_event() == PTRACE_EVENT_SECCOMP);
+  } else {
+    // Restore stomped registers.
+    t->set_regs(regs);
   }
   t->set_status(restore_wait_status);
+  if (restore_sigmask) {
+    static_cast<RecordTask*>(t)->set_sigmask(sigmask_to_restore);
+  }
+  if (need_sigpending_renable) {
+    // The purpose of this PTRACE_INTERRUPT is to re-enable TIF_SIGPENDING on
+    // the tracee, without forcing any actual signals on it. Since PTRACE_INTERRUPT
+    // needs to be able to interrupt re-startable system calls, it is required
+    // to set TIF_SIGPENDING, but the fact that this works is of course a very
+    // deep implementation detail.
+    t->do_ptrace_interrupt();
+  }
 }
 
 static bool ignore_signal(Task* t) {
@@ -277,7 +350,8 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
   callregs.set_syscallno(syscallno);
   t->set_regs(callregs);
 
-  if (use_singlestep_path) {
+  bool from_seccomp = initial_at_seccomp && t->ptrace_event() == PTRACE_EVENT_SECCOMP;
+  if (use_singlestep_path && !from_seccomp) {
     while (true) {
       t->resume_execution(RESUME_SINGLESTEP, RESUME_WAIT, RESUME_NO_TICKS);
       LOG(debug) << "Used singlestep path; status=" << t->status();
@@ -303,15 +377,18 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
       ASSERT(t, false) << "Unexpected status " << t->status();
     }
   } else {
-    if (initial_at_seccomp && t->ptrace_event() == PTRACE_EVENT_SECCOMP) {
+    bool exited = false;
+    if (from_seccomp) {
       LOG(debug) << "Skipping enter_syscall - already at seccomp stop";
     } else {
-      t->enter_syscall();
+      exited = !t->enter_syscall(true);
+      LOG(debug) << "Used enter_syscall; status=" << t->status();
     }
-    LOG(debug) << "Used enter_syscall; status=" << t->status();
     // proceed to syscall exit
-    t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
-    LOG(debug) << "syscall exit status=" << t->status();
+    if (!exited) {
+      t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
+      LOG(debug) << "syscall exit status=" << t->status();
+    }
   }
   while (true) {
     // If the syscall caused the task to exit, just stop now with that status.
@@ -334,7 +411,10 @@ long AutoRemoteSyscalls::syscall_base(int syscallno, Registers& callregs) {
     }
     if (ignore_signal(t)) {
       if (t->regs().syscall_may_restart()) {
-        t->enter_syscall();
+        if (!t->enter_syscall(true)) {
+          // We died, just let it be
+          break;
+        }
         LOG(debug) << "signal ignored; restarting syscall, status="
                    << t->status();
         t->resume_execution(RESUME_SYSCALL, RESUME_WAIT, RESUME_NO_TICKS);
@@ -463,6 +543,8 @@ static long child_recvmsg(AutoRemoteSyscalls& remote, int child_sock) {
     sizeof(msg), &msg, &ok);
 
   if (!ok) {
+    ASSERT(remote.task(), errno == ESRCH);
+    LOG(debug) << "Failed to write memory";
     return -ESRCH;
   }
   int ret = 0;
@@ -472,19 +554,26 @@ static long child_recvmsg(AutoRemoteSyscalls& remote, int child_sock) {
     ret = remote.syscall(Arch::recvmsg, child_sock, msg.remote_msg(), 0);
   }
   if (ret < 0) {
+    LOG(debug) << "Failed to recvmsg " << ret;
     return ret;
   }
   int their_fd = remote.task()->read_mem(msg.remote_cmsgdata(), &ok);
   if (!ok) {
+    ASSERT(remote.task(), errno == ESRCH);
+    LOG(debug) << "Failed to read msg";
     return -ESRCH;
   }
   return their_fd;
 }
 
-static int recvmsg_socket(ScopedFd& sock) {
+static int recvmsg_socket(ScopedFd& sock, bool blocking=true) {
   fd_message<NativeArch> msg;
   struct msghdr *msgp = (struct msghdr*)&msg.msg;
-  if (0 > recvmsg(sock, msgp, MSG_CMSG_CLOEXEC)) {
+  int flag = MSG_CMSG_CLOEXEC;
+  if (!blocking) {
+    flag |= MSG_DONTWAIT;
+  }
+  if (0 > recvmsg(sock, msgp, flag)) {
     return -1;
   }
 
@@ -541,9 +630,15 @@ template <typename Arch> int AutoRemoteSyscalls::send_fd_arch(const ScopedFd &ou
       child_recvmsg<Arch>(*this, task()->session().tracee_fd_number());
   if (child_syscall_result == -ESRCH) {
     /* The child did not receive the message. Read it out of the socket
-       buffer so it doesn't get read by another child later! */
-    int fd = recvmsg_socket(task()->session().tracee_socket_receiver_fd());
+     * buffer so it doesn't get read by another child later!
+     * Note that if the child was killed, we might get an error
+     * (i.e. process exit stop before syscall exit stop) even
+     * if the fd is already removed from the socket.
+     * Because of this, we use a non-blocking recvmsg to avoid blocking forever.
+     */
+    int fd = recvmsg_socket(task()->session().tracee_socket_receiver_fd(), false);
     if (fd >= 0) {
+      ASSERT(t, fd != our_fd.get()) << "This should always return a fresh fd!";
       close(fd);
     }
     return -ESRCH;
@@ -557,33 +652,46 @@ int AutoRemoteSyscalls::send_fd(const ScopedFd &our_fd) {
   RR_ARCH_FUNCTION(send_fd_arch, arch(), our_fd);
 }
 
-void AutoRemoteSyscalls::infallible_send_fd_dup(const ScopedFd& our_fd, int dup_to) {
-  int remote_fd = send_fd(our_fd);
-  ASSERT(task(), remote_fd >= 0);
+void AutoRemoteSyscalls::infallible_close_syscall_if_alive(int child_fd) {
+  infallible_syscall_if_alive(syscall_number_for_close(arch()), child_fd);
+}
+
+int AutoRemoteSyscalls::infallible_send_fd_if_alive(const ScopedFd &our_fd) {
+  int child_fd = send_fd(our_fd);
+  ASSERT(t, child_fd >= 0 || (child_fd == -ESRCH && !t->session().is_replaying()))
+    << "Failed to send fd; err=" << errno_name(-child_fd);
+  return child_fd;
+}
+
+void AutoRemoteSyscalls::infallible_send_fd_dup(const ScopedFd& our_fd, int dup_to, int dup3_flags) {
+  int remote_fd = infallible_send_fd_if_alive(our_fd);
+  ASSERT(t, remote_fd >= 0);
   if (remote_fd != dup_to) {
     long ret = infallible_syscall(syscall_number_for_dup3(arch()), remote_fd,
-                                  dup_to, O_CLOEXEC);
+                                  dup_to, dup3_flags);
     ASSERT(task(), ret == dup_to);
-    infallible_syscall(syscall_number_for_close(arch()), remote_fd);
+    infallible_close_syscall_if_alive(remote_fd);
   }
 }
 
-
-remote_ptr<void> AutoRemoteSyscalls::infallible_mmap_syscall(
+remote_ptr<void> AutoRemoteSyscalls::infallible_mmap_syscall_if_alive(
     remote_ptr<void> addr, size_t length, int prot, int flags, int child_fd,
-    uint64_t offset_pages) {
+    uint64_t offset_bytes) {
+  ASSERT(t, offset_bytes % page_size() == 0)
+    << "mmap offset (" << offset_bytes << ") must be multiple of page size ("
+    << page_size() << ")";
   // The first syscall argument is called "arg 1", so
   // our syscall-arg-index template parameter starts
   // with "1".
   remote_ptr<void> ret =
       has_mmap2_syscall(arch())
-          ? infallible_syscall_ptr(syscall_number_for_mmap2(arch()), addr,
-                                   length, prot, flags, child_fd,
-                                   (off_t)offset_pages)
-          : infallible_syscall_ptr(syscall_number_for_mmap(arch()), addr,
-                                   length, prot, flags, child_fd,
-                                   offset_pages * page_size());
-  if (flags & MAP_FIXED) {
+          ? infallible_syscall_ptr_if_alive(syscall_number_for_mmap2(arch()), addr,
+                                            length, prot, flags, child_fd,
+                                            (off_t)offset_bytes / 4096)
+          : infallible_syscall_ptr_if_alive(syscall_number_for_mmap(arch()), addr,
+                                            length, prot, flags, child_fd,
+                                            offset_bytes);
+  if (ret && (flags & MAP_FIXED)) {
     ASSERT(t, addr == ret) << "MAP_FIXED at " << addr << " but got " << ret;
   }
   return ret;
@@ -599,6 +707,7 @@ int64_t AutoRemoteSyscalls::infallible_lseek_syscall(int fd, int64_t offset,
       return t->read_mem(mem.get().cast<int64_t>());
     }
     case x86_64:
+    case aarch64:
       return infallible_syscall(syscall_number_for_lseek(arch()), fd, offset,
                                 whence);
     default:
@@ -612,7 +721,7 @@ void AutoRemoteSyscalls::check_syscall_result(long ret, int syscallno, bool allo
     // Sign-extend ret because it can be a 32-bit negative errno
     ret = (int)ret;
   }
-  if (allow_death && ret == -ESRCH) {
+  if (ret == -ESRCH && allow_death && !t->session().is_replaying()) {
     return;
   }
   if (-4096 < ret && ret < 0) {
@@ -632,13 +741,13 @@ void AutoRemoteSyscalls::finish_direct_mmap(
                                int prot, int flags,
                                const string& backing_file_name,
                                int backing_file_open_flags,
-                               off64_t backing_offset_pages,
+                               off64_t backing_offset_bytes,
                                struct stat& real_file, string& real_file_name) {
   int fd;
 
   LOG(debug) << "directly mmap'ing " << length << " bytes of "
-             << backing_file_name << " at page offset "
-             << HEX(backing_offset_pages);
+             << backing_file_name << " at offset "
+             << HEX(backing_offset_bytes);
 
   ASSERT(task(), !(flags & MAP_GROWSDOWN));
 
@@ -651,7 +760,7 @@ void AutoRemoteSyscalls::finish_direct_mmap(
                             backing_file_open_flags);
   }
   /* And mmap that file. */
-  infallible_mmap_syscall(rec_addr, length,
+  infallible_mmap_syscall_if_alive(rec_addr, length,
                           /* (We let SHARED|WRITEABLE
                           * mappings go through while
                           * they're not handled properly,
@@ -662,7 +771,7 @@ void AutoRemoteSyscalls::finish_direct_mmap(
                           * memory devices (requires
                           * MAP_SHARED_VALIDATE). Drop it for the
                           * backing file. */
-                          backing_offset_pages);
+                          backing_offset_bytes);
 
   // While it's open, grab the link reference.
   real_file = task()->stat_fd(fd);
@@ -670,7 +779,7 @@ void AutoRemoteSyscalls::finish_direct_mmap(
 
   /* Don't leak the tmp fd.  The mmap doesn't need the fd to
    * stay open. */
-  infallible_syscall(syscall_number_for_close(arch()), fd);
+  infallible_close_syscall_if_alive(fd);
 }
 
 
