@@ -2,6 +2,8 @@
 
 #include "DiversionSession.h"
 
+#include <linux/prctl.h>
+
 #include "AutoRemoteSyscalls.h"
 #include "ReplaySession.h"
 #include "core.h"
@@ -12,7 +14,8 @@ using namespace std;
 
 namespace rr {
 
-DiversionSession::DiversionSession() : emu_fs(EmuFs::create()) {}
+DiversionSession::DiversionSession(int cpu_binding) :
+  emu_fs(EmuFs::create()), fake_rdstc(uint64_t(1) << 60), cpu_binding_(cpu_binding) {}
 
 DiversionSession::~DiversionSession() {
   // We won't permanently leak any OS resources by not ensuring
@@ -47,6 +50,12 @@ static void execute_syscall(Task* t) {
   remote.regs().set_syscall_result(t->regs().syscall_result());
 }
 
+uint64_t DiversionSession::next_rdtsc_value() {
+  uint64_t rdtsc_value = fake_rdstc;
+  fake_rdstc += 1 << 20; // 1M cycles
+  return rdtsc_value;
+}
+
 template <typename Arch>
 static void process_syscall_arch(Task* t, int syscallno) {
   LOG(debug) << "Processing " << syscall_name(syscallno, Arch::arch());
@@ -60,6 +69,15 @@ static void process_syscall_arch(Task* t, int syscallno) {
     return;
   }
 
+  if (syscallno == t->session().syscall_number_for_rrcall_rdtsc()) {
+    uint64_t rdtsc_value = static_cast<DiversionSession*>(&t->session())->next_rdtsc_value();
+    LOG(debug) << "Faking rrcall_rdtsc syscall with value " << rdtsc_value;
+    remote_ptr<uint64_t> out_param(t->regs().arg1());
+    t->write_mem(out_param, rdtsc_value);
+    finish_emulated_syscall_with_ret(t, 0);
+    return;
+  }
+
   switch (syscallno) {
     // We blacklist these syscalls because the params include
     // namespaced identifiers that are different in replay than
@@ -85,7 +103,12 @@ static void process_syscall_arch(Task* t, int syscallno) {
     case Arch::rt_sigqueueinfo:
     case Arch::rt_tgsigqueueinfo:
     case Arch::tgkill:
-    case Arch::tkill: {
+    case Arch::tkill:
+    // fork/vfork/clone are likely to lead to disaster because we only
+    // ever allow a single task to run.
+    case Arch::fork:
+    case Arch::vfork:
+    case Arch::clone: {
       LOG(debug) << "Suppressing syscall "
                  << syscall_name(syscallno, t->arch());
       Registers r = t->regs();
@@ -94,6 +117,19 @@ static void process_syscall_arch(Task* t, int syscallno) {
       return;
     }
 
+    case Arch::prctl: {
+      Registers r = t->regs();
+      int op = r.arg1();
+      if (op == PR_SET_TSC) {
+        LOG(debug) << "Suppressing syscall "
+                   << syscall_name(syscallno, t->arch());
+        r.set_syscall_result(-ENOSYS);
+        t->set_regs(r);
+        return;
+      }
+      break;
+    }
+
     case Arch::gettid: {
       auto tid = t->own_namespace_tid();
       LOG(debug) << "Emulating gettid with " << tid;
@@ -114,7 +150,7 @@ static void process_syscall_arch(Task* t, int syscallno) {
   }
 
   LOG(debug) << "Executing syscall " << syscall_name(syscallno, t->arch());
-  return execute_syscall(t);
+  execute_syscall(t);
 }
 
 static void process_syscall(Task* t, int syscallno){
@@ -127,6 +163,20 @@ static void handle_ptrace_exit_event(Task *t) {
   delete t;
 }
 
+static bool maybe_handle_task_exit(Task* t, TaskContext* context,
+                                   DiversionSession::DiversionResult* result) {
+  if (t->ptrace_event() != PTRACE_EVENT_EXIT) {
+    return false;
+  }
+  handle_ptrace_exit_event(t);
+  // This is now a dangling pointer, so clear it.
+  context->task = nullptr;
+  result->status = DiversionSession::DIVERSION_EXITED;
+  result->break_status.task_context = *context;
+  result->break_status.task_exit = true;
+  return true;
+}
+
 /**
  * Advance execution until either a signal is received (including a SIGTRAP
  * generated by a single-step) or a syscall is made.
@@ -137,60 +187,70 @@ DiversionSession::DiversionResult DiversionSession::diversion_step(
   assert_fully_initialized();
 
   DiversionResult result;
+  TaskContext context(t);
 
   // An exit might have occurred while processing a previous syscall.
-  if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
-    // We're about to destroy the task, so capture the context while
-    // we can.
-    TaskContext context(t);
-    handle_ptrace_exit_event(t);
-    // This is now a dangling pointer, so clear it.
-    context.task = nullptr;
-    result.status = DIVERSION_EXITED;
-    result.break_status.task_context = context;
-    result.break_status.task_exit = true;
+  if (maybe_handle_task_exit(t, &context, &result)) {
     return result;
   }
 
   t->set_in_diversion(true);
 
-  switch (command) {
-    case RUN_CONTINUE:
-      LOG(debug) << "Continuing to next syscall";
-      t->resume_execution(RESUME_SYSEMU, RESUME_WAIT, RESUME_UNLIMITED_TICKS,
-                          signal_to_deliver);
-      break;
-    case RUN_SINGLESTEP:
-      LOG(debug) << "Stepping to next insn/syscall";
-      t->resume_execution(RESUME_SYSEMU_SINGLESTEP, RESUME_WAIT,
-                          RESUME_UNLIMITED_TICKS, signal_to_deliver);
-      break;
-    default:
-      FATAL() << "Illegal run command " << command;
-  }
-
-  if (t->ptrace_event() == PTRACE_EVENT_EXIT) {
-    handle_ptrace_exit_event(t);
-    result.status = DIVERSION_EXITED;
-    return result;
-  }
-
-  result.status = DIVERSION_CONTINUE;
-  if (t->stop_sig()) {
-    LOG(debug) << "Pending signal: " << t->get_siginfo();
-    result.break_status = diagnose_debugger_trap(t, command);
-    if (!result.break_status.breakpoint_hit && result.break_status.watchpoints_hit.empty() && !result.break_status.singlestep_complete && (t->stop_sig() == SIGTRAP)) {
-      result.break_status.signal = unique_ptr<siginfo_t>(new siginfo_t(t->get_siginfo()));
-      result.break_status.signal->si_signo = t->stop_sig();
+  while (true) {
+    switch (command) {
+      case RUN_CONTINUE:
+        LOG(debug) << "Continuing to next syscall";
+        t->resume_execution(RESUME_SYSEMU, RESUME_WAIT, RESUME_UNLIMITED_TICKS,
+                            signal_to_deliver);
+        break;
+      case RUN_SINGLESTEP:
+        LOG(debug) << "Stepping to next insn/syscall";
+        t->resume_execution(RESUME_SYSEMU_SINGLESTEP, RESUME_WAIT,
+                           RESUME_UNLIMITED_TICKS, signal_to_deliver);
+        break;
+      default:
+        FATAL() << "Illegal run command " << command;
     }
-    LOG(debug) << "Diversion break at ip=" << (void*)t->ip().register_value()
-               << "; break=" << result.break_status.breakpoint_hit
-               << ", watch=" << !result.break_status.watchpoints_hit.empty()
-               << ", singlestep=" << result.break_status.singlestep_complete;
-    ASSERT(t,
-           !result.break_status.singlestep_complete ||
-               command == RUN_SINGLESTEP);
-    return result;
+
+    if (maybe_handle_task_exit(t, &context, &result)) {
+      return result;
+    }
+
+    result.status = DIVERSION_CONTINUE;
+    if (t->stop_sig()) {
+      LOG(debug) << "Pending signal: " << t->get_siginfo();
+      result.break_status = diagnose_debugger_trap(t, command);
+      if (t->stop_sig() == SIGTRAP &&
+          !result.break_status.breakpoint_hit &&
+          result.break_status.watchpoints_hit.empty() &&
+          !result.break_status.singlestep_complete) {
+        result.break_status.signal = unique_ptr<siginfo_t>(new siginfo_t(t->get_siginfo()));
+        result.break_status.signal->si_signo = t->stop_sig();
+      } else if (t->stop_sig() == SIGSEGV) {
+        auto trapped_instruction = trapped_instruction_at(t, t->ip());
+        if (trapped_instruction == TrappedInstruction::RDTSC) {
+          size_t len = trapped_instruction_len(trapped_instruction);
+          uint64_t rdtsc_value = next_rdtsc_value();
+          LOG(debug) << "Faking RDTSC instruction with value " << rdtsc_value;
+          Registers r = t->regs();
+          r.set_ip(r.ip() + len);
+          r.set_ax((uint32_t)rdtsc_value);
+          r.set_dx(rdtsc_value >> 32);
+          t->set_regs(r);
+          result.break_status = BreakStatus();
+          continue;
+        }
+      }
+      LOG(debug) << "Diversion break at ip=" << (void*)t->ip().register_value()
+                 << "; break=" << result.break_status.breakpoint_hit
+                 << ", watch=" << !result.break_status.watchpoints_hit.empty()
+                 << ", singlestep=" << result.break_status.singlestep_complete;
+      ASSERT(t,
+             !result.break_status.singlestep_complete ||
+                 command == RUN_SINGLESTEP);
+      return result;
+    }
+    break;
   }
 
   if (t->status().is_syscall()) {
