@@ -24,8 +24,6 @@ static const int st_reg_space = 16;
 static const int xmm_regs_offset = 160;
 static const int xmm_reg_space = 16;
 
-static const int xsave_feature_pkru = 9;
-
 static const uint8_t fxsave_387_ctrl_offsets[] = {
   // The Intel documentation says that the following layout is only valid in
   // 32-bit mode, or when fxsave is executed in 64-bit mode without an
@@ -43,6 +41,7 @@ static const uint8_t fxsave_387_ctrl_offsets[] = {
 };
 
 static const int fip_offset = 8;
+static const int fop_offset = 6;
 static const int fdp_offset = 16;
 static const int mxcsr_offset = 24;
 
@@ -67,6 +66,7 @@ static bool reg_in_range(GdbRegister regno, GdbRegister low, GdbRegister high,
 }
 
 static const int AVX_FEATURE_BIT = 2;
+static const int PKRU_FEATURE_BIT = 9;
 
 static const size_t xsave_header_offset = 512;
 static const size_t xsave_header_size = 64;
@@ -92,11 +92,12 @@ static RegData xsave_register_data(SupportedArch arch, GdbRegister regno) {
         regno = (GdbRegister)(regno - DREG_YMM0H + DREG_64_YMM0H);
         break;
       }
-      if (regno < DREG_FIRST_FXSAVE_REG || regno > DREG_LAST_FXSAVE_REG) {
-        return RegData();
-      }
       if (regno == DREG_MXCSR) {
         regno = DREG_64_MXCSR;
+      } else if (regno == DREG_PKRU) {
+        regno = DREG_64_PKRU;
+      } else if (regno < DREG_FIRST_FXSAVE_REG || regno > DREG_LAST_FXSAVE_REG) {
+        return RegData();
       } else {
         regno = (GdbRegister)(regno - DREG_FIRST_FXSAVE_REG +
                               DREG_64_FIRST_FXSAVE_REG);
@@ -125,6 +126,21 @@ static RegData xsave_register_data(SupportedArch arch, GdbRegister regno) {
     return result;
   }
 
+  if (regno == DREG_64_PKRU) {
+    const XSaveLayout& layout = xsave_native_layout();
+    if (PKRU_FEATURE_BIT > layout.feature_layouts.size()) {
+      return RegData();
+    }
+
+    const XSaveFeatureLayout& fl = layout.feature_layouts[PKRU_FEATURE_BIT];
+    result.offset = fl.offset;
+    // NB: the PKRU *region* may be 8 bytes to maintain alignment but the
+    // PKRU *register* is only the first 4 bytes.
+    result.size = 4;
+    result.xsave_feature_bit = PKRU_FEATURE_BIT;
+    return result;
+  }
+
   if (regno < DREG_64_FIRST_FXSAVE_REG || regno > DREG_64_LAST_FXSAVE_REG) {
     return RegData();
   }
@@ -139,13 +155,22 @@ static RegData xsave_register_data(SupportedArch arch, GdbRegister regno) {
   return RegData(fxsave_387_ctrl_offsets[regno - DREG_64_FCTRL], 4);
 }
 
-static uint64_t xsave_features(const vector<uint8_t>& data) {
+static const uint64_t* xsave_features(const vector<uint8_t>& data) {
   // If this is just FXSAVE(64) data then we we have no XSAVE header and no
   // XSAVE(64) features enabled.
   return data.size() < xsave_header_offset + xsave_header_size
-             ? 0
-             : *reinterpret_cast<const uint64_t*>(data.data() +
-                                                  xsave_header_offset);
+             ? nullptr
+             : reinterpret_cast<const uint64_t*>(data.data() +
+                                                 xsave_header_offset);
+
+}
+
+static uint64_t* xsave_features(vector<uint8_t>& data) {
+  // If this is just FXSAVE(64) data then we we have no XSAVE header and no
+  // XSAVE(64) features enabled.
+  return data.size() < xsave_header_offset + xsave_header_size
+             ? nullptr
+             : reinterpret_cast<uint64_t*>(data.data() + xsave_header_offset);
 }
 
 size_t ExtraRegisters::read_register(uint8_t* buf, GdbRegister regno,
@@ -194,8 +219,10 @@ size_t ExtraRegisters::read_register(uint8_t* buf, GdbRegister regno,
 
   // Apparently before any AVX registers are used, the feature bit is not set
   // in the XSAVE data, so we'll just return 0 for them here.
+  const uint64_t* xsave_features_ = xsave_features(data_);
   if (reg_data.xsave_feature_bit >= 0 &&
-      !(xsave_features(data_) & (1 << reg_data.xsave_feature_bit))) {
+      (!xsave_features_ ||
+       !(*xsave_features_ & (1 << reg_data.xsave_feature_bit)))) {
     memset(buf, 0, reg_data.size);
   } else {
     DEBUG_ASSERT(size_t(reg_data.offset + reg_data.size) <= data_.size());
@@ -204,6 +231,68 @@ size_t ExtraRegisters::read_register(uint8_t* buf, GdbRegister regno,
   return reg_data.size;
 }
 
+bool ExtraRegisters::write_register(GdbRegister regno, const void* value,
+                                    size_t value_size) {
+  if (format_ == NT_FPR) {
+    if (arch() != aarch64) {
+      return false;
+    }
+
+    RegData reg_data;
+    if (DREG_V0 <= regno && regno <= DREG_V31) {
+      reg_data = RegData(offsetof(ARM64Arch::user_fpsimd_state, vregs[0]) +
+        ((regno - DREG_V0) * 16), 16);
+    } else if (regno == DREG_FPSR) {
+      reg_data = RegData(offsetof(ARM64Arch::user_fpsimd_state, fpsr),
+                         sizeof(uint32_t));
+    } else if (regno == DREG_FPCR) {
+      reg_data = RegData(offsetof(ARM64Arch::user_fpsimd_state, fpcr),
+                         sizeof(uint32_t));
+    } else {
+      return false;
+    }
+
+    DEBUG_ASSERT(reg_data.size > 0);
+    if ((size_t)reg_data.size != value_size) {
+      LOG(warn) << "Register " << regno << "has mismatched sizes ("
+                << reg_data.size << " vs " << value_size << ")";
+      return false;
+    }
+
+    DEBUG_ASSERT(size_t(reg_data.offset + reg_data.size) <= data_.size());
+    memcpy(data_.data() + reg_data.offset, value, value_size);
+    return true;
+  }
+
+  if (format_ != XSAVE) {
+    return false;
+  }
+
+  auto reg_data = xsave_register_data(arch(), regno);
+  if (reg_data.offset < 0 || empty()) {
+    return false;
+  }
+
+  DEBUG_ASSERT(reg_data.size > 0);
+  if ((size_t)reg_data.size != value_size) {
+    LOG(warn) << "Register " << regno << "has mismatched sizes ("
+              << reg_data.size << " vs " << value_size << ")";
+    return false;
+  }
+
+  if (reg_data.xsave_feature_bit >= 0) {
+    uint64_t* xsave_features_ = xsave_features(data_);
+    if (!xsave_features_) {
+      return false;
+    }
+
+    *xsave_features_ |= (1 << reg_data.xsave_feature_bit);
+  }
+
+  memcpy(data_.data() + reg_data.offset, value, value_size);
+  return true;
+}
+
 static const int xinuse_offset = 512;
 
 uint64_t ExtraRegisters::read_xinuse(bool* defined) const {
@@ -228,6 +317,17 @@ uint64_t ExtraRegisters::read_fip(bool* defined) const {
   return ret;
 }
 
+uint16_t ExtraRegisters::read_fop(bool* defined) const {
+  if (format_ != XSAVE) {
+    *defined = false;
+    return 0;
+  }
+
+  uint16_t ret;
+  memcpy(&ret, data_.data() + fop_offset, sizeof(ret));
+  return ret;
+}
+
 uint32_t ExtraRegisters::read_mxcsr(bool* defined) const {
   if (format_ != XSAVE) {
     *defined = false;
@@ -269,8 +369,8 @@ void ExtraRegisters::validate(Task* t) {
   if (data_.size() > offset) {
     ASSERT(t, data_.size() >= offset + 64);
     offset += 64;
-    uint64_t features = xsave_features(data_);
-    if (features & (1 << AVX_FEATURE_BIT)) {
+    const uint64_t* features = xsave_features(data_);
+    if (features && (*features & (1 << AVX_FEATURE_BIT))) {
       ASSERT(t, data_.size() >= offset + 256);
     }
   }
@@ -387,29 +487,9 @@ template <typename T> static vector<uint8_t> to_vector(const T& v) {
   return result;
 }
 
-static bool all_zeroes(const uint8_t* data, size_t size) {
-  for (size_t i = 0; i < size; ++i) {
-    if (data[i]) {
-      return false;
-    }
-  }
-  return true;
-}
-
-static uint32_t features_used(const uint8_t* data,
-                              const XSaveLayout& layout) {
+static uint32_t features_used(const uint8_t* data) {
   uint64_t features;
   memcpy(&features, data + xsave_header_offset, sizeof(features));
-  uint64_t pkru_bit = uint64_t(1) << xsave_feature_pkru;
-  if ((features & pkru_bit) &&
-      xsave_feature_pkru < layout.feature_layouts.size()) {
-    // Check if it's really used
-    const XSaveFeatureLayout& fl = layout.feature_layouts[xsave_feature_pkru];
-    if (uint64_t(fl.offset) + fl.size <= layout.full_size &&
-        all_zeroes(data + fl.offset, fl.size)) {
-      features &= ~pkru_bit;
-    }
-  }
   return features;
 }
 
@@ -474,7 +554,28 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
 
   // Check for unsupported features being used
   if (layout.full_size >= xsave_header_end) {
-    uint64_t features = features_used(data, layout);
+    uint64_t features = features_used(data);
+    /* Mask off the PKRU bit unconditionally here.
+     * We want traces that are recorded on machines with PKRU but
+     * that don't actually use PKRU to be replayable on machines
+     * without PKRU. Linux, however, sets the PKRU register to
+     * 0x55555554 (only the default key is allowed to access memory),
+     * while the default hardware value is 0, so in some sense
+     * PKRU is always in use.
+     *
+     * There are three classes of side effects of the pkey feature.
+     * 1. The direct effects of syscalls such as pkey_alloc/pkey_mprotect
+     *    on registers.
+     * 2. Traps generated by the CPU when the protection keys are violated.
+     * 3. The RDPKRU instruction writing to EAX.
+     *
+     * The first two are replayed exactly by rr. The latter will trigger
+     * SIGILL on any machine without PKRU, which is no different from
+     * any other new CPU instruction that doesn't have its own XSAVE
+     * feature bit. So ignore the PKRU bit here and leave users on their
+     * own with respect to RDPKRU.
+     */
+    features &= ~(uint64_t(1) << PKRU_FEATURE_BIT);
     if (features & ~native_layout.supported_feature_bits) {
       LOG(error) << "Unsupported CPU features found: got " << HEX(features)
                  << " (" << xsave_feature_string(features)
@@ -501,7 +602,7 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
     return true;
   }
 
-  uint64_t features = features_used(data, layout);
+  uint64_t features = features_used(data);
   // OK, now both our native layout and the input layout are using the full
   // XSAVE header. Copy the header. Make sure to use our updated `features`.
   memcpy(data_.data() + xsave_header_offset, &features, sizeof(features));
@@ -544,7 +645,6 @@ bool ExtraRegisters::set_to_raw_data(SupportedArch a, Format format,
 
 vector<uint8_t> ExtraRegisters::get_user_fpregs_struct(
     SupportedArch arch) const {
-  DEBUG_ASSERT(format_ == XSAVE);
   switch (arch) {
     case x86:
       DEBUG_ASSERT(format_ == XSAVE);
@@ -570,9 +670,9 @@ vector<uint8_t> ExtraRegisters::get_user_fpregs_struct(
 
 void ExtraRegisters::set_user_fpregs_struct(Task* t, SupportedArch arch,
                                             void* data, size_t size) {
-  DEBUG_ASSERT(format_ == XSAVE);
   switch (arch) {
     case x86:
+      DEBUG_ASSERT(format_ == XSAVE);
       ASSERT(t, size >= sizeof(X86Arch::user_fpregs_struct));
       ASSERT(t, data_.size() >= sizeof(X86Arch::user_fpxregs_struct));
       convert_x86_fpregs_to_fxsave(
@@ -580,10 +680,17 @@ void ExtraRegisters::set_user_fpregs_struct(Task* t, SupportedArch arch,
           reinterpret_cast<X86Arch::user_fpxregs_struct*>(data_.data()));
       return;
     case x86_64:
+      DEBUG_ASSERT(format_ == XSAVE);
       ASSERT(t, data_.size() >= sizeof(X64Arch::user_fpregs_struct));
       ASSERT(t, size >= sizeof(X64Arch::user_fpregs_struct));
       memcpy(data_.data(), data, sizeof(X64Arch::user_fpregs_struct));
       return;
+    case aarch64:
+      DEBUG_ASSERT(format_ == NT_FPR);
+      ASSERT(t, size >= sizeof(ARM64Arch::user_fpregs_struct));
+      ASSERT(t, data_.size() >= sizeof(ARM64Arch::user_fpregs_struct));
+      memcpy(data_.data(), data, sizeof(ARM64Arch::user_fpregs_struct));
+      return;
     default:
       DEBUG_ASSERT(0 && "Unknown arch");
   }
@@ -642,8 +749,12 @@ void ExtraRegisters::reset() {
       * Avoid this issue by setting the bit if the feature is supported by the
       * CPU.
       */
-      uint64_t pkru_bit = uint64_t(1) << xsave_feature_pkru;
+      uint64_t pkru_bit = uint64_t(1) << PKRU_FEATURE_BIT;
       if (xcr0() & pkru_bit) {
+        RegData d = xsave_register_data(arch(), arch() == x86_64 ? DREG_64_PKRU : DREG_PKRU);
+        DEBUG_ASSERT(d.xsave_feature_bit == PKRU_FEATURE_BIT);
+        DEBUG_ASSERT(d.offset + d.size <= (int)data_.size());
+        *reinterpret_cast<int*>(data_.data() + d.offset) = 0x55555554;
         xinuse |= pkru_bit;
       }
 
