@@ -83,6 +83,11 @@ public:
    */
   uint64_t read_fip(bool* defined) const;
 
+  /**
+   * Read FOP field
+   */
+  uint16_t read_fop(bool* defined) const;
+
   /**
    * Read MXCSR field
    */
@@ -100,6 +105,12 @@ public:
    */
   size_t read_register(uint8_t* buf, GdbRegister regno, bool* defined) const;
 
+  /**
+   * Like |Registers::write_register()|, except attempts to write
+   * the value of an "extra register" (floating point / vector).
+   */
+  bool write_register(GdbRegister regno, const void* value, size_t value_size);
+
   /**
    * Get a user_fpregs_struct for a particular Arch from these ExtraRegisters.
    */
