@@ -34,9 +34,12 @@ public:
     ProcMem,
     Stdio,
     VirtualPerfCounter,
+    NonvirtualPerfCounter,
     SysCpu,
     ProcStat,
-    RRPage
+    RRPage,
+    ODirect,
+    BpfMap,
   };
 
   virtual Type type() { return Base; }
@@ -127,6 +130,8 @@ public:
   }
 };
 
+std::string file_monitor_type_name(FileMonitor::Type t);
+
 } // namespace rr
 
 #endif /* RR_FILE_MONITOR_H_ */
