@@ -368,20 +368,22 @@ static GdbThreadId parse_threadid(const char* str, char** endptr) {
   }
   t.pid = strtol(str, &endp, 16);
   parser_assert(endp);
-  if ('\0' == *endp) {
+
+  /* terminators (single process, no PID or TID, depending on 'p' prefix) */ 
+  if (*endp == '\0' || *endp == ';' || *endp == ',') {
     if (multiprocess) {
       t.tid = -1;
     } else {
       t.tid = t.pid;
       t.pid = -1;
     }
-    *endptr = endp;
-    return t;
+  /* multiprocess syntax "<pid>.<tid>" */
+  } else if (*endp == '.') {
+    str = endp + 1;
+    t.tid = strtol(str, &endp, 16);
   }
 
-  parser_assert('.' == *endp);
-  str = endp + 1;
-  t.tid = strtol(str, &endp, 16);
+  parser_assert(*endp == '\0' || *endp == ';' || *endp == ',');
 
   *endptr = endp;
   return t;
@@ -406,7 +408,12 @@ void GdbConnection::write_xfer_response(const void* data, size_t size,
 }
 
 static string read_target_desc(const char* file_name) {
-  string path = resource_path() + "share/rr/" + string(file_name);
+#ifdef __BIONIC__
+  const char* share_path = "usr/share/rr/";
+#else
+  const char* share_path = "share/rr/";
+#endif
+  string path = resource_path() + share_path + string(file_name);
   stringstream ss;
   FILE* f = fopen(path.c_str(), "r");
   DEBUG_ASSERT(f);
@@ -432,6 +439,10 @@ static const char* target_description_name(uint32_t cpu_features) {
       return "i386-avx-linux.xml";
     case GdbConnection::CPU_X86_64 | GdbConnection::CPU_AVX:
       return "amd64-avx-linux.xml";
+    case GdbConnection::CPU_PKU | GdbConnection::CPU_AVX:
+      return "i386-pkeys-linux.xml";
+    case GdbConnection::CPU_X86_64 | GdbConnection::CPU_PKU | GdbConnection::CPU_AVX:
+      return "amd64-pkeys-linux.xml";
     case GdbConnection::CPU_AARCH64:
       return "aarch64-core.xml";
     default:
@@ -670,7 +681,8 @@ bool GdbConnection::query(char* payload) {
                  ";hwbreak+"
                  ";swbreak+"
                  ";ConditionalBreakpoints+"
-                 ";vContSupported+";
+                 ";vContSupported+"
+                 ";QPassSignals+";
     if (features().reverse_execution) {
       supported << ";ReverseContinue+"
                    ";ReverseStep+";
@@ -679,6 +691,7 @@ bool GdbConnection::query(char* payload) {
     return false;
   }
   if (!strcmp(name, "Symbol")) {
+#ifdef PROC_SERVICE_H
     LOG(debug) << "gdb is ready for symbol lookups";
     const char* colon = strchr(args, ':');
     parser_assert(colon != nullptr);
@@ -693,6 +706,11 @@ bool GdbConnection::query(char* payload) {
     ++args;
     req.sym().name = decode_ascii_encoded_hex_str(args);
     return true;
+#else
+    LOG(debug) << "gdb is ready for symbol lookups, but we don't support them";
+    write_packet("");
+    return false;
+#endif
   }
   if (strstr(name, "ThreadExtraInfo") == name) {
     // ThreadExtraInfo is a special snowflake that
@@ -767,6 +785,29 @@ bool GdbConnection::set_var(char* payload) {
     return false;
   }
 
+  if (!strncmp(name, "PassSignals", sizeof("PassSignals"))) {
+    pass_signals.clear();
+    while (*args != '\0') {
+      char *next = nullptr;
+      int sig = std::strtol(args, &next, 16);
+      parser_assert(next != nullptr);
+
+      LOG(debug) << "registered " << sig << " by QPassSignal";
+      pass_signals.insert(sig);
+
+      args = next;
+      if (*args == '\0') {
+        break;
+      }
+
+      parser_assert(*args == ';');
+      args++;
+    }
+
+    write_packet("OK");
+    return false;
+  }
+
   UNHANDLED_REQ() << "Unhandled gdb set: Q" << name;
   return false;
 }
@@ -1477,6 +1518,14 @@ void GdbConnection::notify_stop(GdbThreadId thread, int sig,
                                 const char *reason) {
   DEBUG_ASSERT(req.is_resume_request() || req.type == DREQ_INTERRUPT);
 
+  // don't pass this signal to gdb if it is specified not to
+  if (pass_signals.find(to_gdb_signum(sig)) != pass_signals.end()) {
+    LOG(debug) << "discarding stop notification for signal " << sig 
+                << " on thread " << thread << " as specified by QPassSignal";
+
+    return;
+  }
+
   if (tgid != thread.pid) {
     LOG(debug) << "ignoring stop of " << thread
                << " because we're debugging tgid " << tgid;
@@ -1907,4 +1956,6 @@ void GdbConnection::send_file_error_reply(int system_errno) {
 
 bool GdbConnection::is_connection_alive() { return connection_alive_; }
 
+bool GdbConnection::is_pass_signal(int sig) { return pass_signals.find(to_gdb_signum(sig)) != pass_signals.end(); }
+
 } // namespace rr
