@@ -2,6 +2,7 @@
 
 #include "PerfCounters.h"
 
+#include <dirent.h>
 #include <err.h>
 #include <fcntl.h>
 #include <linux/perf_event.h>
@@ -15,6 +16,9 @@
 #include <unistd.h>
 
 #include <algorithm>
+#include <fstream>
+#include <limits>
+#include <regex>
 #include <string>
 
 #include "Flags.h"
@@ -34,16 +38,26 @@ namespace rr {
 static bool attributes_initialized;
 // At some point we might support multiple kinds of ticks for the same CPU arch.
 // At that point this will need to become more complicated.
-static struct perf_event_attr ticks_attr;
-static struct perf_event_attr minus_ticks_attr;
-static struct perf_event_attr cycles_attr;
-static struct perf_event_attr hw_interrupts_attr;
-static struct perf_event_attr llsc_fail_attr;
-static uint32_t pmu_flags;
-static uint32_t skid_size;
-static bool has_ioc_period_bug;
-static bool only_one_counter;
-static bool activate_useless_counter;
+struct perf_event_attrs {
+  // bug_flags is an architecture dependent flags to determine
+  // what bugs need to be checked.
+  // Current, this is simply the uarch on x86 and unused on aarch64.
+  int bug_flags = 0;
+  perf_event_attr ticks{};
+  perf_event_attr minus_ticks{};
+  perf_event_attr cycles{};
+  perf_event_attr llsc_fail{};
+  const char *pmu_name = nullptr;
+  uint32_t pmu_flags = 0;
+  uint32_t skid_size = 0;
+  bool checked = false;
+  bool has_ioc_period_bug = false;
+  bool only_one_counter = false;
+  bool activate_useless_counter = false;
+};
+// If this contains more than one element, it's indexed by the CPU index.
+static std::vector<perf_event_attrs> perf_attrs;
+static uint32_t pmu_semantics_flags;
 
 /*
  * Find out the cpu model using the cpuid instruction.
@@ -69,14 +83,27 @@ enum CpuMicroarch {
   IntelCometlake,
   IntelIcelake,
   IntelTigerlake,
-  LastIntel = IntelTigerlake,
+  IntelRocketlake,
+  IntelAlderlake,
+  LastIntel = IntelAlderlake,
   FirstAMD,
   AMDF15R30 = FirstAMD,
   AMDZen,
   LastAMD = AMDZen,
   FirstARM,
   ARMNeoverseN1 = FirstARM,
-  LastARM = ARMNeoverseN1,
+  ARMNeoverseE1,
+  ARMNeoverseV1,
+  ARMNeoverseN2,
+  ARMCortexA55,
+  ARMCortexA75,
+  ARMCortexA76,
+  ARMCortexA77,
+  ARMCortexA78,
+  ARMCortexX1,
+  AppleM1Icestorm,
+  AppleM1Firestorm,
+  LastARM = AppleM1Firestorm,
 };
 
 /*
@@ -106,10 +133,13 @@ struct PmuConfig {
   const char* name;
   unsigned rcb_cntr_event;
   unsigned minus_ticks_cntr_event;
-  unsigned hw_intr_cntr_event;
   unsigned llsc_cntr_event;
   uint32_t skid_size;
   uint32_t flags;
+  const char* pmu_name = nullptr; // ARM only
+  unsigned cycle_event = PERF_COUNT_HW_CPU_CYCLES;
+  int cycle_type = PERF_TYPE_HARDWARE;
+  int event_type = PERF_TYPE_RAW;
 };
 
 // XXX please only edit this if you really know what you're doing.
@@ -124,29 +154,64 @@ struct PmuConfig {
 // See Intel 64 and IA32 Architectures Performance Monitoring Events.
 // See check_events from libpfm4.
 static const PmuConfig pmu_configs[] = {
-  { IntelTigerlake, "Intel Tigerlake", 0x5111c4, 0, 0, 0, 100, PMU_TICKS_RCB },
-  { IntelIcelake, "Intel Icelake", 0x5111c4, 0, 0, 0, 100, PMU_TICKS_RCB },
-  { IntelCometlake, "Intel Cometlake", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelKabylake, "Intel Kabylake", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelSilvermont, "Intel Silvermont", 0x517ec4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelGoldmont, "Intel Goldmont", 0x517ec4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelSkylake, "Intel Skylake", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelBroadwell, "Intel Broadwell", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelHaswell, "Intel Haswell", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelIvyBridge, "Intel Ivy Bridge", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelSandyBridge, "Intel Sandy Bridge", 0x5101c4, 0, 0x5301cb, 0, 100, PMU_TICKS_RCB },
-  { IntelNehalem, "Intel Nehalem", 0x5101c4, 0, 0x50011d, 0, 100, PMU_TICKS_RCB },
-  { IntelWestmere, "Intel Westmere", 0x5101c4, 0, 0x50011d, 0, 100, PMU_TICKS_RCB },
-  { IntelPenryn, "Intel Penryn", 0, 0, 0, 0, 100, 0 },
-  { IntelMerom, "Intel Merom", 0, 0, 0, 0, 100, 0 },
-  { AMDF15R30, "AMD Family 15h Revision 30h", 0xc4, 0xc6, 0, 0, 250, PMU_TICKS_TAKEN_BRANCHES },
+  { IntelAlderlake, "Intel Alderlake", 0x5111c4, 0, 0, 125, PMU_TICKS_RCB },
+  { IntelRocketlake, "Intel Rocketlake", 0x5111c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelTigerlake, "Intel Tigerlake", 0x5111c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelIcelake, "Intel Icelake", 0x5111c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelCometlake, "Intel Cometlake", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelKabylake, "Intel Kabylake", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelSilvermont, "Intel Silvermont", 0x517ec4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelGoldmont, "Intel Goldmont", 0x517ec4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelSkylake, "Intel Skylake", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelBroadwell, "Intel Broadwell", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelHaswell, "Intel Haswell", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelIvyBridge, "Intel Ivy Bridge", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelSandyBridge, "Intel Sandy Bridge", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelNehalem, "Intel Nehalem", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelWestmere, "Intel Westmere", 0x5101c4, 0, 0, 100, PMU_TICKS_RCB },
+  { IntelPenryn, "Intel Penryn", 0, 0, 0, 100, 0 },
+  { IntelMerom, "Intel Merom", 0, 0, 0, 100, 0 },
+  { AMDF15R30, "AMD Family 15h Revision 30h", 0xc4, 0xc6, 0, 250, PMU_TICKS_TAKEN_BRANCHES },
   // 0xd1 == RETIRED_CONDITIONAL_BRANCH_INSTRUCTIONS - Number of retired conditional branch instructions
   // 0x2c == INTERRUPT_TAKEN - Counts the number of interrupts taken
   // Both counters are available on Zen, Zen+ and Zen2.
-  { AMDZen, "AMD Zen", 0x5100d1, 0, 0x51002c, 0, 10000, PMU_TICKS_RCB },
+  { AMDZen, "AMD Zen", 0x5100d1, 0, 0, 10000, PMU_TICKS_RCB },
+  // Performance cores from ARM from cortex-a76 on (including neoverse-n1 and later)
+  // have the following counters that are reliable enough for us.
   // 0x21 == BR_RETIRED - Architecturally retired taken branches
   // 0x6F == STREX_SPEC - Speculatively executed strex instructions
-  { ARMNeoverseN1, "ARM Neoverse N1", 0x21, 0, 0, 0x6F, 1000, PMU_TICKS_TAKEN_BRANCHES }
+  // 0x11 == CPU_CYCLES - Cycle
+  { ARMNeoverseN1, "ARM Neoverse N1", 0x21, 0, 0x6F, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3_0", 0x11, -1, -1 },
+  { ARMNeoverseV1, "ARM Neoverse V1", 0x21, 0, 0x6F, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3_0", 0x11, -1, -1 },
+  { ARMNeoverseN2, "ARM Neoverse N2", 0x21, 0, 0x6F, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3_0", 0x11, -1, -1 },
+  { ARMCortexA76, "ARM Cortex A76", 0x21, 0, 0x6F, 10000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3", 0x11, -1, -1 },
+  { ARMCortexA77, "ARM Cortex A77", 0x21, 0, 0x6F, 10000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3", 0x11, -1, -1 },
+  { ARMCortexA78, "ARM Cortex A78", 0x21, 0, 0x6F, 10000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3", 0x11, -1, -1 },
+  { ARMCortexX1, "ARM Cortex X1", 0x21, 0, 0x6F, 10000, PMU_TICKS_TAKEN_BRANCHES,
+    "armv8_pmuv3", 0x11, -1, -1 },
+  // cortex-a55, cortex-a75 and neoverse-e1 counts uarch ISB
+  // as retired branches so the BR_RETIRED counter is not reliable.
+  // There are some counters that are somewhat more reliable than
+  // the total branch count (0x21) including
+  // 0x0D (BR_IMMED_RETIRED) 0x0E (BR_RETURN_RETIRED)
+  // 0xCD (BR_INDIRECT_ADDR_PRED) 0x76 (PC_WRITE_SPEC)
+  // 0x78 (BR_IMMED_SPEC), 0xC9 (BR_COND_PRED)
+  // 0xCD (BR_INDIRECT_ADDR_PRED)
+  // but according to tests on the LITTLE core on a snapdragon 865
+  // none of them (including the sums) seems to be useful/reliable enough.
+  { ARMNeoverseE1, "ARM Neoverse E1", 0, 0, 0, 0, 0 },
+  { ARMCortexA55, "ARM Cortex A55", 0, 0, 0, 0, 0 },
+  { ARMCortexA75, "ARM Cortex A75", 0, 0, 0, 0, 0 },
+  { AppleM1Icestorm, "Apple M1 Icestorm", 0x90, 0, 0, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "apple_icestorm_pmu", 0x8c, -1, -1 },
+  { AppleM1Firestorm, "Apple M1 Firestorm", 0x90, 0, 0, 1000, PMU_TICKS_TAKEN_BRANCHES,
+    "apple_firestorm_pmu", 0x8c, -1, -1 },
 };
 
 #define RR_SKID_MAX 10000
@@ -157,10 +222,34 @@ static string lowercase(const string& s) {
   return c;
 }
 
+// The index of the PMU we are using within perf_attrs.
+// This is always 0 if we detected a single PMU type
+// and will be the same as the CPU index if we detected multiple PMU types.
+static int get_pmu_index(int cpu_binding)
+{
+  if (cpu_binding < 0) {
+    if (perf_attrs.size() > 1) {
+      CLEAN_FATAL() << "\nMultiple PMU types detected. Unbinding CPU is not supported.";
+    }
+    return 0;
+  }
+  if (!PerfCounters::support_cpu(cpu_binding)) {
+    CLEAN_FATAL() << "\nPMU on cpu " << cpu_binding << " is not supported.";
+  }
+  if (perf_attrs.size() == 1) {
+    // Single PMU type.
+    return 0;
+  }
+  if ((size_t)cpu_binding > perf_attrs.size()) {
+    CLEAN_FATAL() << "\nUnable to find PMU type for CPU " << cpu_binding;
+  }
+  return cpu_binding;
+}
+
 static void init_perf_event_attr(struct perf_event_attr* attr,
-                                 perf_type_id type, unsigned config) {
+                                 unsigned type, unsigned config) {
   memset(attr, 0, sizeof(*attr));
-  attr->type = type;
+  attr->type = perf_type_id(type);
   attr->size = sizeof(*attr);
   attr->config = config;
   // rr requires that its events count userspace tracee code
@@ -219,12 +308,12 @@ static ScopedFd start_counter(pid_t tid, int group_fd,
     }
     FATAL() << "Failed to initialize counter";
   }
-  return fd;
+  return ScopedFd(fd);
 }
 
-static void check_for_ioc_period_bug() {
+static void check_for_ioc_period_bug(perf_event_attrs &perf_attr) {
   // Start a cycles counter
-  struct perf_event_attr attr = rr::ticks_attr;
+  struct perf_event_attr attr = perf_attr.ticks;
   attr.sample_period = 0xffffffff;
   attr.exclude_kernel = 1;
   ScopedFd bug_fd = start_counter(0, -1, &attr);
@@ -237,8 +326,8 @@ static void check_for_ioc_period_bug() {
   struct pollfd poll_bug_fd = {.fd = bug_fd, .events = POLL_IN, .revents = 0 };
   poll(&poll_bug_fd, 1, 0);
 
-  has_ioc_period_bug = poll_bug_fd.revents == 0;
-  LOG(debug) << "has_ioc_period_bug=" << has_ioc_period_bug;
+  perf_attr.has_ioc_period_bug = poll_bug_fd.revents == 0;
+  LOG(debug) << "has_ioc_period_bug=" << perf_attr.has_ioc_period_bug;
 }
 
 static const int NUM_BRANCHES = 500;
@@ -265,11 +354,11 @@ static void do_branches() {
 #error Must define microarchitecture detection code for this architecture
 #endif
 
-static void check_working_counters() {
-  struct perf_event_attr attr = rr::ticks_attr;
-  attr.sample_period = 0;
-  struct perf_event_attr attr2 = rr::cycles_attr;
+static void check_working_counters(perf_event_attrs &perf_attr) {
+  struct perf_event_attr attr = perf_attr.ticks;
   attr.sample_period = 0;
+  struct perf_event_attr attr2 = perf_attr.cycles;
+  attr2.sample_period = 0;
   ScopedFd fd = start_counter(0, -1, &attr);
   ScopedFd fd2 = start_counter(0, -1, &attr2);
   do_branches();
@@ -278,16 +367,23 @@ static void check_working_counters() {
 
   if (events < NUM_BRANCHES) {
     char config[100];
-    sprintf(config, "%llx", (long long)ticks_attr.config);
+    sprintf(config, "%llx", (long long)perf_attr.ticks.config);
+    std::string perf_cmdline = "perf stat -e ";
+    if (perf_attr.pmu_name) {
+      perf_cmdline = perf_cmdline + perf_attr.pmu_name + "/r" + config + "/ true";
+    }
+    else {
+      perf_cmdline = perf_cmdline + "r" + config + " true";
+    }
     FATAL()
         << "\nGot " << events << " branch events, expected at least "
         << NUM_BRANCHES
         << ".\n"
            "\nThe hardware performance counter seems to not be working. Check\n"
            "that hardware performance counters are working by running\n"
-           "  perf stat -e r"
-        << config
-        << " true\n"
+           "  "
+        << perf_cmdline
+        << "\n"
            "and checking that it reports a nonzero number of events.\n"
            "If performance counters seem to be working with 'perf', file an\n"
            "rr issue, otherwise check your hardware/OS/VM configuration. Also\n"
@@ -295,27 +391,23 @@ static void check_working_counters() {
            "this CPU.";
   }
 
-  only_one_counter = events2 == 0;
-  LOG(debug) << "only_one_counter=" << only_one_counter;
+  perf_attr.only_one_counter = events2 == 0;
+  LOG(debug) << "only_one_counter=" << perf_attr.only_one_counter;
 
-  if (only_one_counter) {
+  if (perf_attr.only_one_counter) {
     arch_check_restricted_counter();
   }
 }
 
-static void check_for_bugs(CpuMicroarch uarch) {
-  if (running_under_rr()) {
-    // Under rr we emulate idealized performance counters, so we can assume
-    // none of the bugs apply.
-    return;
-  }
+static void check_for_bugs(perf_event_attrs &perf_attr) {
+  DEBUG_ASSERT(!running_under_rr());
 
-  check_for_ioc_period_bug();
-  check_working_counters();
-  check_for_arch_bugs(uarch);
+  check_for_ioc_period_bug(perf_attr);
+  check_working_counters(perf_attr);
+  check_for_arch_bugs(perf_attr);
 }
 
-static CpuMicroarch get_cpu_microarch() {
+static std::vector<CpuMicroarch> get_cpu_microarchs() {
   string forced_uarch = lowercase(Flags::get().forced_uarch);
   if (!forced_uarch.empty()) {
     for (size_t i = 0; i < array_length(pmu_configs); ++i) {
@@ -323,13 +415,43 @@ static CpuMicroarch get_cpu_microarch() {
       string name = lowercase(pmu.name);
       if (name.npos != name.find(forced_uarch)) {
         LOG(info) << "Using forced uarch " << pmu.name;
-        return pmu.uarch;
+        return { pmu.uarch };
       }
     }
     CLEAN_FATAL() << "Forced uarch " << Flags::get().forced_uarch
                   << " isn't known.";
   }
-  return compute_cpu_microarch();
+  return compute_cpu_microarchs();
+}
+
+// Similar to rr::perf_attrs, if this contains more than one element,
+// it's indexed by the CPU index.
+static std::vector<PmuConfig> get_pmu_microarchs() {
+  std::vector<PmuConfig> pmu_uarchs;
+  auto uarchs = get_cpu_microarchs();
+  bool found_working_pmu = false;
+  for (auto uarch : uarchs) {
+    bool found = false;
+    for (size_t i = 0; i < array_length(pmu_configs); ++i) {
+      if (uarch == pmu_configs[i].uarch) {
+        found = true;
+        if (pmu_configs[i].flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES)) {
+          found_working_pmu |= true;
+        }
+        pmu_uarchs.push_back(pmu_configs[i]);
+        break;
+      }
+    }
+    DEBUG_ASSERT(found);
+  }
+  if (!found_working_pmu) {
+    CLEAN_FATAL() << "No supported microarchitectures found.";
+  }
+  DEBUG_ASSERT(!pmu_uarchs.empty());
+  // Note that the `uarch` field after processed by `post_init_pmu_uarchs`
+  // is used to store the bug_flags and may not be the actual uarch.
+  post_init_pmu_uarchs(pmu_uarchs);
+  return pmu_uarchs;
 }
 
 static void init_attributes() {
@@ -338,58 +460,108 @@ static void init_attributes() {
   }
   attributes_initialized = true;
 
-  CpuMicroarch uarch = get_cpu_microarch();
-  const PmuConfig* pmu = nullptr;
-  for (size_t i = 0; i < array_length(pmu_configs); ++i) {
-    if (uarch == pmu_configs[i].uarch) {
-      pmu = &pmu_configs[i];
-      break;
+  auto pmu_uarchs = get_pmu_microarchs();
+  pmu_semantics_flags = PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES;
+  for (auto &pmu_uarch : pmu_uarchs) {
+    if (!(pmu_uarch.flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES))) {
+      continue;
     }
+    pmu_semantics_flags = pmu_semantics_flags & pmu_uarch.flags;
   }
-  DEBUG_ASSERT(pmu);
-
-  if (!(pmu->flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES))) {
-    FATAL() << "Microarchitecture `" << pmu->name << "' currently unsupported.";
+  if (!(pmu_semantics_flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES))) {
+    if (pmu_uarchs.size() == 1) {
+      FATAL() << "Microarchitecture `" << pmu_uarchs[0].name
+              << "' currently unsupported.";
+    } else {
+      std::string uarch_list;
+      for (auto &pmu_uarch : pmu_uarchs) {
+        uarch_list += "\n  ";
+        uarch_list += pmu_uarch.name;
+      }
+      FATAL() << "Microarchitecture combination currently unsupported:"
+              << uarch_list;
+    }
   }
 
   if (running_under_rr()) {
-    init_perf_event_attr(&ticks_attr, PERF_TYPE_HARDWARE, PERF_COUNT_RR);
-    skid_size = RR_SKID_MAX;
-    pmu_flags = pmu->flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES);
+    perf_attrs.resize(1);
+    init_perf_event_attr(&perf_attrs[0].ticks, PERF_TYPE_HARDWARE, PERF_COUNT_RR);
+    perf_attrs[0].skid_size = RR_SKID_MAX;
+    perf_attrs[0].pmu_flags = pmu_semantics_flags;
   } else {
-    skid_size = pmu->skid_size;
-    pmu_flags = pmu->flags;
-    init_perf_event_attr(&ticks_attr, PERF_TYPE_RAW, pmu->rcb_cntr_event);
-    if (pmu->minus_ticks_cntr_event != 0) {
-      init_perf_event_attr(&minus_ticks_attr, PERF_TYPE_RAW,
-                           pmu->minus_ticks_cntr_event);
+    auto npmus = pmu_uarchs.size();
+    perf_attrs.resize(npmus);
+    for (size_t i = 0; i < npmus; i++) {
+      auto &perf_attr = perf_attrs[i];
+      auto &pmu_uarch = pmu_uarchs[i];
+      if (!(pmu_uarch.flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES))) {
+        perf_attr.pmu_flags = 0; // Mark as unsupported
+        continue;
+      }
+      perf_attr.pmu_name = pmu_uarch.pmu_name;
+      perf_attr.skid_size = pmu_uarch.skid_size;
+      perf_attr.pmu_flags = pmu_uarch.flags;
+      perf_attr.bug_flags = (int)pmu_uarch.uarch;
+      init_perf_event_attr(&perf_attr.ticks, pmu_uarch.event_type,
+                           pmu_uarch.rcb_cntr_event);
+      if (pmu_uarch.minus_ticks_cntr_event != 0) {
+        init_perf_event_attr(&perf_attr.minus_ticks, pmu_uarch.event_type,
+                             pmu_uarch.minus_ticks_cntr_event);
+      }
+      init_perf_event_attr(&perf_attr.cycles, pmu_uarch.cycle_type,
+                           pmu_uarch.cycle_event);
+      init_perf_event_attr(&perf_attr.llsc_fail, pmu_uarch.event_type,
+                           pmu_uarch.llsc_cntr_event);
     }
-    init_perf_event_attr(&cycles_attr, PERF_TYPE_HARDWARE,
-                         PERF_COUNT_HW_CPU_CYCLES);
-    init_perf_event_attr(&hw_interrupts_attr, PERF_TYPE_RAW,
-                         pmu->hw_intr_cntr_event);
-    init_perf_event_attr(&llsc_fail_attr, PERF_TYPE_RAW,
-                         pmu->llsc_cntr_event);
-    // libpfm encodes the event with this bit set, so we'll do the
-    // same thing.  Unclear if necessary.
-    hw_interrupts_attr.exclude_hv = 1;
-
-    check_for_bugs(uarch);
-    /*
-     * For maintainability, and since it doesn't impact performance when not
-     * needed, we always activate this. If it ever turns out to be a problem,
-     * this can be set to pmu->flags & PMU_BENEFITS_FROM_USELESS_COUNTER,
-     * instead.
-     *
-     * We also disable this counter when running under rr. Even though it's the
-     * same event for the same task as the outer rr, the linux kernel does not
-     * coalesce them and tries to schedule the new one on a general purpose PMC.
-     * On CPUs with only 2 general PMCs (e.g. KNL), we'd run out.
-     */
-    activate_useless_counter = has_ioc_period_bug && !running_under_rr();
   }
 }
 
+bool PerfCounters::support_cpu(int cpu)
+{
+  // We could probably make cpu=-1 mean whether all CPUs are supported
+  // if there's a need for it...
+  DEBUG_ASSERT(cpu >= 0);
+  init_attributes();
+
+  auto nattrs = (int)perf_attrs.size();
+  if (nattrs == 1) {
+    cpu = 0;
+  }
+  if (cpu >= nattrs) {
+    return false;
+  }
+  auto &perf_attr = perf_attrs[cpu];
+  return perf_attr.pmu_flags & (PMU_TICKS_RCB | PMU_TICKS_TAKEN_BRANCHES);
+}
+
+static void check_pmu(int pmu_index) {
+  auto &perf_attr = perf_attrs[pmu_index];
+  if (perf_attr.checked) {
+    return;
+  }
+  perf_attr.checked = true;
+
+  // Under rr we emulate idealized performance counters, so we can assume
+  // none of the bugs apply.
+  if (running_under_rr()) {
+    return;
+  }
+
+  check_for_bugs(perf_attr);
+  /*
+   * For maintainability, and since it doesn't impact performance when not
+   * needed, we always activate this. If it ever turns out to be a problem,
+   * this can be set to pmu->flags & PMU_BENEFITS_FROM_USELESS_COUNTER,
+   * instead.
+   *
+   * We also disable this counter when running under rr. Even though it's the
+   * same event for the same task as the outer rr, the linux kernel does not
+   * coalesce them and tries to schedule the new one on a general purpose PMC.
+   * On CPUs with only 2 general PMCs (e.g. KNL), we'd run out.
+   */
+  perf_attr.activate_useless_counter = perf_attr.has_ioc_period_bug;
+}
+
 bool PerfCounters::is_rr_ticks_attr(const perf_event_attr& attr) {
   return attr.type == PERF_TYPE_HARDWARE && attr.config == PERF_COUNT_RR;
 }
@@ -398,9 +570,9 @@ bool PerfCounters::supports_ticks_semantics(TicksSemantics ticks_semantics) {
   init_attributes();
   switch (ticks_semantics) {
   case TICKS_RETIRED_CONDITIONAL_BRANCHES:
-    return (pmu_flags & PMU_TICKS_RCB) != 0;
+    return (pmu_semantics_flags & PMU_TICKS_RCB) != 0;
   case TICKS_TAKEN_BRANCHES:
-    return (pmu_flags & PMU_TICKS_TAKEN_BRANCHES) != 0;
+    return (pmu_semantics_flags & PMU_TICKS_TAKEN_BRANCHES) != 0;
   default:
     FATAL() << "Unknown ticks_semantics " << ticks_semantics;
     return false;
@@ -409,10 +581,10 @@ bool PerfCounters::supports_ticks_semantics(TicksSemantics ticks_semantics) {
 
 TicksSemantics PerfCounters::default_ticks_semantics() {
   init_attributes();
-  if (pmu_flags & PMU_TICKS_TAKEN_BRANCHES) {
+  if (pmu_semantics_flags & PMU_TICKS_TAKEN_BRANCHES) {
     return TICKS_TAKEN_BRANCHES;
   }
-  if (pmu_flags & PMU_TICKS_RCB) {
+  if (pmu_semantics_flags & PMU_TICKS_RCB) {
     return TICKS_RETIRED_CONDITIONAL_BRANCHES;
   }
   FATAL() << "Unsupported architecture";
@@ -420,12 +592,14 @@ TicksSemantics PerfCounters::default_ticks_semantics() {
 }
 
 uint32_t PerfCounters::skid_size() {
-  init_attributes();
-  return rr::skid_size;
+  DEBUG_ASSERT(attributes_initialized);
+  DEBUG_ASSERT(perf_attrs[pmu_index].checked);
+  return perf_attrs[pmu_index].skid_size;
 }
 
-PerfCounters::PerfCounters(pid_t tid, TicksSemantics ticks_semantics)
-    : tid(tid), ticks_semantics_(ticks_semantics), started(false), counting(false) {
+PerfCounters::PerfCounters(pid_t tid, int cpu_binding,
+                           TicksSemantics ticks_semantics)
+    : tid(tid), pmu_index(get_pmu_index(cpu_binding)), ticks_semantics_(ticks_semantics), started(false), counting(false) {
   if (!supports_ticks_semantics(ticks_semantics)) {
     FATAL() << "Ticks semantics " << ticks_semantics << " not supported";
   }
@@ -440,8 +614,10 @@ static void make_counter_async(ScopedFd& fd, int signal) {
 
 void PerfCounters::reset(Ticks ticks_period) {
   DEBUG_ASSERT(ticks_period >= 0);
+  check_pmu(pmu_index);
 
-  if (ticks_period == 0 && !always_recreate_counters()) {
+  auto &perf_attr = perf_attrs[pmu_index];
+  if (ticks_period == 0 && !always_recreate_counters(perf_attr)) {
     // We can't switch a counter between sampling and non-sampling via
     // PERF_EVENT_IOC_PERIOD so just turn 0 into a very big number.
     ticks_period = uint64_t(1) << 60;
@@ -450,22 +626,22 @@ void PerfCounters::reset(Ticks ticks_period) {
   if (!started) {
     LOG(debug) << "Recreating counters with period " << ticks_period;
 
-    struct perf_event_attr attr = rr::ticks_attr;
-    struct perf_event_attr minus_attr = rr::minus_ticks_attr;
+    struct perf_event_attr attr = perf_attr.ticks;
+    struct perf_event_attr minus_attr = perf_attr.minus_ticks;
     attr.sample_period = ticks_period;
     fd_ticks_interrupt = start_counter(tid, -1, &attr);
     if (minus_attr.config != 0) {
       fd_minus_ticks_measure = start_counter(tid, fd_ticks_interrupt, &minus_attr);
     }
 
-    if (!only_one_counter && !running_under_rr()) {
+    if (!perf_attr.only_one_counter && !running_under_rr()) {
       reset_arch_extras<NativeArch>();
     }
 
-    if (activate_useless_counter && !fd_useless_counter.is_open()) {
+    if (perf_attr.activate_useless_counter && !fd_useless_counter.is_open()) {
       // N.B.: This is deliberately not in the same group as the other counters
       // since we want to keep it scheduled at all times.
-      fd_useless_counter = start_counter(tid, -1, &cycles_attr);
+      fd_useless_counter = start_counter(tid, -1, &perf_attr.cycles);
     }
 
     struct f_owner_ex own;
@@ -542,7 +718,7 @@ void PerfCounters::stop_counting() {
     return;
   }
   counting = false;
-  if (always_recreate_counters()) {
+  if (always_recreate_counters(perf_attrs[pmu_index])) {
     stop();
   } else {
     ioctl(fd_ticks_interrupt, PERF_EVENT_IOC_DISABLE, 0);
@@ -558,12 +734,20 @@ void PerfCounters::stop_counting() {
   }
 }
 
+// Note that on aarch64 this is also used to get the count for `ret`
 Ticks PerfCounters::ticks_for_unconditional_indirect_branch(Task*) {
-  return (pmu_flags & PMU_TICKS_TAKEN_BRANCHES) ? 1 : 0;
+  DEBUG_ASSERT(attributes_initialized);
+  return (pmu_semantics_flags & PMU_TICKS_TAKEN_BRANCHES) ? 1 : 0;
+}
+
+Ticks PerfCounters::ticks_for_unconditional_direct_branch(Task*) {
+  DEBUG_ASSERT(attributes_initialized);
+  return (pmu_semantics_flags & PMU_TICKS_TAKEN_BRANCHES) ? 1 : 0;
 }
 
 Ticks PerfCounters::ticks_for_direct_call(Task*) {
-  return (pmu_flags & PMU_TICKS_TAKEN_BRANCHES) ? 1 : 0;
+  DEBUG_ASSERT(attributes_initialized);
+  return (pmu_semantics_flags & PMU_TICKS_TAKEN_BRANCHES) ? 1 : 0;
 }
 
 Ticks PerfCounters::read_ticks(Task* t) {
