@@ -42,7 +42,7 @@ public:
   /**
    * Create performance counters monitoring the given task.
    */
-  PerfCounters(pid_t tid, TicksSemantics ticks_semantics);
+  PerfCounters(pid_t tid, int cpu_binding, TicksSemantics ticks_semantics);
   ~PerfCounters() { stop(); }
 
   void set_tid(pid_t tid);
@@ -78,11 +78,20 @@ public:
    * Return the number of ticks we need for an emulated branch.
    */
   static Ticks ticks_for_unconditional_indirect_branch(Task*);
+  /**
+   * Return the number of ticks we need for an emulated direct branch.
+   */
+  static Ticks ticks_for_unconditional_direct_branch(Task*);
   /**
    * Return the number of ticks we need for a direct call.
    */
   static Ticks ticks_for_direct_call(Task*);
 
+  /**
+   * Whether PMU on core i is supported.
+   */
+  static bool support_cpu(int cpu);
+
   /**
    * Read the current value of the ticks counter.
    * `t` is used for debugging purposes.
@@ -113,19 +122,20 @@ public:
    * When an interrupt is requested, at most this many ticks may elapse before
    * the interrupt is delivered.
    */
-  static uint32_t skid_size();
+  uint32_t skid_size();
 
   /**
    * Use a separate skid_size for recording since we seem to see more skid
    * in practice during recording, in particular during the
    * async_signal_syscalls tests
    */
-  static uint32_t recording_skid_size() { return skid_size() * 5; }
+  uint32_t recording_skid_size() { return skid_size() * 5; }
 
 private:
   // Only valid while 'counting' is true
   Ticks counting_period;
   pid_t tid;
+  int pmu_index;
   // We use separate fds for counting ticks and for generating interrupts. The
   // former ignores ticks in aborted transactions, and does not support
   // sample_period; the latter does not ignore ticks in aborted transactions,
