@@ -70,7 +70,8 @@ public:
       const TraceUuid* trace_id = nullptr,
       bool use_audit = false,
       bool unmap_vdso = false,
-      bool force_asan_active = false);
+      bool force_asan_active = false,
+      bool force_tsan_active = false);
 
   const DisableCPUIDFeatures& disable_cpuid_features() const {
     return disable_cpuid_features_;
@@ -84,8 +85,14 @@ public:
   int get_ignore_sig() const { return ignore_sig; }
   void set_continue_through_sig(int sig) { continue_through_sig = sig; }
   int get_continue_through_sig() const { return continue_through_sig; }
-  void set_asan_active(bool active) { asan_active_ = active; }
-  bool asan_active() const { return asan_active_; }
+  // Returns ranges to exclude from chaos mode memory allocation.
+  // Used to exclude ranges used by sanitizers.
+  const std::vector<MemoryRange> excluded_ranges() const {
+    return excluded_ranges_;
+  }
+  MemoryRange fixed_global_exclusion_range() const {
+    return fixed_global_exclusion_range_;
+  }
   bool use_audit() const { return use_audit_; }
   bool unmap_vdso() { return unmap_vdso_; }
   uint64_t rr_signal_mask() const;
@@ -172,6 +179,7 @@ public:
 
   RecordTask* find_task(pid_t rec_tid) const;
   RecordTask* find_task(const TaskUid& tuid) const;
+  RecordTask* find_detached_proxy_task(pid_t proxt_tid) const;
 
   void on_proxy_detach(RecordTask *t, pid_t new_tid);
 
@@ -193,6 +201,11 @@ public:
    */
   void term_detached_tasks();
 
+  /**
+   * Forward SIGTERM to initial task
+   */
+  void forward_SIGTERM();
+
 private:
   RecordSession(const std::string& exe_path,
                 const std::vector<std::string>& argv,
@@ -249,12 +262,18 @@ private:
    * When true, try to increase the probability of finding bugs.
    */
   bool enable_chaos_;
-  bool asan_active_;
   /**
    * When true, wait for all tracees to exit before finishing recording.
    */
   bool wait_for_all_;
 
+  std::vector<MemoryRange> excluded_ranges_;
+  MemoryRange fixed_global_exclusion_range_;
+  /**
+   * Keeps track of detached tasks.
+   */
+  std::map<pid_t, RecordTask*> detached_task_map;
+
   std::string output_trace_dir;
 
   bool use_audit_;
