@@ -191,6 +191,7 @@ RecordTask::RecordTask(RecordSession& session, pid_t _tid, uint32_t serial,
       break_at_syscallbuf_traced_syscalls(false),
       break_at_syscallbuf_untraced_syscalls(false),
       break_at_syscallbuf_final_instruction(false),
+      syscallstub_exit_breakpoint(),
       next_pmc_interrupt_is_for_user(false),
       did_record_robust_futex_changes(false),
       waiting_for_reap(false),
@@ -198,7 +199,9 @@ RecordTask::RecordTask(RecordSession& session, pid_t _tid, uint32_t serial,
       waiting_for_ptrace_exit(false),
       retry_syscall_patching(false),
       sent_shutdown_kill(false),
-      tick_request_override((TicksRequest)0) {
+      did_execveat(false),
+      tick_request_override((TicksRequest)0),
+      schedule_frozen(false) {
   push_event(Event::sentinel());
   if (session.tasks().empty()) {
     // Initial tracee. It inherited its state from this process, so set it up.
@@ -253,15 +256,11 @@ RecordTask::~RecordTask() {
   }
 }
 
-void RecordTask::record_exit_event(int fatal_signo, WriteChildTid write_child_tid) {
+void RecordTask::record_exit_event(WriteChildTid write_child_tid) {
   // The kernel explicitly only clears the futex if the address space is shared.
   // If the address space has no other users then the futex will not be cleared
   // even if it lives in shared memory which other tasks can read.
-  // If however, the exit was the result of a fatal, core-dump signal, the futex
-  // is not cleared (both to preserve the coredump and because any other users
-  // of the same address space were also shot down).
-  if (!is_coredumping_signal(fatal_signo) &&
-     !tid_futex.is_null() && as->task_set().size() > 1 &&
+  if (!tid_futex.is_null() && as->task_set().size() > 1 &&
      as->has_mapping(tid_futex)) {
     int val = 0;
     record_local(tid_futex, &val);
@@ -343,18 +342,19 @@ void RecordTask::post_wait_clone(Task* cloned_from, int flags) {
 }
 
 void RecordTask::post_exec() {
-  // Change syscall number to execve *for the new arch*. If we don't do this,
+  // Change syscall number to execve/execveat *for the new arch*. If we don't do this,
   // and the arch changes, then the syscall number for execve in the old arch/
   // is treated as the syscall we're executing in the new arch, with hilarious
   // results.
-  int syscallno = syscall_number_for_execve(arch());
-  registers.set_original_syscallno(syscallno);
+  int new_syscallno = did_execveat ? syscall_number_for_execveat(arch())
+      : syscall_number_for_execve(arch());
+  registers.set_original_syscallno(new_syscallno);
   // Fix event architecture and syscall number
-  ev().Syscall().number = syscallno;
+  ev().Syscall().number = new_syscallno;
   ev().Syscall().set_arch(arch());
 
   // The signal mask is inherited across execve so we don't need to invalidate.
-  Task::post_exec(this->exe_path());
+  Task::post_exec(exe_path(), exe_path());
   if (emulated_ptracer) {
     ASSERT(this, !(emulated_ptracer->arch() == x86 && arch() == x86_64))
         << "We don't support a 32-bit process tracing a 64-bit process";
@@ -374,8 +374,10 @@ void RecordTask::post_exec() {
 
 template <typename Arch> static void do_preload_init_arch(RecordTask* t) {
   auto params = t->read_mem(
-      remote_ptr<rrcall_init_preload_params<Arch>>(t->regs().arg1()));
+      remote_ptr<rrcall_init_preload_params<Arch>>(t->regs().orig_arg1()));
 
+  t->syscallbuf_code_layout.syscallbuf_syscall_hook =
+      params.syscallbuf_syscall_hook.rptr().as_int();
   t->syscallbuf_code_layout.syscallbuf_final_exit_instruction =
       params.syscallbuf_final_exit_instruction.rptr().as_int();
   t->syscallbuf_code_layout.syscallbuf_code_start =
@@ -392,11 +394,6 @@ template <typename Arch> static void do_preload_init_arch(RecordTask* t) {
   t->write_mem(in_chaos_ptr, in_chaos);
   t->record_local(in_chaos_ptr, &in_chaos);
 
-  int cores = t->session().scheduler().pretend_num_cores();
-  auto cores_ptr = REMOTE_PTR_FIELD(params.globals.rptr(), pretend_num_cores);
-  t->write_mem(cores_ptr, cores);
-  t->record_local(cores_ptr, &cores);
-
   auto desched_sig = t->session().syscallbuf_desched_sig();
   auto desched_sig_ptr = REMOTE_PTR_FIELD(params.globals.rptr(), desched_sig);
   t->write_mem(desched_sig_ptr, desched_sig);
@@ -447,7 +444,7 @@ template <typename Arch> void RecordTask::init_buffers_arch() {
   AutoRemoteSyscalls remote(this);
 
   // Arguments to the rrcall.
-  remote_ptr<rrcall_init_buffers_params<Arch>> child_args = regs().arg1();
+  remote_ptr<rrcall_init_buffers_params<Arch>> child_args = regs().orig_arg1();
   auto args = read_mem(child_args);
 
   args.cloned_file_data_fd = -1;
@@ -455,45 +452,56 @@ template <typename Arch> void RecordTask::init_buffers_arch() {
     args.syscallbuf_size = syscallbuf_size = session().syscall_buffer_size();
     KernelMapping syscallbuf_km = init_syscall_buffer(remote, nullptr);
     args.syscallbuf_ptr = syscallbuf_child;
+    if (syscallbuf_child != nullptr) {
+      // This needs to be skipped if we couldn't allocate the buffer
+      // since replaying only reads (and advances) the mmap record
+      // if `args.syscallbuf_ptr != nullptr`.
+      auto record_in_trace = trace_writer().write_mapped_region(
+        this, syscallbuf_km, syscallbuf_km.fake_stat(), syscallbuf_km.fsname(),
+        vector<TraceRemoteFd>(),
+        TraceWriter::RR_BUFFER_MAPPING);
+      ASSERT(this, record_in_trace == TraceWriter::DONT_RECORD_IN_TRACE);
+    } else {
+      // This can fail, e.g. if the tracee died unexpectedly.
+      LOG(debug) << "Syscallbuf initialization failed";
+      args.syscallbuf_size = 0;
+    }
+  } else {
+    args.syscallbuf_ptr = remote_ptr<void>(nullptr);
+    args.syscallbuf_size = 0;
+  }
+
+  if (args.syscallbuf_ptr) {
     desched_fd_child = args.desched_counter_fd;
     // Prevent the child from closing this fd
     fds->add_monitor(this, desched_fd_child, new PreserveFileMonitor());
     desched_fd = remote.retrieve_fd(desched_fd_child);
 
-    auto record_in_trace = trace_writer().write_mapped_region(
-        this, syscallbuf_km, syscallbuf_km.fake_stat(), syscallbuf_km.fsname(),
-        vector<TraceRemoteFd>(),
-        TraceWriter::RR_BUFFER_MAPPING);
-    ASSERT(this, record_in_trace == TraceWriter::DONT_RECORD_IN_TRACE);
-
     if (trace_writer().supports_file_data_cloning() &&
         session().use_read_cloning()) {
       cloned_file_data_fname = trace_writer().file_data_clone_file_name(tuid());
       ScopedFd clone_file(cloned_file_data_fname.c_str(), O_RDWR | O_CREAT, 0600);
-      int cloned_file_data = remote.send_fd(clone_file.get());
-      ASSERT(this, cloned_file_data >= 0);
-      int free_fd = find_free_file_descriptor(tid);
-      cloned_file_data_fd_child =
-          remote.syscall(syscall_number_for_dup3(arch()), cloned_file_data,
-                          free_fd, O_CLOEXEC);
-      if (cloned_file_data_fd_child != free_fd) {
-        ASSERT(this, cloned_file_data_fd_child < 0);
-        LOG(warn) << "Couldn't dup clone-data file to free fd";
-        cloned_file_data_fd_child = cloned_file_data;
-      } else {
-        // Prevent the child from closing this fd. We're going to close it
-        // ourselves and we don't want the child closing it and then reopening
-        // its own file with this fd.
-        fds->add_monitor(this, cloned_file_data_fd_child,
-                          new PreserveFileMonitor());
-        remote.infallible_syscall(syscall_number_for_close(arch()),
-                                  cloned_file_data);
+      int cloned_file_data = remote.infallible_send_fd_if_alive(clone_file);
+      if (cloned_file_data >= 0) {
+        int free_fd = find_free_file_descriptor(tid);
+        cloned_file_data_fd_child =
+            remote.syscall(syscall_number_for_dup3(arch()), cloned_file_data,
+                            free_fd, O_CLOEXEC);
+        if (cloned_file_data_fd_child != free_fd) {
+          ASSERT(this, cloned_file_data_fd_child < 0);
+          LOG(warn) << "Couldn't dup clone-data file to free fd";
+          cloned_file_data_fd_child = cloned_file_data;
+        } else {
+          // Prevent the child from closing this fd. We're going to close it
+          // ourselves and we don't want the child closing it and then reopening
+          // its own file with this fd.
+          fds->add_monitor(this, cloned_file_data_fd_child,
+                            new PreserveFileMonitor());
+          remote.infallible_close_syscall_if_alive(cloned_file_data);
+        }
+        args.cloned_file_data_fd = cloned_file_data_fd_child;
       }
-      args.cloned_file_data_fd = cloned_file_data_fd_child;
     }
-  } else {
-    args.syscallbuf_ptr = remote_ptr<void>(nullptr);
-    args.syscallbuf_size = 0;
   }
   args.scratch_buf = scratch_ptr;
   args.usable_scratch_size = usable_scratch_size();
@@ -579,7 +587,15 @@ bool RecordTask::is_at_syscallbuf_final_instruction_breakpoint() {
   return i == syscallbuf_code_layout.syscallbuf_final_exit_instruction;
 }
 
-void RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
+bool RecordTask::is_at_syscallstub_exit_breakpoint() {
+  if (!break_at_syscallbuf_final_instruction || !syscallstub_exit_breakpoint) {
+    return false;
+  }
+  auto i = ip().undo_executed_bkpt(arch());
+  return i == syscallstub_exit_breakpoint;
+}
+
+bool RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
                                        TicksRequest ticks_request, int sig) {
   // We may execute user code, which could lead to an RDTSC or grow-map
   // operation which unblocks SIGSEGV, and we'll need to know whether to
@@ -610,16 +626,8 @@ void RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
       // We're injecting a signal, so make sure that signal is unblocked.
       sigset &= ~signal_bit(sig);
     }
-    int ret = fallible_ptrace(PTRACE_SETSIGMASK, remote_ptr<void>(8), &sigset);
-    if (ret < 0) {
-      if (errno == EIO) {
-        FATAL() << "PTRACE_SETSIGMASK not supported; rr requires Linux kernel >= 3.11";
-      }
-      ASSERT(this, errno == EINVAL);
-    } else {
-      LOG(debug) << "Set signal mask to block all signals (bar "
-                 << "SYSCALLBUF_DESCHED_SIGNAL/TIME_SLICE_SIGNAL) while we "
-                 << " have a stashed signal";
+    if (!set_sigmask(sigset)) {
+      return false;
     }
   }
 
@@ -651,8 +659,14 @@ void RecordTask::will_resume_execution(ResumeRequest, WaitRequest,
       vm()->add_breakpoint(
           syscallbuf_code_layout.syscallbuf_final_exit_instruction,
           BKPT_INTERNAL);
+      auto stub_bp = as->monkeypatcher().get_jump_stub_exit_breakpoint(ip(), this);
+      if (stub_bp) {
+        syscallstub_exit_breakpoint = stub_bp;
+        vm()->add_breakpoint(stub_bp, BKPT_INTERNAL);
+      }
     }
   }
+  return true;
 }
 
 vector<remote_code_ptr> RecordTask::syscallbuf_syscall_entry_breakpoints() {
@@ -682,6 +696,9 @@ void RecordTask::did_wait() {
         syscallbuf_code_layout.syscallbuf_final_exit_instruction,
         BKPT_INTERNAL);
   }
+  if (syscallstub_exit_breakpoint) {
+    vm()->remove_breakpoint(syscallstub_exit_breakpoint, BKPT_INTERNAL);
+  }
 
   if (stashed_signals_blocking_more_signals) {
     // Saved 'blocked_sigs' must still be correct regardless of syscallbuf
@@ -724,7 +741,7 @@ void RecordTask::set_emulated_ptracer(RecordTask* tracer) {
   }
 }
 
-bool RecordTask::emulate_ptrace_stop(WaitStatus status,
+bool RecordTask::emulate_ptrace_stop(WaitStatus status, EmulatedStopType stop_type,
                                      const siginfo_t* siginfo, int si_code) {
   ASSERT(this, emulated_stop_type == NOT_STOPPED);
   if (!emulated_ptracer) {
@@ -744,12 +761,12 @@ bool RecordTask::emulate_ptrace_stop(WaitStatus status,
     }
     save_ptrace_signal_siginfo(si);
   }
-  force_emulate_ptrace_stop(status);
+  force_emulate_ptrace_stop(status, stop_type);
   return true;
 }
 
-void RecordTask::force_emulate_ptrace_stop(WaitStatus status) {
-  emulated_stop_type = status.group_stop() ? GROUP_STOP : SIGNAL_DELIVERY_STOP;
+void RecordTask::force_emulate_ptrace_stop(WaitStatus status, EmulatedStopType stop_type) {
+  emulated_stop_type = stop_type;
   emulated_stop_code = status;
   emulated_stop_pending = true;
   emulated_ptrace_SIGCHLD_pending = true;
@@ -1233,6 +1250,10 @@ sig_set_t RecordTask::read_sigmask_from_process() {
   }
 
   auto results = read_proc_status_fields(tid, "SigBlk");
+  if (results.empty()) {
+    // Read failed, process probably died
+    return 0;
+  }
   ASSERT(this, results.size() == 1);
   return strtoull(results[0].c_str(), NULL, 16);
 }
@@ -1247,21 +1268,34 @@ sig_set_t RecordTask::get_sigmask() {
   return blocked_sigs;
 }
 
-void RecordTask::unblock_signal(int sig) {
+bool RecordTask::unblock_signal(int sig) {
   sig_set_t mask = get_sigmask();
   mask &= ~signal_bit(sig);
+  if (!set_sigmask(mask)) {
+    return false;
+  }
+  invalidate_sigmask();
+  return true;
+}
+
+bool RecordTask::set_sigmask(sig_set_t mask) {
   int ret = fallible_ptrace(PTRACE_SETSIGMASK, remote_ptr<void>(8), &mask);
   if (ret < 0) {
     if (errno == EIO) {
       FATAL() << "PTRACE_SETSIGMASK not supported; rr requires Linux kernel >= 3.11";
     }
+    if (errno == ESRCH) {
+      // Task most likely died while we at the ptrace stop.
+      detected_unexpected_exit = true;
+      return false;
+    }
     ASSERT(this, errno == EINVAL);
   } else {
     LOG(debug) << "Set signal mask to block all signals (bar "
                << "SYSCALLBUF_DESCHED_SIGNAL/TIME_SLICE_SIGNAL) while we "
                << " have a stashed signal";
   }
-  invalidate_sigmask();
+  return true;
 }
 
 void RecordTask::set_sig_handler_default(int sig) {
@@ -1299,6 +1333,10 @@ void RecordTask::verify_signal_states() {
   }
 
   auto results = read_proc_status_fields(tid, "SigBlk", "SigIgn", "SigCgt");
+  if (results.empty()) {
+    // Read failed, process probably died
+    return;
+  }
   ASSERT(this, results.size() == 3);
   sig_set_t blocked = strtoull(results[0].c_str(), NULL, 16);
   sig_set_t ignored = strtoull(results[1].c_str(), NULL, 16);
@@ -1354,13 +1392,14 @@ void RecordTask::stash_sig() {
   }
 
   const siginfo_t& si = get_siginfo();
-  stashed_signals.push_back(StashedSignal(si, is_deterministic_signal(this)));
+  stashed_signals.push_back(StashedSignal(si, is_deterministic_signal(this), ip()));
   // Once we've stashed a signal, stop at the next traced/untraced syscall to
   // check whether we need to process the signal before it runs.
   stashed_signals_blocking_more_signals =
       break_at_syscallbuf_final_instruction =
           break_at_syscallbuf_traced_syscalls =
               break_at_syscallbuf_untraced_syscalls = true;
+  syscallstub_exit_breakpoint = nullptr;
 }
 
 void RecordTask::stash_synthetic_sig(const siginfo_t& si,
@@ -1387,11 +1426,12 @@ void RecordTask::stash_synthetic_sig(const siginfo_t& si,
   }
 
   stashed_signals.insert(stashed_signals.begin(),
-                         StashedSignal(si, deterministic));
+                         StashedSignal(si, deterministic, ip()));
   stashed_signals_blocking_more_signals =
       break_at_syscallbuf_final_instruction =
           break_at_syscallbuf_traced_syscalls =
               break_at_syscallbuf_untraced_syscalls = true;
+  syscallstub_exit_breakpoint = nullptr;
 }
 
 bool RecordTask::has_stashed_sig(int sig) const {
@@ -1403,10 +1443,10 @@ bool RecordTask::has_stashed_sig(int sig) const {
   return false;
 }
 
-const siginfo_t* RecordTask::stashed_sig_not_synthetic_SIGCHLD() const {
+const RecordTask::StashedSignal* RecordTask::stashed_sig_not_synthetic_SIGCHLD() const {
   for (auto it = stashed_signals.begin(); it != stashed_signals.end(); ++it) {
     if (!is_synthetic_SIGCHLD(it->siginfo)) {
-      return &it->siginfo;
+      return &*it;
     }
   }
   return nullptr;
@@ -1426,6 +1466,7 @@ void RecordTask::stashed_signal_processed() {
   break_at_syscallbuf_final_instruction = break_at_syscallbuf_traced_syscalls =
       break_at_syscallbuf_untraced_syscalls =
           stashed_signals_blocking_more_signals = has_stashed_sig();
+  syscallstub_exit_breakpoint = nullptr;
 }
 
 const RecordTask::StashedSignal* RecordTask::peek_stashed_sig_to_deliver()
@@ -1482,7 +1523,7 @@ bool RecordTask::is_syscall_restart() {
 
   {
     const Registers& old_regs = ev().Syscall().regs;
-    if (!(old_regs.arg1() == regs().arg1() &&
+    if (!(old_regs.orig_arg1() == regs().arg1() &&
           old_regs.arg2() == regs().arg2() &&
           old_regs.arg3() == regs().arg3() &&
           old_regs.arg4() == regs().arg4() &&
@@ -1523,12 +1564,19 @@ bool RecordTask::is_in_syscallbuf() {
                           p < syscallbuf_code_layout.get_pc_thunks_end)) {
     // Look at the caller to see if we're in the syscallbuf or not.
     bool ok = true;
-    uint64_t addr = read_ptr(this, regs().sp(), &ok);
+    uint64_t addr;
+    if (arch() == aarch64) {
+      addr = regs().xlr();
+    }
+    else {
+      ASSERT(this, is_x86ish(arch())) << "Unknown architecture";
+      addr = read_ptr(this, regs().sp(), &ok);
+    }
     if (ok) {
       p = addr;
     }
   }
-  return as->monkeypatcher().is_jump_stub_instruction(p) ||
+  return as->monkeypatcher().is_jump_stub_instruction(p, false) ||
          (syscallbuf_code_layout.syscallbuf_code_start <= p &&
           p < syscallbuf_code_layout.syscallbuf_code_end);
 }
@@ -1644,12 +1692,19 @@ void RecordTask::record_remote_writable(remote_ptr<void> addr,
   ASSERT(this, num_bytes >= 0);
 
   remote_ptr<void> p = addr;
+  bool seen_rr_mapping = false;
+  int mapping_count = 0;
   while (p < addr + num_bytes) {
     if (!as->has_mapping(p)) {
       break;
     }
+    ++mapping_count;
     auto m = as->mapping_of(p);
-    if (!(m.map.prot() & PROT_WRITE)) {
+    if (m.flags) {
+      seen_rr_mapping = true;
+    }
+    if (!(m.map.prot() & PROT_WRITE) ||
+        (seen_rr_mapping && mapping_count > 1)) {
       break;
     }
     p = m.map.end();
@@ -1846,16 +1901,50 @@ void RecordTask::record_event(const Event& ev, FlushSyscallbuf flush,
     }
   }
 
-  if (ev.is_syscall_event() && ev.Syscall().state == EXITING_SYSCALL) {
-    ticks_at_last_recorded_syscall_exit = tick_count();
-    ip_at_last_recorded_syscall_exit = registers->ip();
+  if (ev.is_syscall_event()) {
+    auto state = ev.Syscall().state;
+    if (state == EXITING_SYSCALL) {
+      ticks_at_last_recorded_syscall_exit = tick_count();
+      ip_at_last_recorded_syscall_exit = registers->ip();
+      if (ticks_at_last_recorded_syscall_exit == ticks_at_last_syscall_entry &&
+          ip_at_last_recorded_syscall_exit == ip_at_last_syscall_entry) {
+        // We've done processing this syscall so we can forget about the entry now
+        // This makes sure that any restarted syscalls would not be treated
+        // as the same entry.
+        ticks_at_last_syscall_entry = 0;
+        ip_at_last_syscall_entry = nullptr;
+        last_syscall_entry_recorded = false;
+      }
+    } else if (state == ENTERING_SYSCALL || state == ENTERING_SYSCALL_PTRACE) {
+      if (tick_count() == ticks_at_last_syscall_entry &&
+          registers->ip() == ip_at_last_syscall_entry) {
+        // Let the process handler know that we've recorded the entry already
+        last_syscall_entry_recorded = true;
+      }
+    }
+  }
+
+  remote_code_ptr rseq_new_ip = ip();
+  bool invalid_rseq_cs = false;
+  if (should_apply_rseq_abort(ev.type(), &rseq_new_ip, &invalid_rseq_cs)) {
+    auto addr = REMOTE_PTR_FIELD(rseq_state->ptr.cast<typename NativeArch::rseq_t>(), rseq_cs);
+    uint64_t value = 0;
+    write_mem(addr, value);
+    record_local(addr, &value);
+  } else {
+    ASSERT(this, !invalid_rseq_cs) << "Invalid rseq_cs found, not currently emulated properly by rr (should segfault)";
   }
 
   trace_writer().write_frame(this, ev, registers, extra_registers);
   LOG(debug) << "Wrote event " << ev << " for time " << current_time;
 
+  if (rseq_new_ip != ip()) {
+    Registers r = regs();
+    r.set_ip(rseq_new_ip);
+    set_regs(r);
+  }
+
   if (!ev.has_ticks_slop() && reset == ALLOW_RESET_SYSCALLBUF) {
-    ASSERT(this, flush == FLUSH_SYSCALLBUF);
     // After we've output an event, it's safe to reset the syscallbuf (if not
     // explicitly delayed) since we will have exited the syscallbuf code that
     // consumed the syscallbuf data.
