@@ -10,6 +10,10 @@
 namespace rr {
 
 struct Sighandlers;
+class TaskSyscallStateBase {
+public:
+  virtual ~TaskSyscallStateBase() {}
+};
 
 /** Different kinds of waits a task can do.
  */
@@ -32,6 +36,9 @@ enum EmulatedStopType {
   NOT_STOPPED,
   GROUP_STOP,          // stopped by a signal. This applies to non-ptracees too.
   SIGNAL_DELIVERY_STOP,// Stopped before delivering a signal. ptracees only.
+  SYSCALL_ENTRY_STOP,  // Stopped at syscall entry. ptracees only
+  SYSCALL_EXIT_STOP,   // Stopped at syscall exit. ptracees only
+  SECCOMP_STOP,        // Stopped at seccomp stop. ptracees only
   CHILD_STOP           // All other kinds of non-ptrace stops
 };
 
@@ -46,6 +53,7 @@ struct SyscallbufCodeLayout {
   remote_code_ptr syscallbuf_code_end;
   remote_code_ptr get_pc_thunks_start;
   remote_code_ptr get_pc_thunks_end;
+  remote_code_ptr syscallbuf_syscall_hook;
   remote_code_ptr syscallbuf_final_exit_instruction;
 };
 
@@ -69,13 +77,14 @@ public:
   virtual void post_wait_clone(Task* cloned_from, int flags) override;
   virtual void on_syscall_exit(int syscallno, SupportedArch arch,
                                const Registers& regs) override;
-  virtual void will_resume_execution(ResumeRequest, WaitRequest, TicksRequest,
+  virtual bool will_resume_execution(ResumeRequest, WaitRequest, TicksRequest,
                                      int /*sig*/) override;
   virtual void did_wait() override;
 
   std::vector<remote_code_ptr> syscallbuf_syscall_entry_breakpoints();
   bool is_at_syscallbuf_syscall_entry_breakpoint();
   bool is_at_syscallbuf_final_instruction_breakpoint();
+  bool is_at_syscallstub_exit_breakpoint();
 
   /**
    * Initialize tracee buffers in this, i.e., implement
@@ -107,11 +116,17 @@ public:
    * Returns true if the task is stopped-for-emulated-ptrace, false otherwise.
    */
   bool emulate_ptrace_stop(WaitStatus status,
+                           const siginfo_t* siginfo = nullptr, int si_code = 0) {
+    return emulate_ptrace_stop(status, status.group_stop() ? GROUP_STOP : SIGNAL_DELIVERY_STOP,
+      siginfo, si_code);
+  }
+  bool emulate_ptrace_stop(WaitStatus status, EmulatedStopType stop_type,
                            const siginfo_t* siginfo = nullptr, int si_code = 0);
+
   /**
    * Force the ptrace-stop state no matter what state the task is currently in.
    */
-  void force_emulate_ptrace_stop(WaitStatus status);
+  void force_emulate_ptrace_stop(WaitStatus status, EmulatedStopType stop_type);
   /**
    * If necessary, signal the ptracer that this task has exited.
    */
@@ -124,7 +139,7 @@ public:
     KERNEL_WRITES_CHILD_TID,
     WRITE_CHILD_TID,
   };
-  void record_exit_event(int exitsig = 0, WriteChildTid write_child_tid = KERNEL_WRITES_CHILD_TID);
+  void record_exit_event(WriteChildTid write_child_tid = KERNEL_WRITES_CHILD_TID);
   /**
    * Called when we're about to deliver a signal to this task. If it's a
    * synthetic SIGCHLD and there's a ptraced task that needs to SIGCHLD,
@@ -273,14 +288,16 @@ public:
   void stash_synthetic_sig(const siginfo_t& si,
                            SignalDeterministic deterministic);
   bool has_stashed_sig() const { return !stashed_signals.empty(); }
-  const siginfo_t* stashed_sig_not_synthetic_SIGCHLD() const;
-  bool has_stashed_sig(int sig) const;
   struct StashedSignal {
-    StashedSignal(const siginfo_t& siginfo, SignalDeterministic deterministic)
-        : siginfo(siginfo), deterministic(deterministic) {}
+    StashedSignal(const siginfo_t& siginfo, SignalDeterministic deterministic,
+                  remote_code_ptr ip)
+        : siginfo(siginfo), deterministic(deterministic), ip(ip) {}
     siginfo_t siginfo;
     SignalDeterministic deterministic;
+    remote_code_ptr ip;
   };
+  const StashedSignal* stashed_sig_not_synthetic_SIGCHLD() const;
+  bool has_stashed_sig(int sig) const;
   const StashedSignal* peek_stashed_sig_to_deliver() const;
   void pop_stash_sig(const StashedSignal* stashed);
   void stashed_signal_processed();
@@ -341,6 +358,10 @@ public:
   /**
    * Return true if this is within the syscallbuf library.  This
    * *does not* imply that $ip is at a buffered syscall.
+   * This also includes the runtime stub code that runs
+   * before entering syscallbuf but does not include the "safe area".
+   * Returning true from this function implies that the code will execute
+   * `_syscallbuf_final_exit_instruction` before returning to normal code.
    */
   bool is_in_syscallbuf();
   /**
@@ -388,7 +409,8 @@ public:
   ssize_t record_remote_fallible(remote_ptr<void> addr, uintptr_t num_bytes,
                                  const std::vector<WriteHole>& holes = std::vector<WriteHole>());
   // Record as much as we can of the bytes in this range. Will record only
-  // contiguous mapped-writable data starting at `addr`.
+  // contiguous mapped-writable data starting at `addr`. rr mappings (e.g. syscallbuf)
+  // are treated as non-contiguous with any other mapping.
   void record_remote_writable(remote_ptr<void> addr, ssize_t num_bytes);
 
   // Simple helper that attempts to use the local mapping to record if one
@@ -542,8 +564,9 @@ public:
   sig_set_t read_sigmask_from_process();
   /**
    * Unblock the signal for the process.
+   * Returns `false` if the process died underneath us.
    */
-  void unblock_signal(int sig);
+  bool unblock_signal(int sig);
   /**
    * Set the signal handler to default for the process.
    */
@@ -594,6 +617,8 @@ public:
    */
   void send_synthetic_SIGCHLD_if_necessary();
 
+  bool set_sigmask(sig_set_t mask);
+
 private:
   /* Retrieve the tid of this task from the tracee and store it */
   void update_own_namespace_tid();
@@ -633,6 +658,7 @@ private:
   virtual bool post_vm_clone(CloneReason reason, int flags, Task* origin) override;
 
 public:
+  std::unique_ptr<TaskSyscallStateBase> syscall_state;
   Ticks ticks_at_last_recorded_syscall_exit;
   remote_code_ptr ip_at_last_recorded_syscall_exit;
 
@@ -761,6 +787,7 @@ public:
   bool break_at_syscallbuf_traced_syscalls;
   bool break_at_syscallbuf_untraced_syscalls;
   bool break_at_syscallbuf_final_instruction;
+  remote_code_ptr syscallstub_exit_breakpoint;
 
   // The pmc is programmed to interrupt at a value requested by the tracee, not
   // by rr.
@@ -784,9 +811,16 @@ public:
   // We've sent a SIGKILL during shutdown for this task.
   bool sent_shutdown_kill;
 
+  // Last exec system call was an execveat
+  bool did_execveat;
+
   // Set if the tracee requested an override of the ticks request.
   // Used for testing.
   TicksRequest tick_request_override;
+
+  // Set to prevent the scheduler from scheduling this tid, even
+  // if it is otherwise considered runnable. Used for testing.
+  bool schedule_frozen;
 };
 
 } // namespace rr
