@@ -360,6 +360,11 @@ public:
   bool set_cx(uintptr_t value) { return RR_SET_REG_X86(ecx, rcx, value); }
 
   uintptr_t ax() const { return RR_GET_REG_X86(eax, rax); }
+  bool set_ax(uintptr_t value) { return RR_SET_REG_X86(eax, rax, value); }
+
+  uintptr_t dx() const { return RR_GET_REG_X86(edx, rdx); }
+  bool set_dx(uintptr_t value) { return RR_SET_REG_X86(edx, rdx, value); }
+
   uintptr_t bp() const { return RR_GET_REG_X86(ebp, rbp); }
 
   uintptr_t flags() const { return RR_GET_REG_X86(eflags, eflags); };
@@ -409,6 +414,16 @@ public:
     u.arm64regs.x[7] = x7;
   }
 
+  void set_x15(uintptr_t x15) {
+    DEBUG_ASSERT(arch() == aarch64);
+    u.arm64regs.x[15] = x15;
+  }
+
+  void set_xlr(uintptr_t xlr) {
+    DEBUG_ASSERT(arch() == aarch64);
+    u.arm64regs.x[30] = xlr;
+  }
+
   uintptr_t x1() const {
     DEBUG_ASSERT(arch() == aarch64);
     return u.arm64regs.x[1];
@@ -418,6 +433,11 @@ public:
     DEBUG_ASSERT(arch() == aarch64);
     return u.arm64regs.x[7];
   }
+
+  uintptr_t xlr() const {
+    DEBUG_ASSERT(arch() == aarch64);
+    return u.arm64regs.x[30];
+  }
   // End of aarch64 specific accessors
 
   /**
@@ -491,7 +511,7 @@ public:
    * Update the register named |reg_name| to |value| with
    * |value_size| number of bytes.
    */
-  void write_register(GdbRegister reg_name, const void* value,
+  bool write_register(GdbRegister reg_name, const void* value,
                       size_t value_size);
 
   /**
@@ -558,7 +578,7 @@ private:
                                            bool* defined) const;
 
   template <typename Arch>
-  void write_register_arch(GdbRegister regno, const void* value,
+  bool write_register_arch(GdbRegister regno, const void* value,
                            size_t value_size);
 
   template <typename Arch>
