@@ -6,6 +6,7 @@
 
 #include <linux/futex.h>
 #include <syscall.h>
+#include <sys/prctl.h>
 
 #include <algorithm>
 
@@ -161,11 +162,12 @@ ReplaySession::ReplaySession(const std::string& dir, const Flags& flags)
       current_step(),
       ticks_at_start_of_event(0),
       flags_(flags),
+      skip_next_execution_event(false),
       trace_start_time(0) {
   if (trace_in.required_forward_compatibility_version() > FORWARD_COMPATIBILITY_VERSION) {
     CLEAN_FATAL()
-      << "This rr build is too old to replay the trace (we support forward compability version "
-      << FORWARD_COMPATIBILITY_VERSION << " but the trace needs " << trace_in.required_forward_compatibility_version();
+      << "This rr build is too old to replay the trace (we support forward compatibility version "
+      << FORWARD_COMPATIBILITY_VERSION << " but the trace needs " << trace_in.required_forward_compatibility_version() << ")";
   }
 
   ticks_semantics_ = trace_in.ticks_semantics();
@@ -183,25 +185,27 @@ ReplaySession::ReplaySession(const std::string& dir, const Flags& flags)
 
   trace_start_time = trace_frame.monotonic_time();
 
-  if (!PerfCounters::supports_ticks_semantics(ticks_semantics_)) {
-    CLEAN_FATAL()
-        << "Trace was recorded on a machine that defines ticks differently\n"
-           "to this machine; replay will not work.";
-  }
-
-  if (is_x86ish(trace_in.arch())) {
-    if (trace_in.uses_cpuid_faulting() && !has_cpuid_faulting()) {
+  if (!flags.replay_stops_at_first_execve) {
+    if (!PerfCounters::supports_ticks_semantics(ticks_semantics_)) {
       CLEAN_FATAL()
-          << "Trace was recorded with CPUID faulting enabled, but this\n"
-            "system does not support CPUID faulting.";
-    }
-    if (!has_cpuid_faulting() && !cpuid_compatible(trace_in.cpuid_records())) {
-      CLEAN_FATAL()
-          << "Trace was recorded on a machine with different CPUID values\n"
-            "and CPUID faulting is not enabled; replay will not work.";
+          << "Trace was recorded on a machine that defines ticks differently\n"
+             "to this machine; replay will not work.";
     }
 
-    check_xsave_compatibility(trace_in);
+    if (is_x86ish(trace_in.arch())) {
+      if (trace_in.uses_cpuid_faulting() && !has_cpuid_faulting()) {
+        CLEAN_FATAL()
+            << "Trace was recorded with CPUID faulting enabled, but this\n"
+              "system does not support CPUID faulting.";
+      }
+      if (!has_cpuid_faulting() && !cpuid_compatible(trace_in.cpuid_records())) {
+        CLEAN_FATAL()
+            << "Trace was recorded on a machine with different CPUID values\n"
+              "and CPUID faulting is not enabled; replay will not work.";
+      }
+
+      check_xsave_compatibility(trace_in);
+    }
   }
 }
 
@@ -217,6 +221,7 @@ ReplaySession::ReplaySession(const ReplaySession& other)
       last_siginfo_(other.last_siginfo_),
       flags_(other.flags_),
       fast_forward_status(other.fast_forward_status),
+      skip_next_execution_event(other.skip_next_execution_event),
       trace_start_time(other.trace_start_time) {}
 
 ReplaySession::~ReplaySession() {
@@ -248,8 +253,7 @@ ReplaySession::shr_ptr ReplaySession::clone() {
  * Return true if it's possible/meaningful to make a checkpoint at the
  * |frame| that |t| will replay.
  */
-static bool can_checkpoint_at(const TraceFrame& frame) {
-  const Event& ev = frame.event();
+static bool can_checkpoint_at(const Event& ev) {
   if (ev.has_ticks_slop()) {
     return false;
   }
@@ -273,7 +277,7 @@ bool ReplaySession::can_clone() {
   finish_initializing();
 
   ReplayTask* t = current_task();
-  return t && done_initial_exec() && can_checkpoint_at(current_trace_frame());
+  return t && done_initial_exec() && can_checkpoint_at(current_trace_frame().event());
 }
 
 DiversionSession::shr_ptr ReplaySession::clone_diversion() {
@@ -283,7 +287,7 @@ DiversionSession::shr_ptr ReplaySession::clone_diversion() {
   LOG(debug) << "Deepforking ReplaySession " << this
              << " to DiversionSession...";
 
-  DiversionSession::shr_ptr session(new DiversionSession());
+  DiversionSession::shr_ptr session(new DiversionSession(cpu_binding()));
   session->ticks_semantics_ = ticks_semantics_;
   session->tracee_socket = tracee_socket;
   session->tracee_socket_fd_number = tracee_socket_fd_number;
@@ -312,7 +316,7 @@ Task* ReplaySession::new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
   vector<string> argv;
   vector<string> env;
 
-  session->do_bind_cpu(session->trace_in);
+  session->do_bind_cpu();
   ScopedFd error_fd = session->create_spawn_task_error_pipe();
   ReplayTask* t = static_cast<ReplayTask*>(
       Task::spawn(*session, error_fd, &session->tracee_socket_fd(),
@@ -325,11 +329,11 @@ Task* ReplaySession::new_task(pid_t tid, pid_t rec_tid, uint32_t serial,
   return session;
 }
 
-int ReplaySession::cpu_binding(TraceStream& trace) const {
+int ReplaySession::cpu_binding() const {
   if (flags_.cpu_unbound) {
     return -1;
   }
-  return Session::cpu_binding(trace);
+  return Session::cpu_binding();
 }
 
 void ReplaySession::advance_to_next_trace_frame() {
@@ -394,7 +398,7 @@ static bool compute_ticks_request(
   *ticks_request = RESUME_UNLIMITED_TICKS;
   if (constraints.ticks_target > 0) {
     Ticks ticks_period =
-        constraints.ticks_target - PerfCounters::skid_size() - t->tick_count();
+        constraints.ticks_target - t->hpc.skid_size() - t->tick_count();
     if (ticks_period <= 0) {
       // Behave as if we actually executed something. Callers assume we did.
       t->clear_wait_status();
@@ -505,7 +509,7 @@ Completion ReplaySession::cont_syscall_boundary(
       break;
   }
   if (t->stop_sig()) {
-      ASSERT(t, false) << "Replay got unrecorded signal " << t->get_siginfo();
+    ASSERT(t, false) << "Replay got unrecorded signal " << t->get_siginfo();
   }
 
   if (t->seccomp_bpf_enabled &&
@@ -520,6 +524,8 @@ Completion ReplaySession::cont_syscall_boundary(
     t->resume_execution(RESUME_SYSEMU, RESUME_WAIT, ticks_request);
   }
 
+  t->apply_syscall_entry_regs();
+
   auto type = AddressSpace::rr_page_syscall_from_exit_point(t->arch(), t->ip());
   if (type && type->traced == AddressSpace::UNTRACED &&
       type->enabled == AddressSpace::REPLAY_ONLY) {
@@ -606,9 +612,16 @@ Completion ReplaySession::enter_syscall(ReplayTask* t,
     if (cont_syscall_boundary(t, constraints) == INCOMPLETE) {
       bool reached_target = syscall_bp_vm && SIGTRAP == t->stop_sig() &&
                             t->ip().undo_executed_bkpt(t->arch()) ==
-                                syscall_instruction &&
-                            t->vm()->get_breakpoint_type_at_addr(
-                                syscall_instruction) == BKPT_INTERNAL;
+                                syscall_instruction;
+      if (reached_target) {
+        // Check if we've hit a user break/watchpoint
+        if (t->vm()->get_breakpoint_type_at_addr(syscall_instruction) != BKPT_INTERNAL) {
+          reached_target = false;
+        }
+        else if (t->vm()->is_exec_watchpoint(syscall_instruction)) {
+          reached_target = false;
+        }
+      }
       if (reached_target) {
         emulate_syscall_entry(t, current_trace_frame(), syscall_instruction);
         clear_syscall_bp();
@@ -669,7 +682,9 @@ void ReplaySession::check_pending_sig(ReplayTask* t) {
 }
 
 static bool do_replay_assist(Task* t) {
-  auto type = AddressSpace::rr_page_syscall_from_exit_point(t->arch(), t->ip());
+  auto orig_ip = t->ip();
+  auto exit_ip = orig_ip.advance_past_executed_bkpt(t->arch());
+  auto type = AddressSpace::rr_page_syscall_from_exit_point(t->arch(), exit_ip);
   if (!type || type->enabled != AddressSpace::REPLAY_ASSIST) {
     return false;
   }
@@ -679,6 +694,11 @@ static bool do_replay_assist(Task* t) {
   Registers regs = t->regs();
   regs.set_syscall_result(next_rec.ret);
   t->on_syscall_exit(next_rec.syscallno, t->arch(), regs);
+  if (orig_ip != exit_ip) {
+    auto r = t->regs();
+    r.set_ip(exit_ip);
+    t->set_regs(r);
+  }
   return true;
 }
 
@@ -727,13 +747,15 @@ Completion ReplaySession::continue_or_step(ReplayTask* t,
         // would make it harder to track down bugs. There is a performance hit
         // to stopping for each mprotect, but replaying recordings of replays
         // is not fast anyway.)
+        t->apply_syscall_entry_regs();
         perform_interrupted_syscall(t);
         return INCOMPLETE;
       }
     } else if (t->stop_sig() == SIGTRAP) {
       // Detect replay assist but handle it later in flush_syscallbuf
-      auto type = AddressSpace::rr_page_syscall_from_exit_point(t->arch(), t->ip());
+      auto type = AddressSpace::rr_page_syscall_from_exit_point(t->arch(), t->ip().advance_past_executed_bkpt(t->arch()));
       if (type && type->enabled == AddressSpace::REPLAY_ASSIST) {
+        t->apply_syscall_entry_regs();
         return INCOMPLETE;
       }
     } else if (handle_unrecorded_cpuid_fault(t, constraints)) {
@@ -793,7 +815,8 @@ static void guard_unexpected_signal(ReplayTask* t) {
 static bool is_same_execution_point(ReplayTask* t, const Registers& rec_regs,
                                     Ticks ticks_left,
                                     Registers* mismatched_regs,
-                                    const Registers** mismatched_regs_ptr) {
+                                    const Registers** mismatched_regs_ptr,
+                                    bool in_syscallbuf) {
   MismatchBehavior behavior =
       IS_LOGGING(debug) ? LOG_MISMATCHES : EXPECT_MISMATCHES;
 
@@ -806,8 +829,21 @@ static bool is_same_execution_point(ReplayTask* t, const Registers& rec_regs,
     }
     return false;
   }
-  if (!Registers::compare_register_files(t, "rep", t->regs(), "rec", rec_regs,
-                                         behavior)) {
+  if (in_syscallbuf) {
+    // In the syscallbuf, only check IP. The values of registers may diverge between
+    // recording and replay.
+    // If this is too loose (i.e. we can reach the same IP with no ticks in between),
+    // we should add instructions to the syscallbuf code to explicitly increase the tick
+    // count, e.g. dummy conditional branches.
+    if (t->ip() != rec_regs.ip()) {
+      LOG(debug) << "  not same execution point: expected IP " << rec_regs.ip()
+                 << ", got " << t->ip();
+      *mismatched_regs = t->regs();
+      *mismatched_regs_ptr = mismatched_regs;
+      return false;
+    }
+  } else if (!Registers::compare_register_files(t, "rep", t->regs(), "rec", rec_regs,
+                                                behavior)) {
     LOG(debug) << "  not same execution point: regs differ (@" << rec_regs.ip()
                << ")";
     *mismatched_regs = t->regs();
@@ -826,12 +862,16 @@ static bool is_same_execution_point(ReplayTask* t, const Registers& rec_regs,
  * from traps related to replaying execution.  |ticks| is an inout param
  * that will be decremented by branches retired during this attempted
  * step.
+ * If in_syscallbuf_syscall_hook is non-null we'll stop if execution
+ * reaches that address and return COMPLETE.
  */
 Completion ReplaySession::emulate_async_signal(
-    ReplayTask* t, const StepConstraints& constraints, Ticks ticks) {
+    ReplayTask* t, const StepConstraints& constraints, Ticks ticks,
+    remote_code_ptr in_syscallbuf_syscall_hook) {
+  bool in_syscallbuf = !in_syscallbuf_syscall_hook.is_null();
+
   const Registers& regs = trace_frame.regs();
   remote_code_ptr ip = regs.ip();
-  bool did_set_internal_breakpoint = false;
 
   /* Step 1: advance to the target ticks (minus a slack region) as
    * quickly as possible by programming the hpc. */
@@ -840,22 +880,35 @@ Completion ReplaySession::emulate_async_signal(
   LOG(debug) << "advancing " << ticks_left << " ticks to reach " << ticks << "/"
              << ip;
 
+  auto skid_size = t->hpc.skid_size();
   /* XXX should we only do this if (ticks > 10000)? */
-  while (ticks_left - PerfCounters::skid_size() > PerfCounters::skid_size()) {
+  while (ticks_left - skid_size > skid_size) {
     LOG(debug) << "  programming interrupt for "
-               << (ticks_left - PerfCounters::skid_size()) << " ticks";
+               << (ticks_left - skid_size) << " ticks";
 
     // Avoid overflow. If ticks_left > MAX_TICKS_REQUEST, execution will stop
     // early but we'll treat that just like a stray TIME_SLICE_SIGNAL and
     // continue as needed.
+    if (in_syscallbuf_syscall_hook) {
+      // Advance no further than syscall_hook.
+      t->vm()->add_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+    }
     continue_or_step(t, constraints,
                      (TicksRequest)(min<Ticks>(MAX_TICKS_REQUEST, ticks_left) -
-                                    PerfCounters::skid_size()));
+                                    skid_size));
     guard_unexpected_signal(t);
+    if (in_syscallbuf_syscall_hook) {
+      t->vm()->remove_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+    }
 
     ticks_left = ticks - t->tick_count();
 
     if (SIGTRAP == t->stop_sig()) {
+      if (t->ip().undo_executed_bkpt(t->arch()) == in_syscallbuf_syscall_hook) {
+        t->move_ip_before_breakpoint();
+        // Advance no further.
+        return COMPLETE;
+      }
       /* We proved we're not at the execution
        * target, and we haven't set any internal
        * breakpoints, and we're not temporarily
@@ -871,6 +924,7 @@ Completion ReplaySession::emulate_async_signal(
   /* True when our advancing has triggered a tracee SIGTRAP that needs to
    * be dealt with. */
   bool pending_SIGTRAP = false;
+  bool did_set_internal_breakpoints = false;
   RunCommand SIGTRAP_run_command = RUN_CONTINUE;
 
   /* Step 2: more slowly, find our way to the target ticks and
@@ -903,7 +957,7 @@ Completion ReplaySession::emulate_async_signal(
      * Determining whether we're at a debugger trap is
      * surprisingly complicated. */
     bool at_target = is_same_execution_point(
-        t, regs, ticks_left, &mismatched_regs, &mismatched_regs_ptr);
+        t, regs, ticks_left, &mismatched_regs, &mismatched_regs_ptr, in_syscallbuf);
     if (pending_SIGTRAP) {
       TrapReasons trap_reasons = t->compute_trap_reasons();
       BreakpointType breakpoint_type =
@@ -917,8 +971,11 @@ Completion ReplaySession::emulate_async_signal(
           (trap_reasons.breakpoint && BKPT_USER == breakpoint_type)) {
         /* Case (0) above: interrupt for the debugger. */
         LOG(debug) << "    trap was debugger singlestep/breakpoint";
-        if (did_set_internal_breakpoint) {
+        if (did_set_internal_breakpoints) {
           t->vm()->remove_breakpoint(ip, BKPT_INTERNAL);
+          if (in_syscallbuf_syscall_hook) {
+            t->vm()->remove_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+          }
         }
         return INCOMPLETE;
       }
@@ -928,11 +985,17 @@ Completion ReplaySession::emulate_async_signal(
         // breakpoint instruction in the tracee would have triggered a
         // deterministic signal instead of an async one.
         // So we must have hit our internal breakpoint.
-        ASSERT(t, did_set_internal_breakpoint);
-        ASSERT(t, regs.ip() == t->ip().undo_executed_bkpt(t->arch()));
+        ASSERT(t, did_set_internal_breakpoints);
         // We didn't do an internal singlestep, and if we'd done a
         // user-requested singlestep we would have hit the above case.
         ASSERT(t, !trap_reasons.singlestep);
+        if (t->ip().undo_executed_bkpt(t->arch()) == in_syscallbuf_syscall_hook) {
+          t->vm()->remove_breakpoint(ip, BKPT_INTERNAL);
+          t->vm()->remove_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+          t->move_ip_before_breakpoint();
+          return COMPLETE;
+        }
+        ASSERT(t, regs.ip() == t->ip().undo_executed_bkpt(t->arch()));
         /* Case (1) above: cover the tracks of
          * our internal breakpoint, and go
          * check again if we're at the
@@ -966,9 +1029,12 @@ Completion ReplaySession::emulate_async_signal(
      * to resume execution in one of a variety of ways,
      * and it's simpler to start out knowing that the
      * breakpoint isn't set. */
-    if (did_set_internal_breakpoint) {
+    if (did_set_internal_breakpoints) {
       t->vm()->remove_breakpoint(ip, BKPT_INTERNAL);
-      did_set_internal_breakpoint = false;
+      if (in_syscallbuf_syscall_hook) {
+        t->vm()->remove_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+      }
+      did_set_internal_breakpoints = false;
     }
 
     if (at_target) {
@@ -978,7 +1044,7 @@ Completion ReplaySession::emulate_async_signal(
 
     /* At this point, we've proven that we're not at the
      * target execution point, and we've ensured the
-     * internal breakpoint is unset. */
+     * internal breakpoints are unset. */
     if (USE_BREAKPOINT_TARGET && regs.ip() != t->regs().ip()) {
       /* Case (4) above: set a breakpoint on the
        * target $ip and PTRACE_CONT in an attempt to
@@ -995,7 +1061,10 @@ Completion ReplaySession::emulate_async_signal(
        * the target execution point. */
       LOG(debug) << "    breaking on target $ip";
       t->vm()->add_breakpoint(ip, BKPT_INTERNAL);
-      did_set_internal_breakpoint = true;
+      if (in_syscallbuf_syscall_hook) {
+        t->vm()->add_breakpoint(in_syscallbuf_syscall_hook, BKPT_INTERNAL);
+      }
+      did_set_internal_breakpoints = true;
       continue_or_step(t, constraints, RESUME_UNLIMITED_TICKS);
       SIGTRAP_run_command = constraints.command;
     } else {
@@ -1039,7 +1108,7 @@ Completion ReplaySession::emulate_async_signal(
      * reach the desired state.
      */
     if (!is_same_execution_point(t, regs, ticks_left, &mismatched_regs,
-                                 &mismatched_regs_ptr)) {
+                                 &mismatched_regs_ptr, in_syscallbuf)) {
       guard_unexpected_signal(t);
     }
 
@@ -1180,7 +1249,7 @@ Completion ReplaySession::emulate_deterministic_signal(
  * Restore the recorded syscallbuf data to the tracee, preparing the
  * tracee for replaying the records.
  */
-void ReplaySession::prepare_syscallbuf_records(ReplayTask* t) {
+void ReplaySession::prepare_syscallbuf_records(ReplayTask* t, Ticks ticks) {
   // Read the recorded syscall buffer back into the buffer
   // region.
   auto buf = t->trace_reader().read_raw_data();
@@ -1201,6 +1270,7 @@ void ReplaySession::prepare_syscallbuf_records(ReplayTask* t) {
              t->syscallbuf_size);
 
   current_step.flush.stop_breakpoint_offset = recorded_hdr.num_rec_bytes / 8;
+  current_step.flush.recorded_ticks = ticks;
 
   LOG(debug) << "Prepared " << (uint32_t)recorded_hdr.num_rec_bytes
              << " bytes of syscall records";
@@ -1339,10 +1409,18 @@ Completion ReplaySession::flush_syscallbuf(ReplayTask* t,
   r.set_ip(r.ip().increment_by_movrm_insn_length(t->arch()));
   t->set_regs(r);
 
+  if (current_step.flush.recorded_ticks <= t->tick_count() && has_trace_quirk(TraceReader::BufferedSyscallForcedTick)) {
+    // We've reached the breakpoint and executed at least as many ticks as were recorded for the FLUSH_SYSCALLBUF.
+    // That means the flush was actually performed before we left the syscallbuf code, i.e. due to a SIGKILL
+    // in syscallbuf code. We will have recorded another execution event that triggered a flush, but just ignore it,
+    // since we probably already passed it and this task is about to die anyway.
+    skip_next_execution_event = true;
+  }
+
   return COMPLETE;
 }
 
-Completion ReplaySession::patch_vsyscall(ReplayTask* t, const StepConstraints& constraints)
+Completion ReplaySession::patch_ip(ReplayTask* t, const StepConstraints& constraints)
 {
   TicksRequest ticks_request;
   if (!compute_ticks_request(t, constraints, &ticks_request)) {
@@ -1375,24 +1453,14 @@ Completion ReplaySession::patch_vsyscall(ReplayTask* t, const StepConstraints& c
     return INCOMPLETE;
   }
 
-  t->apply_all_data_records_from_trace();
+  apply_patch_data(t);
   Registers r = t->regs();
   r.set_ip(vsyscall_entry);
   t->set_regs(r);
   return COMPLETE;
 }
 
-Completion ReplaySession::patch_next_syscall(
-    ReplayTask* t, const StepConstraints& constraints, bool before_syscall) {
-  if (before_syscall) {
-    if (cont_syscall_boundary(t, constraints) == INCOMPLETE) {
-      return INCOMPLETE;
-    }
-
-    t->canonicalize_regs(t->arch());
-    t->exit_syscall_and_prepare_restart();
-  }
-
+void ReplaySession::apply_patch_data(ReplayTask* t) {
   // All patching effects have been recorded to the trace.
   // First, replay any memory mapping done by Monkeypatcher. There should be
   // at most one but we might as well be general.
@@ -1405,8 +1473,8 @@ Completion ReplaySession::patch_next_syscall(
     }
     AutoRemoteSyscalls remote(t);
     ASSERT(t, km.flags() & MAP_ANONYMOUS);
-    remote.infallible_mmap_syscall(km.start(), km.size(), km.prot(),
-                                   km.flags() | MAP_FIXED, -1, 0);
+    remote.infallible_mmap_syscall_if_alive(km.start(), km.size(), km.prot(),
+                                            km.flags() | MAP_FIXED, -1, 0);
     t->vm()->map(t, km.start(), km.size(), km.prot(), km.flags(), 0, string(),
                  KernelMapping::NO_DEVICE, KernelMapping::NO_INODE, nullptr,
                  &km);
@@ -1416,6 +1484,19 @@ Completion ReplaySession::patch_next_syscall(
 
   // Now replay all data records.
   t->apply_all_data_records_from_trace();
+}
+
+Completion ReplaySession::patch_next_syscall(
+    ReplayTask* t, const StepConstraints& constraints, bool before_syscall) {
+  if (before_syscall) {
+    if (cont_syscall_boundary(t, constraints) == INCOMPLETE) {
+      return INCOMPLETE;
+    }
+
+    t->canonicalize_regs(t->arch());
+    t->exit_syscall_and_prepare_restart();
+  }
+  apply_patch_data(t);
   return COMPLETE;
 }
 
@@ -1441,7 +1522,7 @@ void ReplaySession::check_approaching_ticks_target(
     BreakStatus& break_status) {
   if (constraints.ticks_target > 0) {
     Ticks ticks_left = constraints.ticks_target - t->tick_count();
-    if (ticks_left <= PerfCounters::skid_size()) {
+    if (ticks_left <= t->hpc.skid_size()) {
       break_status.approaching_ticks_target = true;
     }
   }
@@ -1469,7 +1550,7 @@ Completion ReplaySession::advance_to_ticks_target(
 Completion ReplaySession::try_one_trace_step(
     ReplayTask* t, const StepConstraints& constraints) {
   if (constraints.ticks_target > 0 && !trace_frame.event().has_ticks_slop() &&
-      t->current_trace_frame().ticks() > constraints.ticks_target) {
+      current_trace_frame().ticks() > constraints.ticks_target) {
     // Instead of doing this step, just advance to the ticks_target, since
     // that happens before this event completes.
     // Unfortunately we can't do this for TSTEP_FLUSH_SYSCALLBUF
@@ -1488,14 +1569,27 @@ Completion ReplaySession::try_one_trace_step(
     case TSTEP_DETERMINISTIC_SIGNAL:
       return emulate_deterministic_signal(t, current_step.target.signo,
                                           constraints);
-    case TSTEP_PROGRAM_ASYNC_SIGNAL_INTERRUPT:
-      return emulate_async_signal(t, constraints, current_step.target.ticks);
+    case TSTEP_PROGRAM_ASYNC_SIGNAL_INTERRUPT: {
+      Completion completion = emulate_async_signal(t, constraints, current_step.target.ticks, current_step.target.in_syscallbuf_syscall_hook);
+      if (completion == COMPLETE) {
+        remote_code_ptr rseq_new_ip = t->ip();
+        bool invalid_rseq_cs;
+        if (t->should_apply_rseq_abort(current_trace_frame().event().type(), &rseq_new_ip, &invalid_rseq_cs)
+            && t->ip() != rseq_new_ip) {
+          Registers r = t->regs();
+          r.set_ip(rseq_new_ip);
+          t->set_regs(r);
+        }
+        t->apply_all_data_records_from_trace();
+      }
+      return completion;
+    }
     case TSTEP_DELIVER_SIGNAL:
       return emulate_signal_delivery(t, current_step.target.signo);
     case TSTEP_FLUSH_SYSCALLBUF:
       return flush_syscallbuf(t, constraints);
-    case TSTEP_PATCH_VSYSCALL:
-      return patch_vsyscall(t, constraints);
+    case TSTEP_PATCH_IP:
+      return patch_ip(t, constraints);
     case TSTEP_PATCH_SYSCALL:
       return patch_next_syscall(t, constraints, true);
     case TSTEP_PATCH_AFTER_SYSCALL:
@@ -1554,8 +1648,7 @@ Completion ReplaySession::exit_task(ReplayTask* t) {
 
 ReplayTask* ReplaySession::revive_task_for_exec() {
   const Event& ev = trace_frame.event();
-  if (!ev.is_syscall_event() ||
-      !is_execve_syscall(ev.Syscall().number, ev.Syscall().arch())) {
+  if (!ev.is_syscall_event() || !ev.Syscall().is_exec()) {
     FATAL() << "Can't find task, but we're not in an execve";
   }
 
@@ -1631,7 +1724,7 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
       break;
     case EV_SYSCALLBUF_FLUSH:
       current_step.action = TSTEP_FLUSH_SYSCALLBUF;
-      prepare_syscallbuf_records(t);
+      prepare_syscallbuf_records(t, trace_frame.ticks());
       break;
     case EV_SYSCALLBUF_RESET:
       // Reset syscallbuf_hdr->num_rec_bytes and zero out the recorded data.
@@ -1644,8 +1737,9 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
       current_step.action = TSTEP_RETIRE;
       break;
     case EV_PATCH_SYSCALL:
-      if (ev.PatchSyscall().patch_vsyscall) {
-        current_step.action = TSTEP_PATCH_VSYSCALL;
+      if (ev.PatchSyscall().patch_trapping_instruction ||
+          ev.PatchSyscall().patch_vsyscall) {
+        current_step.action = TSTEP_PATCH_IP;
       } else if (ev.PatchSyscall().patch_after_syscall) {
         current_step.action = TSTEP_PATCH_AFTER_SYSCALL;
       } else {
@@ -1653,20 +1747,33 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
       }
       break;
     case EV_SCHED:
+      if (skip_next_execution_event) {
+        current_step.action = TSTEP_NONE;
+        break;
+      }
       current_step.action = TSTEP_PROGRAM_ASYNC_SIGNAL_INTERRUPT;
       current_step.target.ticks = trace_frame.ticks();
       current_step.target.signo = 0;
+      current_step.target.in_syscallbuf_syscall_hook = ev.Sched().in_syscallbuf_syscall_hook.register_value();
+      if (current_step.target.in_syscallbuf_syscall_hook) {
+        t->note_sched_in_syscallbuf_syscall_hook();
+      }
       break;
     case EV_INSTRUCTION_TRAP:
       current_step.action = TSTEP_DETERMINISTIC_SIGNAL;
       current_step.target.ticks = -1;
       current_step.target.signo = SIGSEGV;
+      current_step.target.in_syscallbuf_syscall_hook = 0;
       break;
     case EV_GROW_MAP:
       process_grow_map(t);
       current_step.action = TSTEP_RETIRE;
       break;
     case EV_SIGNAL: {
+      if (skip_next_execution_event) {
+        current_step.action = TSTEP_NONE;
+        break;
+      }
       last_siginfo_ = ev.Signal().siginfo;
       if (treat_signal_event_as_deterministic(ev.Signal())) {
         current_step.action = TSTEP_DETERMINISTIC_SIGNAL;
@@ -1677,14 +1784,20 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
         current_step.target.signo = ev.Signal().siginfo.si_signo;
         current_step.target.ticks = trace_frame.ticks();
       }
+      current_step.target.in_syscallbuf_syscall_hook = 0;
       break;
     }
     case EV_SIGNAL_DELIVERY:
     case EV_SIGNAL_HANDLER:
       current_step.action = TSTEP_DELIVER_SIGNAL;
       current_step.target.signo = ev.Signal().siginfo.si_signo;
+      current_step.target.in_syscallbuf_syscall_hook = 0;
       break;
     case EV_SYSCALL:
+      if (skip_next_execution_event) {
+        current_step.action = TSTEP_NONE;
+        break;
+      }
       if (trace_frame.event().Syscall().state == ENTERING_SYSCALL ||
           trace_frame.event().Syscall().state == ENTERING_SYSCALL_PTRACE) {
         rep_prepare_run_to_syscall(t, &current_step);
@@ -1721,14 +1834,16 @@ ReplayTask* ReplaySession::setup_replay_one_trace_frame(ReplayTask* t) {
       FATAL() << "Unexpected event " << ev;
   }
 
+  skip_next_execution_event = false;
   return t;
 }
 
-bool ReplaySession::next_step_is_successful_syscall_exit(int syscallno) {
+bool ReplaySession::next_step_is_successful_exec_syscall_exit() {
+  const Event& ev = trace_frame.event();
   return current_step.action == TSTEP_NONE &&
-         trace_frame.event().is_syscall_event() &&
-         trace_frame.event().Syscall().number == syscallno &&
-         trace_frame.event().Syscall().state == EXITING_SYSCALL &&
+         ev.is_syscall_event() &&
+         ev.Syscall().is_exec() &&
+         ev.Syscall().state == EXITING_SYSCALL &&
          !trace_frame.regs().syscall_failed();
 }
 
@@ -1899,4 +2014,68 @@ double ReplaySession::get_trace_start_time(){
   return trace_start_time;
 }
 
+void ReplaySession::prepare_to_detach_tasks() {
+  finish_initializing();
+
+  for (auto& entry : task_map) {
+    Task* t = entry.second;
+    t->flush_regs();
+  }
+}
+
+void ReplaySession::forget_tasks() {
+  while (!task_map.empty()) {
+    Task* t = task_map.begin()->second;
+    t->forget();
+    delete t;
+  }
+  while (!vm_map.empty()) {
+    AddressSpace* a = vm_map.begin()->second;
+    delete a;
+  }
+}
+
+void ReplaySession::detach_tasks(pid_t new_ptracer, ScopedFd& new_tracee_socket_receiver) {
+  // First tell Yama to let new_ptracer ptrace the tracees.
+  // Do this before sending SIGSTOP to any tracees because SIGSTOP
+  // might stop threads before we do their PR_SET_PTRACER.
+  // Also push the new control socket into all tracees.
+  for (auto& entry : task_map) {
+    Task* t = entry.second;
+    AutoRemoteSyscalls remote(t);
+    long ret = remote.syscall(syscall_number_for_prctl(t->arch()), PR_SET_PTRACER, new_ptracer);
+    ASSERT(t, ret >= 0 || ret == -EINVAL) << "Failed PR_SET_PTRACER";
+    remote.infallible_send_fd_dup(new_tracee_socket_receiver, tracee_socket_fd_number, 0);
+  }
+  // Now PTRACE_DETACH and stop them all with SIGSTOP.
+  for (auto& entry : task_map) {
+    Task* t = entry.second;
+    t->flush_regs();
+    t->xptrace(PTRACE_DETACH, nullptr, (void*)SIGSTOP);
+  }
+  forget_tasks();
+}
+
+void ReplaySession::reattach_tasks(ScopedFd new_tracee_socket, ScopedFd new_tracee_socket_receiver) {
+  tracee_socket = make_shared<ScopedFd>(move(new_tracee_socket));
+  tracee_socket_receiver = make_shared<ScopedFd>(move(new_tracee_socket_receiver));
+  // Seize all tasks.
+  for (auto& entry : task_map) {
+    Task* t = entry.second;
+    long ret = Task::ptrace_seize(t->tid, *this);
+    ASSERT(t, ret >= 0) << "Failed to PTRACE_SEIZE";
+  }
+  // Get stop events for all tasks
+  for (auto& entry : task_map) {
+    Task* t = entry.second;
+    t->wait();
+    if (SIGSTOP != t->status().group_stop()) {
+      WaitStatus failed_status = t->status();
+      FATAL() << "Unexpected stop " << failed_status << " for " << t->tid;
+    }
+    t->clear_wait_status();
+    t->open_mem_fd();
+  }
+}
+
 } // namespace rr
