@@ -29,6 +29,8 @@ class ReplayTask;
 struct ReplayFlushBufferedSyscallState {
   /* The offset in the syscallbuf (in 8-byte units) at which we want to stop */
   uintptr_t stop_breakpoint_offset;
+  /* This includes slop */
+  Ticks recorded_ticks;
 };
 
 /**
@@ -63,8 +65,8 @@ enum ReplayTraceStepType {
   /* Replay until we exit the next syscall, then patch it. */
   TSTEP_PATCH_AFTER_SYSCALL,
 
-  /* Replay until we hit the ip recorded in the event, then patch the vsyscall caller. */
-  TSTEP_PATCH_VSYSCALL,
+  /* Replay until we hit the ip recorded in the event, then patch the site. */
+  TSTEP_PATCH_IP,
 
   /* Exit the task */
   TSTEP_EXIT_TASK,
@@ -92,6 +94,8 @@ struct ReplayTraceStep {
     struct {
       Ticks ticks;
       int signo;
+      // Not remote_code_ptr because this has to have plain data
+      uint64_t in_syscallbuf_syscall_hook;
     } target;
 
     ReplayFlushBufferedSyscallState flush;
@@ -219,7 +223,7 @@ public:
    * Returns true if the next step for this session is to exit a syscall with
    * the given number.
    */
-  bool next_step_is_successful_syscall_exit(int syscallno);
+  bool next_step_is_successful_exec_syscall_exit();
 
   /**
    * The current ReplayStepKey.
@@ -236,11 +240,13 @@ public:
     Flags()
       : redirect_stdio(false)
       , share_private_mappings(false)
+      , replay_stops_at_first_execve(false)
       , cpu_unbound(false) {}
-    Flags(const Flags& other) = default;
+    Flags(const Flags&) = default;
     bool redirect_stdio;
     std::string redirect_stdio_file;
     bool share_private_mappings;
+    bool replay_stops_at_first_execve;
     bool cpu_unbound;
   };
 
@@ -252,9 +258,8 @@ public:
 
   struct StepConstraints {
     explicit StepConstraints(RunCommand command)
-        : command(command), stop_at_time(0), ticks_target(0) {}
+        : command(command), ticks_target(0) {}
     RunCommand command;
-    FrameTime stop_at_time;
     Ticks ticks_target;
     // When the RunCommand is RUN_SINGLESTEP_FAST_FORWARD, stop if the next
     // singlestep would enter one of the register states in this list.
@@ -269,16 +274,11 @@ public:
   };
   /**
    * Take a single replay step.
-   * Ensure we stop at event stop_at_time. If this is not specified,
-   * optimizations may cause a replay_step to pass straight through
-   * stop_at_time.
    * Outside of replay_step, no internal breakpoints will be set for any
    * task in this session.
-   * Stop when the current event reaches stop_at_time (i.e. this event has
-   * is the next event to be replayed).
    * If ticks_target is nonzero, stop before the current task's ticks
-   * reaches ticks_target (but not too far before, unless we hit a breakpoint
-   * or stop_at_time). Only useful for RUN_CONTINUE.
+   * reaches ticks_target (but not too far before, unless we hit a breakpoint).
+   * Only useful for RUN_CONTINUE.
    * Always stops on a switch to a new task.
    */
   ReplayResult replay_step(const StepConstraints& constraints);
@@ -314,7 +314,7 @@ public:
 
   virtual TraceStream* trace_stream() override { return &trace_in; }
 
-  virtual int cpu_binding(TraceStream& trace) const override;
+  virtual int cpu_binding() const override;
 
   bool has_trace_quirk(TraceReader::TraceQuirks quirk) { return trace_in.quirks() & quirk; }
 
@@ -322,6 +322,33 @@ public:
     return tracee_output_fd_.get() ? tracee_output_fd_->get() : dflt;
   }
 
+  /**
+   * Get ready to detach these tasks and reattach them in a child process. Call this
+   * before forking the child.
+   */
+  void prepare_to_detach_tasks();
+  /**
+   * This ReplaySession is in a forked child. The real ReplaySession is still running in
+   * the parent, so we don't really own tasks and other shared resources. Forget about
+   * them so we don't try to tear them down when this ReplaySession is destroyed.
+   */
+  void forget_tasks();
+  /**
+   * The shared resources associated with this ReplaySession are being transferred to
+   * the child process `new_ptracer`. Prepare them for transfer (e.g. ptrace-detach the
+   * tracees) and prepare them to be traced by `new_ptracer`, and forget about them.
+   * `new_sock_fd` is the new control fd pushed into all tasks.
+   */
+  void detach_tasks(pid_t new_ptracer, ScopedFd& new_tracee_socket_receiver);
+  /**
+   * The shared resources associated with this ReplaySession are being transferred to
+   * the child process `new_ptracer`. Receive them in the child process by ptrace-attaching
+   * to them etc.
+   * `new_sock_fd` is the control fd that has been assigned to all tasks,
+   * `new_sock_receiver_fd` is its receiver end.
+   */
+  void reattach_tasks(ScopedFd new_tracee_socket, ScopedFd new_tracee_socket_receiver);
+
 private:
   ReplaySession(const std::string& dir, const Flags& flags);
   ReplaySession(const ReplaySession& other);
@@ -350,14 +377,16 @@ private:
                                           const StepConstraints& constraints);
   Completion emulate_async_signal(ReplayTask* t,
                                   const StepConstraints& constraints,
-                                  Ticks ticks);
-  void prepare_syscallbuf_records(ReplayTask* t);
+                                  Ticks ticks,
+                                  remote_code_ptr in_syscallbuf_syscall_hook);
+  void prepare_syscallbuf_records(ReplayTask* t, Ticks ticks);
   Completion flush_syscallbuf(ReplayTask* t,
                               const StepConstraints& constraints);
   Completion patch_next_syscall(ReplayTask* t,
                                 const StepConstraints& constraints,
                                 bool before_syscall);
-  Completion patch_vsyscall(ReplayTask* t, const StepConstraints& constraints);
+  Completion patch_ip(ReplayTask* t, const StepConstraints& constraints);
+  void apply_patch_data(ReplayTask* t);
   void check_approaching_ticks_target(ReplayTask* t,
                                       const StepConstraints& constraints,
                                       BreakStatus& break_status);
@@ -374,6 +403,7 @@ private:
   siginfo_t last_siginfo_;
   Flags flags_;
   FastForwardStatus fast_forward_status;
+  bool skip_next_execution_event;
 
   // The clock_gettime(CLOCK_MONOTONIC) timestamp of the first trace event, used
   // during 'replay' to calculate the elapsed time between the first event and
