@@ -14,7 +14,9 @@ namespace rr {
 
 ReplayTask::ReplayTask(ReplaySession& session, pid_t _tid, pid_t _rec_tid,
                        uint32_t serial, SupportedArch a)
-    : Task(session, _tid, _rec_tid, serial, a) {}
+    : Task(session, _tid, _rec_tid, serial, a),
+      seen_sched_in_syscallbuf_syscall_hook(false)
+{}
 
 ReplaySession& ReplayTask::session() const {
   return *Task::session().as_replay();
@@ -49,7 +51,7 @@ void ReplayTask::init_buffers_arch(remote_ptr<void> map_hint) {
       cloned_file_data_fname = trace_reader().file_data_clone_file_name(tuid());
       ScopedFd clone_file(cloned_file_data_fname.c_str(), O_RDONLY);
       ASSERT(this, clone_file.is_open());
-      remote.infallible_send_fd_dup(clone_file, cloned_file_data_fd_child);
+      remote.infallible_send_fd_dup(clone_file, cloned_file_data_fd_child, O_CLOEXEC);
       fds->add_monitor(this, cloned_file_data_fd_child, new PreserveFileMonitor());
     }
   }
@@ -61,8 +63,8 @@ void ReplayTask::init_buffers(remote_ptr<void> map_hint) {
   RR_ARCH_FUNCTION(init_buffers_arch, arch(), map_hint);
 }
 
-void ReplayTask::post_exec_syscall(const string& replay_exe) {
-  Task::post_exec(replay_exe);
+void ReplayTask::post_exec_syscall(const string& replay_exe, const string& original_replay_exe) {
+  Task::post_exec(replay_exe, original_replay_exe);
 
   // Perform post-exec-syscall tasks now (e.g. opening mem_fd) before we
   // switch registers. This lets us perform AutoRemoteSyscalls using the
@@ -86,6 +88,10 @@ void ReplayTask::validate_regs(uint32_t flags) {
   if (!session().done_initial_exec()) {
     return;
   }
+  if (seen_sched_in_syscallbuf_syscall_hook) {
+    /* Registers may diverge here */
+    return;
+  }
 
   Registers rec_regs = current_trace_frame().regs();
 
@@ -142,6 +148,11 @@ void ReplayTask::set_return_value_from_trace() {
   // seccomp filters) we need to emulate a change to the original_syscallno
   // (to -1 in that case).
   r.set_original_syscallno(current_trace_frame().regs().original_syscallno());
+  if (r.original_syscallno() == session().syscall_number_for_rrcall_rdtsc()) {
+    ASSERT(this, is_x86ish(arch()));
+    // EAX has been set via set_syscall_result above
+    r.set_dx(current_trace_frame().regs().dx());
+  }
   set_regs(r);
 }
 
