@@ -426,7 +426,6 @@ void ReplayTimeline::seek_to_ticks(FrameTime time, Ticks ticks) {
 
   while (current->trace_reader().time() < time) {
     ReplaySession::StepConstraints constraints(RUN_CONTINUE);
-    constraints.stop_at_time = time;
     ReplayResult result = current->replay_step(constraints);
     if (result.status != REPLAY_CONTINUE) {
       FATAL() << "Trace finished before target was reached";
@@ -465,7 +464,6 @@ ReplayResult ReplayTimeline::replay_step_to_mark(
     // we should.
     ReplaySession::StepConstraints constraints =
         strategy.setup_step_constraints();
-    constraints.stop_at_time = mark.ptr->proto.key.trace_time;
     result = current->replay_step(constraints);
     update_strategy_and_fix_watchpoint_quirk(strategy, constraints, result,
                                              before);
@@ -554,7 +552,6 @@ void ReplayTimeline::seek_to_proto_mark(const ProtoMark& pmark) {
   while (!pmark.equal_states(*current)) {
     if (current->trace_reader().time() < pmark.key.trace_time) {
       ReplaySession::StepConstraints constraints(RUN_CONTINUE);
-      constraints.stop_at_time = pmark.key.trace_time;
       current->replay_step(constraints);
     } else {
       ReplayTask* t = current->current_task();
@@ -830,7 +827,6 @@ bool ReplayTimeline::run_forward_to_intermediate_point(const Mark& end,
   FrameTime mid = (now + end.ptr->proto.key.trace_time) / 2;
   if (now < mid && mid < end.ptr->proto.key.trace_time) {
     ReplaySession::StepConstraints constraints(RUN_CONTINUE);
-    constraints.stop_at_time = mid;
     while (current->trace_reader().time() < mid) {
       current->replay_step(constraints);
     }
@@ -842,7 +838,6 @@ bool ReplayTimeline::run_forward_to_intermediate_point(const Mark& end,
   if (current->trace_reader().time() < end.ptr->proto.key.trace_time &&
       end.ptr->ticks_at_event_start < end.ptr->proto.key.ticks) {
     ReplaySession::StepConstraints constraints(RUN_CONTINUE);
-    constraints.stop_at_time = end.ptr->proto.key.trace_time;
     while (current->trace_reader().time() < end.ptr->proto.key.trace_time) {
       current->replay_step(constraints);
     }
@@ -946,12 +941,13 @@ static bool arch_watch_fires_before_instr(SupportedArch arch) {
 }
 
 ReplayResult ReplayTimeline::reverse_continue(
-    const std::function<bool(ReplayTask* t)>& stop_filter,
+    const std::function<bool(ReplayTask* t, const BreakStatus &)>& stop_filter,
     const std::function<bool()>& interrupt_check) {
   Mark end = mark();
   LOG(debug) << "ReplayTimeline::reverse_continue from " << end;
 
-  bool last_stop_is_watch_or_signal = false;
+  bool last_stop_is_watch = false;
+  bool last_stop_is_signal = false;
   ReplayResult final_result;
   TaskUid final_tuid;
   Ticks final_ticks = 0;
@@ -1026,10 +1022,11 @@ ReplayResult ReplayTimeline::reverse_continue(
         avoidable_stop_ip = result.break_status.task()->ip();
         avoidable_stop_ticks = result.break_status.task()->tick_count();
       }
+      before_watchpoint = false;
 
       evaluate_conditions(result);
       if (result.break_status.any_break() &&
-          !stop_filter(to_replay_task(result.break_status))) {
+          !stop_filter(to_replay_task(result.break_status), result.break_status)) {
         result.break_status = BreakStatus();
       }
 
@@ -1045,12 +1042,16 @@ ReplayResult ReplayTimeline::reverse_continue(
         dest = mark();
         if (result.break_status.signal) {
           LOG(debug) << "Found signal break at " << dest;
+          last_stop_is_watch = false;
+          last_stop_is_signal = true;
         } else {
           LOG(debug) << "Found watch break at " << dest << ", addr="
                      << result.break_status.data_watchpoints_hit()[0].addr;
           if (arch_watch_fires_before_instr(current->arch())) {
             before_watchpoint = true;
           }
+          last_stop_is_signal = false;
+          last_stop_is_watch = true;
         }
         final_result = result;
         final_tuid = result.break_status.task()
@@ -1059,7 +1060,6 @@ ReplayResult ReplayTimeline::reverse_continue(
         final_ticks = result.break_status.task()
                           ? result.break_status.task()->tick_count()
                           : 0;
-        last_stop_is_watch_or_signal = true;
       }
       DEBUG_ASSERT(result.status == REPLAY_CONTINUE);
 
@@ -1071,7 +1071,8 @@ ReplayResult ReplayTimeline::reverse_continue(
         final_result.break_status.task_exit = true;
         final_tuid = final_result.break_status.task()->tuid();
         final_ticks = result.break_status.task()->tick_count();
-        last_stop_is_watch_or_signal = false;
+        last_stop_is_watch = false;
+        last_stop_is_signal = false;
       }
 
       if (at_mark(end)) {
@@ -1092,7 +1093,8 @@ ReplayResult ReplayTimeline::reverse_continue(
         final_ticks = result.break_status.task()
                           ? result.break_status.task()->tick_count()
                           : 0;
-        last_stop_is_watch_or_signal = false;
+        last_stop_is_watch = false;
+        last_stop_is_signal = false;
       }
 
       if (interrupt_check()) {
@@ -1130,12 +1132,23 @@ ReplayResult ReplayTimeline::reverse_continue(
     }
   }
 
-  if (last_stop_is_watch_or_signal) {
-    LOG(debug)
+  if (last_stop_is_watch || last_stop_is_signal) {
+    if (last_stop_is_watch && arch_watch_fires_before_instr(current->arch())) {
+      // GDB expect reverse continue to stop before reversing past the watchpoint
+      // (i.e. pc points to end of instruction) on aarch64
+      // but the dest is pointing to the beginning of the instruction
+      // so we need to single step the instruction at dest.
+      LOG(debug) << "Single step past the watchpoint" << dest;
+      seek_to_mark(dest);
+      unapply_breakpoints_and_watchpoints();
+      current->replay_step(RUN_SINGLESTEP);
+    } else {
+      LOG(debug)
         << "Performing final reverse-singlestep to pass over watch/signal";
-    auto stop_filter = [&](ReplayTask* t) { return t->tuid() == final_tuid; };
-    reverse_singlestep(dest, final_tuid, final_ticks, stop_filter,
-                       interrupt_check);
+      auto stop_filter = [&](ReplayTask* t,const BreakStatus &) { return t->tuid() == final_tuid; };
+      reverse_singlestep(dest, final_tuid, final_ticks, stop_filter,
+                         interrupt_check);
+    }
   } else {
     LOG(debug) << "Seeking to final destination " << dest;
     seek_to_mark(dest);
@@ -1161,7 +1174,7 @@ void ReplayTimeline::update_observable_break_status(
 
 ReplayResult ReplayTimeline::reverse_singlestep(
     const Mark& origin, const TaskUid& step_tuid, Ticks step_ticks,
-    const std::function<bool(ReplayTask* t)>& stop_filter,
+    const std::function<bool(ReplayTask* t,const BreakStatus &)>& stop_filter,
     const std::function<bool()>& interrupt_check) {
   LOG(debug) << "ReplayTimeline::reverse_singlestep from " << origin;
 
@@ -1204,7 +1217,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
       bool seen_other_task_break = false;
       while (!at_mark(end)) {
         ReplayTask* t = current->current_task();
-        if (stop_filter(t) && current->done_initial_exec()) {
+        if (stop_filter(t, BreakStatus()) && current->done_initial_exec()) {
           if (t->tuid() == step_tuid) {
             if (t->tick_count() >= ticks_target) {
               // Don't step any further.
@@ -1258,7 +1271,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
       }
       end = start;
     }
-    DEBUG_ASSERT(stop_filter(current->current_task()) || seen_barrier);
+    DEBUG_ASSERT(stop_filter(current->current_task(), BreakStatus()) || seen_barrier);
 
     Mark destination_candidate;
     Mark step_start = set_short_checkpoint();
@@ -1279,7 +1292,7 @@ ReplayResult ReplayTimeline::reverse_singlestep(
     while (true) {
       Mark now;
       ReplayResult result;
-      if (stop_filter(current->current_task())) {
+      if (stop_filter(current->current_task(), BreakStatus())) {
         apply_breakpoints_and_watchpoints();
         if (current->current_task()->tuid() == step_tuid) {
           Mark before_step = mark();
@@ -1288,7 +1301,9 @@ ReplayResult ReplayTimeline::reverse_singlestep(
           constraints.stop_before_states.push_back(&end.ptr->proto.regs);
           result = current->replay_step(constraints);
           update_observable_break_status(now, result);
-          if (result.break_status.hardware_or_software_breakpoint_hit()) {
+          if (result.break_status.hardware_or_software_breakpoint_hit() ||
+              (!result.break_status.data_watchpoints_hit().empty() &&
+               arch_watch_fires_before_instr(current->arch()))) {
             // If we hit a breakpoint while singlestepping, we didn't
             // make any progress.
             unapply_breakpoints_and_watchpoints();
@@ -1325,7 +1340,9 @@ ReplayResult ReplayTimeline::reverse_singlestep(
           if (result.break_status.any_break()) {
             seen_other_task_break = true;
           }
-          if (result.break_status.hardware_or_software_breakpoint_hit()) {
+          if (result.break_status.hardware_or_software_breakpoint_hit() ||
+              (!result.break_status.data_watchpoints_hit().empty() &&
+               arch_watch_fires_before_instr(current->arch()))) {
             unapply_breakpoints_and_watchpoints();
             result = current->replay_step(RUN_SINGLESTEP_FAST_FORWARD);
             update_observable_break_status(now, result);
@@ -1432,8 +1449,7 @@ void ReplayTimeline::evaluate_conditions(ReplayResult& result) {
   }
 }
 
-ReplayResult ReplayTimeline::replay_step_forward(RunCommand command,
-                                                 FrameTime stop_at_time) {
+ReplayResult ReplayTimeline::replay_step_forward(RunCommand command) {
   DEBUG_ASSERT(command != RUN_SINGLESTEP_FAST_FORWARD);
 
   ReplayResult result;
@@ -1441,7 +1457,6 @@ ReplayResult ReplayTimeline::replay_step_forward(RunCommand command,
   ProtoMark before = proto_mark();
   current->set_visible_execution(true);
   ReplaySession::StepConstraints constraints(command);
-  constraints.stop_at_time = stop_at_time;
   result = current->replay_step(constraints);
   current->set_visible_execution(false);
   if (command == RUN_CONTINUE) {
@@ -1471,7 +1486,7 @@ ReplayResult ReplayTimeline::replay_step_forward(RunCommand command,
 
 ReplayResult ReplayTimeline::reverse_singlestep(
     const TaskUid& tuid, Ticks tuid_ticks,
-    const std::function<bool(ReplayTask* t)>& stop_filter,
+    const std::function<bool(ReplayTask* t, const BreakStatus &)>& stop_filter,
     const std::function<bool()>& interrupt_check) {
   return reverse_singlestep(mark(), tuid, tuid_ticks, stop_filter,
                             interrupt_check);
