@@ -7,6 +7,7 @@
 #include <limits>
 
 #include "Command.h"
+#include "ExportImportCheckpoints.h"
 #include "Flags.h"
 #include "GdbServer.h"
 #include "ReplaySession.h"
@@ -32,6 +33,8 @@ public:
 
 protected:
   RerunCommand(const char* name, const char* help) : Command(name, help) {}
+  // command_for_checkpoint is an in and out parameter.
+  int run_internal(CommandForCheckpoint& command_for_checkpoint);
 
   static RerunCommand singleton;
 };
@@ -44,6 +47,13 @@ RerunCommand RerunCommand::singleton(
     "                             address and jump to <ADDR> to fake call\n"
     "  --singlestep=<REGS>        dump <REGS> after each singlestep\n"
     "  --event-regs=<REGS>        dump <REGS> after each event\n"
+    "  --export-checkpoints=<EVENT>,<NUM>,<FILE>\n"
+    "                             Run to start of <EVENT> and then export checkpoints over\n"
+    "                             Unix socket at <FILE>. Exit after <NUM>\n"
+    "                             connections to the socket.\n"
+    "  --import-checkpoint=<FILE> Start the replay by importing a checkpoint from\n"
+    "                             another rr instance exporting checkpoints at\n"
+    "                             <FILE>\n"
     "  -r, --raw                  dump registers in raw format\n"
     "  -s, --trace-start=<EVENT>  start tracing at <EVENT>\n"
     "  -u, --cpu-unbound          allow replay to run on any CPU. Default is\n"
@@ -75,6 +85,7 @@ enum TraceFieldKind {
   TRACE_XMM_REG,           // outputs 128-bit value
   TRACE_YMM_REG,           // outputs 256-bit value
   TRACE_FIP,               // outputs 64-bit value
+  TRACE_FOP,               // outputs 16-bit value
   TRACE_TID,               // outputs 32-bit value
   TRACE_MXCSR,             // outputs 32-bit value
   TRACE_TICKS,             // outputs 64-bit value
@@ -90,12 +101,18 @@ struct RerunFlags {
   remote_code_ptr function;
   vector<TraceField> singlestep_trace;
   vector<TraceField> event_trace;
+  string import_checkpoint_socket;
+  string export_checkpoints_socket;
+  FrameTime export_checkpoints_event;
+  int export_checkpoints_count;
   bool raw;
   bool cpu_unbound;
 
   RerunFlags()
       : trace_start(0),
         trace_end(numeric_limits<decltype(trace_end)>::max()),
+        export_checkpoints_event(0),
+        export_checkpoints_count(0),
         raw(false),
         cpu_unbound(false) {}
 };
@@ -202,7 +219,7 @@ static void print_regs(Task* t, FrameTime event, uint64_t instruction_count,
   }
   union {
     NativeArch::user_regs_struct gp_regs;
-    uintptr_t regs_values[sizeof(struct user_regs_struct) / sizeof(uintptr_t)];
+    uintptr_t regs_values[sizeof(NativeArch::user_regs_struct) / sizeof(uintptr_t)];
   };
   bool got_gp_regs = false;
   bool first = true;
@@ -349,6 +366,12 @@ static void print_regs(Task* t, FrameTime event, uint64_t instruction_count,
         print_value("fip", &value, sizeof(value), flags, out);
         break;
       }
+      case TRACE_FOP: {
+        bool defined;
+        uint16_t value = t->extra_regs().read_fop(&defined);
+        print_value("fop", &value, sizeof(value), flags, out);
+        break;
+      }
       case TRACE_MXCSR: {
         bool defined;
         uint32_t value = t->extra_regs().read_mxcsr(&defined);
@@ -437,6 +460,8 @@ static bool parse_regs(const string& value, vector<TraceField>* out) {
       out->push_back({ TRACE_XINUSE, 0 });
     } else if (reg == "fip") {
       out->push_back({ TRACE_FIP, 0 });
+    } else if (reg == "fop") {
+      out->push_back({ TRACE_FOP, 0 });
     } else if (reg == "mxcsr") {
       out->push_back({ TRACE_MXCSR, 0});
     } else if (reg == "tid") {
@@ -459,6 +484,8 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
   static const OptionSpec options[] = {
     { 1, "singlestep", HAS_PARAMETER },
     { 2, "event-regs", HAS_PARAMETER },
+    { 3, "export-checkpoints", HAS_PARAMETER },
+    { 4, "import-checkpoint", HAS_PARAMETER },
     { 'e', "trace-end", HAS_PARAMETER },
     { 'f', "function", HAS_PARAMETER },
     { 'r', "raw", NO_PARAMETER },
@@ -481,6 +508,14 @@ static bool parse_rerun_arg(vector<string>& args, RerunFlags& flags) {
         return false;
       }
       break;
+    case 3:
+      if (!parse_export_checkpoints(opt.value, flags.export_checkpoints_event, flags.export_checkpoints_count, flags.export_checkpoints_socket)) {
+        return false;
+      }
+      break;
+    case 4:
+      flags.import_checkpoint_socket = opt.value;
+      break;
     case 'e':
       if (!opt.verify_valid_int(1, UINT32_MAX)) {
         return false;
@@ -572,6 +607,11 @@ static void run_diversion_function(ReplaySession& replay, Task* task,
       ASSERT(task, false) << "Unexpected signal "
                           << *result.break_status.signal;
     }
+
+    if (result.status == DiversionSession::DiversionStatus::DIVERSION_EXITED) {
+        LOG(debug) << "DIVERSION_EXITED, breaking out of diversion_step() loop";
+        break;
+    }
   }
 }
 
@@ -583,8 +623,35 @@ static ReplaySession::Flags session_flags(const RerunFlags& flags) {
   return result;
 }
 
-static int rerun(const string& trace_dir, const RerunFlags& flags) {
-  ReplaySession::shr_ptr replay_session = ReplaySession::create(trace_dir, session_flags(flags));
+static int rerun(const string& trace_dir, const RerunFlags& flags, CommandForCheckpoint& command_for_checkpoint) {
+  ScopedFd export_checkpoints_socket;
+  // Construct the listening socket immediately so importers can connect early and block without polling.
+  // If we need to import a checkpoint, we pass the socket in command_for_checkpoint.fds to the checkpoint
+  // exporter's child process.
+  if (flags.export_checkpoints_event) {
+    if (command_for_checkpoint.session) {
+      export_checkpoints_socket = move(command_for_checkpoint.fds.front());
+    } else {
+      export_checkpoints_socket = bind_export_checkpoints_socket(flags.export_checkpoints_count, flags.export_checkpoints_socket);
+    }
+  }
+
+  ReplaySession::shr_ptr replay_session;
+  if (command_for_checkpoint.session) {
+    replay_session = move(command_for_checkpoint.session);
+  } else if (flags.import_checkpoint_socket.empty()) {
+    replay_session = ReplaySession::create(trace_dir, session_flags(flags));
+    // Now that we've spawned the replay, raise our resource limits if
+    // possible.
+    raise_resource_limits();
+  } else {
+    vector<ScopedFd> fds;
+    if (export_checkpoints_socket.is_open()) {
+      fds.push_back(move(export_checkpoints_socket));
+    }
+    return invoke_checkpoint_command(flags.import_checkpoint_socket, command_for_checkpoint.args, move(fds));
+  }
+
   uint64_t instruction_count_within_event = 0;
   bool done_first_step = false;
   bool need_to_singlestep = !flags.singlestep_trace.empty();
@@ -594,10 +661,6 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
     }
   }
 
-  // Now that we've spawned the replay, raise our resource limits if
-  // possible.
-  raise_resource_limits();
-
   while (replay_session->trace_reader().time() < flags.trace_end) {
     RunCommand cmd = RUN_CONTINUE;
 
@@ -673,16 +736,24 @@ static int rerun(const string& trace_dir, const RerunFlags& flags) {
                  flags.event_trace, stdout);
       instruction_count_within_event = 1;
     }
+
+    if (after_time == flags.export_checkpoints_event) {
+      command_for_checkpoint = export_checkpoints(move(replay_session), flags.export_checkpoints_count,
+          export_checkpoints_socket, flags.export_checkpoints_socket);
+      return 0;
+    }
   }
 
   LOG(info) << "Rerun successfully finished";
   return 0;
 }
 
-int RerunCommand::run(vector<string>& args) {
+int RerunCommand::run_internal(CommandForCheckpoint& command_for_checkpoint) {
+  // parse args first
   bool found_dir = false;
   string trace_dir;
   RerunFlags flags;
+  vector<string> args = command_for_checkpoint.args;
 
   while (!args.empty()) {
     if (parse_rerun_arg(args, flags)) {
@@ -711,7 +782,22 @@ int RerunCommand::run(vector<string>& args) {
     }
   }
 
-  return rerun(trace_dir, flags);
+  return rerun(trace_dir, flags, command_for_checkpoint);
+}
+
+int RerunCommand::run(vector<string>& args) {
+  CommandForCheckpoint command_for_checkpoint;
+  command_for_checkpoint.args = move(args);
+  while (true) {
+    ScopedFd exit_notification_fd = move(command_for_checkpoint.exit_notification_fd);
+    int ret = run_internal(command_for_checkpoint);
+    if (!command_for_checkpoint.session) {
+      if (exit_notification_fd.is_open()) {
+        notify_normal_exit(exit_notification_fd);
+      }
+      return ret;
+    }
+  }
 }
 
 } // namespace rr
