@@ -98,6 +98,8 @@ Scheduler::Scheduler(RecordSession& session)
       enable_poll(false),
       last_reschedule_in_high_priority_only_interval(false),
       unlimited_ticks_mode(false) {
+  std::random_device rd;
+  random.seed(rd());
   regenerate_affinity_mask();
 }
 
@@ -144,7 +146,7 @@ void Scheduler::regenerate_affinity_mask() {
       other_cpus.push_back(i);
     }
   }
-  random_shuffle(other_cpus.begin(), other_cpus.end());
+  shuffle(other_cpus.begin(), other_cpus.end(), random);
   CPU_ZERO(&pretend_affinity_mask_);
   CPU_SET(cpu, &pretend_affinity_mask_);
   for (int i = 0; i < pretend_num_cores_ - 1; ++i) {
@@ -207,7 +209,7 @@ bool Scheduler::is_task_runnable(RecordTask* t, bool* by_waitpid) {
                                "returned a task that must run!";
 
   if (t->detached_proxy) {
-    LOG(debug) << "  " << t->tid << " is waiting a detached proxy";
+    LOG(debug) << "  " << t->tid << " is a detached proxy";
     return false;
   }
 
@@ -220,9 +222,14 @@ bool Scheduler::is_task_runnable(RecordTask* t, bool* by_waitpid) {
     return false;
   }
 
+
   LOG(debug) << "Task event is " << t->ev();
   if (!t->may_be_blocked()) {
     LOG(debug) << "  " << t->tid << " isn't blocked";
+    if (t->schedule_frozen) {
+      LOG(debug) << "  " << t->tid << "  but is frozen";
+      return false;
+    }
     return true;
   }
 
@@ -261,7 +268,11 @@ bool Scheduler::is_task_runnable(RecordTask* t, bool* by_waitpid) {
   if (t->waiting_for_ptrace_exit) {
     LOG(debug) << "  " << t->tid << " is waiting to exit; checking status ...";
   } else if (!t->is_running()) {
-    LOG(debug) << "  was already stopped with status " << t->status();
+    LOG(debug) << "  " << t->tid << "  was already stopped with status " << t->status();
+    if (t->schedule_frozen && t->status().ptrace_event() != PTRACE_EVENT_SECCOMP) {
+      LOG(debug) << "   but is frozen";
+      return false;
+    }
     // If we have may_be_blocked, but we aren't running, then somebody noticed
     // this event earlier and already called did_waitpid for us. Just pretend
     // we did that here.
@@ -271,6 +282,10 @@ bool Scheduler::is_task_runnable(RecordTask* t, bool* by_waitpid) {
   } else if (EV_SYSCALL == t->ev().type() &&
       PROCESSING_SYSCALL == t->ev().Syscall().state &&
       treat_syscall_as_nonblocking(t->ev().Syscall().number, t->arch())) {
+    if (t->schedule_frozen) {
+      LOG(debug) << "  " << t->tid << " is frozen in sched_yield";
+      return false;
+    }
     // These syscalls never really block but the kernel may report that
     // the task is not stopped yet if we pass WNOHANG. To make them
     // behave predictably, do a blocking wait.
@@ -288,10 +303,14 @@ bool Scheduler::is_task_runnable(RecordTask* t, bool* by_waitpid) {
   bool did_wait_for_t;
   did_wait_for_t = t->try_wait();
   if (did_wait_for_t) {
+    LOG(debug) << "  ready with status " << t->status();
+    if (t->schedule_frozen && t->status().ptrace_event() != PTRACE_EVENT_SECCOMP) {
+      LOG(debug) << "   but is frozen";
+      return false;
+    }
     *by_waitpid = true;
     ntasks_running--;
     must_run_task = t;
-    LOG(debug) << "  ready with status " << t->status();
     return true;
   }
   LOG(debug) << "  still blocked";
@@ -319,7 +338,7 @@ RecordTask* Scheduler::find_next_runnable_task(RecordTask* t, bool* by_waitpid,
       for (auto it = same_priority_start; it != same_priority_end; ++it) {
         tasks.push_back(it->second);
       }
-      random_shuffle(tasks.begin(), tasks.end());
+      shuffle(tasks.begin(), tasks.end(), random);
       for (RecordTask* next : tasks) {
         if (is_task_runnable(next, by_waitpid)) {
           return next;
@@ -503,15 +522,43 @@ static RecordTask* find_waited_task(RecordSession& session, pid_t tid, WaitStatu
     }
   }
   if (!waited) {
-    LOG(debug) << "    ... but it's dead";
-    return nullptr;
+    // See if this is one of our detached proxies' original tids.
+    waited = session.find_detached_proxy_task(tid);
+    if (!waited) {
+      LOG(debug) << "    ... but it's dead";
+      return nullptr;
+    }
+
+
+    ASSERT(waited, waited->detached_proxy);
+    ASSERT(waited, status.type() == WaitStatus::FATAL_SIGNAL);
+    LOG(debug) << "    ... but it's a detached proxy";
+
+    // We received an unexpected SIGKILL for one of our detached proxies,
+    // probably because the parent or some other task tried to SIGKILL it
+    // using the original pid. Forward the SIGKILL to the detached process
+    // ... unless the Task has already exited and this SIGKILL raced with
+    // us waiting for the parent to reap it.
+    if (!waited->already_exited()) {
+      LOG(debug) << "        ... sending SIGKILL to detached process " << waited->rec_tid;;
+      ::kill(waited->rec_tid, SIGKILL);
+      int raw_status;
+      pid_t npid = ::waitpid(waited->rec_tid, &raw_status, __WALL | WUNTRACED);
+      ASSERT(waited, npid == waited->rec_tid);
+      status = WaitStatus(raw_status);
+      ASSERT(waited, status.type() == WaitStatus::EXIT ||
+                     status.type() == WaitStatus::FATAL_SIGNAL);
+    } else {
+      LOG(debug) << "        ... but the detached process is already dead";
+    }
   }
   if (waited->detached_proxy) {
+    waited->did_waitpid(status);
     pid_t parent_rec_tid = waited->get_parent_pid();
-    LOG(debug) << "    ... but it's a detached proxy.";
+    LOG(debug) << "    ... but it's a detached process.";
     RecordTask *parent = session.find_task(parent_rec_tid);
-    if (parent) {
-      LOG(debug) << "    ... forwarding to parent.";
+    if (parent && !waited->emulated_stop_pending) {
+      LOG(debug) << "    ... notifying parent.";
       waited->emulated_stop_type = CHILD_STOP;
       waited->emulated_stop_pending = true;
       waited->emulated_SIGCHLD_pending = true;
@@ -528,6 +575,7 @@ static RecordTask* find_waited_task(RecordSession& session, pid_t tid, WaitStatu
       } else {
         // The task is now dead, but so is our parent, so none of our
         // tasks care about this. We can now delete the proxy task.
+        // This will also reap the rec_tid of the proxy task.
         delete waited;
       }
     }
@@ -728,9 +776,9 @@ Scheduler::Rescheduled Scheduler::reschedule(Switchable switchable) {
         result.interrupted_by_signal = true;
         return result;
       }
+      LOG(debug) << "  " << tid << " changed status to " << status;
       next = find_waited_task(session, tid, status);
       now = -1; // invalid, don't use
-      LOG(debug) << "  " << tid << " changed status to " << status;
       if (next) {
         ASSERT(next,
                next->may_be_blocked() ||
