@@ -6,6 +6,7 @@
 #include <sched.h>
 
 #include <deque>
+#include <random>
 #include <set>
 
 #include "Ticks.h"
@@ -182,6 +183,7 @@ private:
   // Tasks sorted by priority.
   typedef std::set<std::pair<int, RecordTask*>> TaskPrioritySet;
   typedef std::deque<RecordTask*> TaskQueue;
+  typedef std::default_random_engine Random;
 
   /**
    * Pull a task from the round-robin queue if available. Otherwise,
@@ -214,6 +216,8 @@ private:
   void validate_scheduled_task();
   void regenerate_affinity_mask();
 
+  Random random;
+
   RecordSession& session;
 
   /**
