@@ -15,6 +15,7 @@
 #include "RecordSession.h"
 #include "TraceStream.h"
 #include "core.h"
+#include "cpp_supplement.h"
 #include "log.h"
 #include "main.h"
 #include "util.h"
@@ -106,7 +107,7 @@ static void base_name(string& s) {
   }
 }
 
-static bool is_absolute(string& s) {
+static bool is_absolute(const string& s) {
   return s[0] == '/';
 }
 
@@ -145,19 +146,23 @@ struct DirExistsCache {
 // the build directory is some ancestor directory of original_file_name.
 // We try making comp_dir/rel_dir/file_name relative to each ancestor directory
 // of original_file_name, and if we find a file there, we return that name.
+// original_file_name must be absolute if not NULL.
 //
 // If non-empty, `comp_dir_substution` should replace `original_comp_dir`
 // in `rel_dir` if `original_comp_dir` is a prefix of `rel_dir`.
-static string resolve_file_name(const char* original_file_name,
-                                const char* comp_dir,
-                                const char* original_comp_dir,
-                                const string& comp_dir_substitution,
-                                const char* rel_dir,
-                                const char* file_name,
-                                DirExistsCache& dir_exists_cache) {
-  string path = file_name;
+// Always returns an absolute file name.
+// Returns true if we got a result, otherwise false.
+static bool resolve_file_name(const char* original_file_name,
+                              const char* comp_dir,
+                              const char* original_comp_dir,
+                              const string& comp_dir_substitution,
+                              const char* rel_dir,
+                              const char* file_name,
+                              DirExistsCache& dir_exists_cache,
+                              string& path) {
+  path = file_name;
   if (is_absolute(path)) {
-    return path;
+    return true;
   }
   if (rel_dir) {
     if (rel_dir[0] == '/' && !comp_dir_substitution.empty() && original_comp_dir &&
@@ -168,27 +173,33 @@ static string resolve_file_name(const char* original_file_name,
       prepend_path(rel_dir, path);
     }
     if (is_absolute(path)) {
-      return path;
+      return true;
     }
   }
   if (comp_dir) {
     prepend_path(comp_dir, path);
     if (is_absolute(path)) {
-      return path;
+      return true;
     }
   }
   if (!original_file_name) {
-    return path;
+    if (is_absolute(path)) {
+      return true;
+    }
+    LOG(warn) << "Path " << path << " is relative and we can't make it absolute";
+    return false;
   }
   string original(original_file_name);
   while (true) {
     dir_name(original);
     if (original.empty()) {
-      return path;
+      LOG(warn) << "Path " << path << " is relative and we can't make it absolute";
+      return false;
     }
     string candidate = original + "/" + path;
     if (dir_exists_cache.dir_exists(candidate)) {
-      return candidate;
+      path = candidate;
+      return true;
     }
   }
 }
@@ -210,11 +221,23 @@ static bool process_compilation_units(ElfFileReader& reader,
                                       set<string>* file_names, vector<DwoInfo>* dwos,
                                       DirExistsCache& dir_exists_cache) {
   DwarfSpan debug_info = reader.dwarf_section(".debug_info");
+  if (debug_info.empty()) {
+    debug_info = reader.dwarf_section(".zdebug_info", true);
+  }
   DwarfSpan debug_abbrev = reader.dwarf_section(".debug_abbrev");
+  if (debug_abbrev.empty()) {
+    debug_abbrev = reader.dwarf_section(".zdebug_abbrev", true);
+  }
   DwarfSpan debug_str = reader.dwarf_section(".debug_str");
+  if (debug_str.empty()) {
+    debug_str = reader.dwarf_section(".zdebug_str", true);
+  }
   DwarfSpan debug_str_sup = sup_reader ? sup_reader->dwarf_section(".debug_str") : DwarfSpan();
   DwarfSpan debug_str_offsets = reader.dwarf_section(".debug_str_offsets");
   DwarfSpan debug_line = reader.dwarf_section(".debug_line");
+  if (debug_line.empty()) {
+    debug_line = reader.dwarf_section(".zdebug_line", true);
+  }
   DwarfSpan debug_line_str = reader.dwarf_section(".debug_line_str");
   if (debug_info.empty() || debug_abbrev.empty() ||
       (debug_str.empty() && debug_str_sup.empty()) ||
@@ -275,12 +298,16 @@ static bool process_compilation_units(ElfFileReader& reader,
         }
       }
       if (has_dwo_id) {
-        string full_name = resolve_file_name(original_file_name.c_str(), comp_dir, original_comp_dir, comp_dir_substitution, nullptr, dwo_name, dir_exists_cache);
-        string c;
-        if (comp_dir) {
-          c = comp_dir;
+        string full_name;
+        if (resolve_file_name(original_file_name.c_str(), comp_dir, original_comp_dir, comp_dir_substitution, nullptr, dwo_name, dir_exists_cache, full_name)) {
+          string c;
+          if (comp_dir) {
+            c = comp_dir;
+          }
+          dwos->push_back({ dwo_name, trace_relative_name, move(c), full_name, dwo_id });
+        } else {
+          FATAL() << "DWO missing due to relative path " << full_name;
         }
-        dwos->push_back({ dwo_name, trace_relative_name, move(c), full_name, dwo_id });
       } else {
         LOG(warn) << "DW_AT_GNU_dwo_name but not DW_AT_GNU_dwo_id";
       }
@@ -290,7 +317,10 @@ static bool process_compilation_units(ElfFileReader& reader,
       continue;
     }
     if (source_file_name) {
-      file_names->insert(resolve_file_name(original_file_name.c_str(), comp_dir, original_comp_dir, comp_dir_substitution, nullptr, source_file_name, dir_exists_cache));
+      string full_name;
+      if (resolve_file_name(original_file_name.c_str(), comp_dir, original_comp_dir, comp_dir_substitution, nullptr, source_file_name, dir_exists_cache, full_name)) {
+        file_names->insert(full_name);
+      }
     }
     intptr_t stmt_list = cu.die().section_ptr_attr(DW_AT_stmt_list, &ok);
     if (stmt_list < 0 || !ok) {
@@ -306,7 +336,10 @@ static bool process_compilation_units(ElfFileReader& reader,
         continue;
       }
       const char* dir = lines.directories()[f.directory_index];
-      file_names->insert(resolve_file_name(original_file_name.c_str(), comp_dir, original_comp_dir, comp_dir_substitution, dir, f.file_name, dir_exists_cache));
+      string full_name;
+      if (resolve_file_name(original_file_name.c_str(), comp_dir, original_comp_dir, comp_dir_substitution, dir, f.file_name, dir_exists_cache, full_name)) {
+        file_names->insert(full_name);
+      }
     }
   } while (!debug_info.empty());
 
@@ -396,8 +429,23 @@ find_auxiliary_file(const string& original_file_name,
     }
     LOG(info) << "Can't find external debuginfo file " << full_file_name;
 
+    // On Ubuntu 20.04 there's both a /lib/x86_64-linux-gnu/libc-2.31.so and a
+    // /usr/lib/x86_64-linux-gnu/libc-2.31.so. They are hardlinked to the same inode,
+    // and glibc debuginfo is present in the location corresponding to
+    // /lib/x86_64-linux-gnu/libc-2.31.so. But the kernel returns the /usr prefixed
+    // path from /proc/<pid>/fd/<fd>. Hack around that here.
+    if (original_file_dir.find("/usr/") == 0) {
+      full_file_name = "/usr/lib/debug" + original_file_dir.substr(sizeof("/usr") - 1) + "/" + aux_file_name;
+      normalize_file_name(full_file_name);
+      fd = ScopedFd(full_file_name.c_str(), O_RDONLY);
+      if (fd.is_open()) {
+        goto found;
+      }
+      LOG(info) << "Can't find external debuginfo file " << full_file_name;
+    }
+
     // If none of those worked, give up.
-    LOG(warn) << "Exhausted auxilliary debuginfo search locations for " << aux_file_name;
+    LOG(warn) << "Exhausted auxiliary debuginfo search locations for " << aux_file_name;
     return nullptr;
   }
 
@@ -541,7 +589,14 @@ static bool has_subdir(string& base, const char* suffix) {
   return !ret;
 }
 
+static void assert_absolute(const string& path) {
+  if (!is_absolute(path)) {
+    FATAL() << "Path " << path << " not absolute";
+  }
+}
+
 static void check_vcs_root(string& path, set<string>* vcs_dirs) {
+  assert_absolute(path);
   if (has_subdir(path, "/.git") || has_subdir(path, "/.hg")) {
     vcs_dirs->insert(path + "/");
   }
@@ -550,7 +605,7 @@ static void check_vcs_root(string& path, set<string>* vcs_dirs) {
 // Returns an empty string if the path does not exist or
 // is not accessible.
 // `path` need not be normalized, i.e. may contain .. or .
-// components.
+// components. It mus be absolute.
 // The result string, if non-empty, will be absolute,
 // normalized, and contain no symlink components.
 // The keys in resolved_dirs are absolute file paths
@@ -563,6 +618,7 @@ static string resolve_symlinks(const string& path,
                                unordered_map<string, string>* resolved_dirs,
                                vector<Symlink>* symlinks,
                                set<string>* vcs_dirs) {
+  assert_absolute(path);
   // Absolute, not normalized. We don't keep this normalized because
   // we want resolved_dirs to work well.
   // This is always a prefix of `path`.
@@ -675,6 +731,7 @@ static string resolve_symlinks(const string& path,
 
 /// Adds to vcs_dirs any directory paths under any
 /// of our resolved directories.
+/// file_names must be absolute.
 static void build_symlink_map(const set<string>& file_names,
                               set<string>* resolved_file_names,
                               vector<Symlink>* symlinks,
@@ -705,6 +762,7 @@ struct OutputCompDirSubstitution {
 
 static int sources(const map<string, string>& binary_file_names, const map<string, string>& comp_dir_substitutions, bool is_explicit) {
   vector<string> relevant_binary_names;
+  // Must be absolute.
   set<string> file_names;
   set<ExternalDebugInfo> external_debug_info;
   vector<DwoInfo> dwos;
