@@ -14,7 +14,6 @@
 #include "ExtraRegisters.h"
 #include "FdTable.h"
 #include "PerfCounters.h"
-#include "PropertyTable.h"
 #include "Registers.h"
 #include "TaskishUid.h"
 #include "ThreadGroup.h"
@@ -113,6 +112,13 @@ struct TrapReasons {
   bool breakpoint;
 };
 
+struct RseqState {
+  remote_ptr<void> ptr;
+  uint32_t abort_prefix_signature;
+  RseqState(remote_ptr<void> ptr, uint32_t abort_prefix_signature)
+    : ptr(ptr), abort_prefix_signature(abort_prefix_signature) {}
+};
+
 /**
  * A "task" is a task in the linux usage: the unit of scheduling.  (OS
  * people sometimes call this a "thread control block".)  Multiple
@@ -302,6 +308,10 @@ public:
    * Emulate a jump to a new IP, updating the ticks counter as appropriate.
    */
   void emulate_jump(remote_code_ptr);
+  void count_direct_jump()
+  {
+    ticks += PerfCounters::ticks_for_unconditional_direct_branch(this);
+  }
 
   /**
    * Return true if this is at an arm-desched-event or
@@ -336,7 +346,14 @@ public:
    * instruction.
    */
   bool is_in_untraced_syscall() {
-    auto t = AddressSpace::rr_page_syscall_from_exit_point(arch(), ip());
+    const AddressSpace::SyscallType *t;
+    if (arch() == aarch64 && stop_sig() > 0) {
+      // On aarch64 we can't distinguish untraced syscall entry and exit
+      // when a signal happened
+      t = AddressSpace::rr_page_syscall_from_entry_point(arch(), ip());
+    } else {
+      t = AddressSpace::rr_page_syscall_from_exit_point(arch(), ip());
+    }
     return t && t->traced == AddressSpace::UNTRACED;
   }
 
@@ -367,9 +384,11 @@ public:
 
   /**
    * Hook called by `resume_execution`.
+   * Returns `false` if the task is in the process of dying and setup could not
+   * be completed, `true` otherwise.
    */
-  virtual void will_resume_execution(ResumeRequest, WaitRequest, TicksRequest,
-                                     int /*sig*/) {}
+  virtual bool will_resume_execution(ResumeRequest, WaitRequest, TicksRequest,
+                                     int /*sig*/) { return true; }
   /**
    * Hook called by `did_waitpid`.
    */
@@ -398,8 +417,10 @@ public:
    * We're currently in user-space with registers set up to perform a system
    * call. Continue into the kernel and stop where we can modify the syscall
    * state.
+   * Return `true` if the syscall entry succeeded.
+   * Return `false` if the tracee exited unexpectedly.
    */
-  void enter_syscall();
+  bool enter_syscall(bool allow_exit=false);
 
   /**
    * We have observed entry to a syscall (either by PTRACE_EVENT_SECCOMP or
@@ -420,8 +441,12 @@ public:
    * Call this method when this task has just performed an |execve()|
    * (so we're in the new address space), but before the system call has
    * returned.
+   * `exe_file` is the name of the executable file in the trace, if there is one,
+   * otherwise the original exe file name --- a best-effort filename we can
+   * pass to gdb for it to read the exe. `original_exe_file` is the
+   * original file exe file name.
    */
-  void post_exec(const std::string& exe_file);
+  void post_exec(const std::string& exe_file, const std::string& original_exe_file);
 
   /**
    * Call this method when this task has exited a successful execve() syscall.
@@ -545,6 +570,11 @@ public:
   /** Set the tracee's extra registers to |regs|. */
   void set_extra_regs(const ExtraRegisters& regs);
 
+  /** Adjust IP for rseq abort if necessary and return true if an abort is required.
+   * Sets *rseq_cs_invalid if it was invalid */
+  bool should_apply_rseq_abort(EventType event_type, remote_code_ptr* new_ip,
+                               bool* invalid_rseq_cs);
+
   /**
    * Read the aarch64 TLS register via ptrace. Returns true on success, false
    * on failure. On success `result` is set to the tracee's TLS register.
@@ -779,6 +809,26 @@ public:
    */
   void open_mem_fd_if_needed();
 
+  /**
+   * Perform a PTRACE_INTERRUPT set up the counter for potential spurious stops
+   * to be detected in `account_for_potential_ptrace_interrupt_stop`.
+   */
+  void do_ptrace_interrupt();
+
+  /**
+   * Sometimes we use PTRACE_INTERRUPT to kick the tracee out of various
+   * undesirable states. Unfortunately, that can (but need not) result in later
+   * undesired GROUP-STOP-SIGTRAP stops which report the PTRACE_INTERRUPT.
+   * This function may be called when examining stops to account for any
+   * such spurious stops.
+   *
+   * Should be called at exactly once for every ptrace stop.
+   *
+   * Returns true if the stop is caused by a PTRACE_INTERRUPT we know about,
+   * false otherwise.
+   */
+  bool account_for_potential_ptrace_interrupt_stop(WaitStatus status);
+
   /* Imagine that task A passes buffer |b| to the read()
    * syscall.  Imagine that, after A is switched out for task B,
    * task B then writes to |b|.  Then B is switched out for A.
@@ -832,6 +882,8 @@ public:
   int cloned_file_data_fd_child;
   /* The filename opened by the child's cloned_file_data_fd */
   std::string cloned_file_data_fname;
+  // Current rseq state if registered
+  std::unique_ptr<RseqState> rseq_state;
 
   PerfCounters hpc;
 
@@ -852,8 +904,6 @@ public:
   typedef uint8_t ThreadLocals[PRELOAD_THREAD_LOCALS_SIZE];
   ThreadLocals thread_locals;
 
-  PropertyTable& properties() { return properties_; }
-
   size_t usable_scratch_size() {
     return std::max<ssize_t>(0, scratch_size - page_size());
   }
@@ -863,7 +913,9 @@ public:
   }
   void setup_preload_thread_locals();
   void setup_preload_thread_locals_from_clone(Task* origin);
-  const ThreadLocals& fetch_preload_thread_locals();
+  // If `fetch_full` is false, avoid fetching the full stub_scratch_2 on aarch64
+  // and only fetch the first two pointers from it.
+  const ThreadLocals& fetch_preload_thread_locals(bool fetch_full);
   void activate_preload_thread_locals();
 
   struct CapturedState {
@@ -879,6 +931,7 @@ public:
     remote_ptr<void> scratch_ptr;
     ssize_t scratch_size;
     remote_ptr<void> top_of_stack;
+    std::unique_ptr<RseqState> rseq_state;
     uint64_t cloned_file_data_offset;
     ThreadLocals thread_locals;
     pid_t rec_tid;
@@ -961,6 +1014,11 @@ public:
                      const std::vector<std::string>& argv,
                      const std::vector<std::string>& envp, pid_t rec_tid = -1);
 
+  /**
+   * Do PTRACE_SEIZE on this tid with the correct ptrace options.
+   */
+  static long ptrace_seize(pid_t tid, Session& session);
+
   /**
    * Do a tgkill to send a specific signal to this task.
    */
@@ -976,6 +1034,19 @@ public:
   // FdTable sharing relationships during a session fork.
   using ClonedFdTables = std::unordered_map<uintptr_t, FdTable::shr_ptr>;
 
+  /**
+   * Just forget that this Task exists. Another rr process will manage it.
+   */
+  void forget();
+
+  // Used on aarch64 to detect whether we've recorded x0 and x8 on syscall entry
+  Ticks ticks_at_last_syscall_entry;
+  remote_code_ptr ip_at_last_syscall_entry;
+  // Whether the syscall entry corresponding to `{ticks,ip}_at_last_syscall_entry`
+  // has been recorded in the trace
+  // (used to avoid double recording on unexpected exit)
+  bool last_syscall_entry_recorded;
+
 protected:
   Task(Session& session, pid_t tid, pid_t rec_tid, uint32_t serial,
        SupportedArch a);
@@ -1205,13 +1276,13 @@ protected:
   // If this is true, `seen_ptrace_exit_event` must be true.
   bool handled_ptrace_exit_event;
 
-  PropertyTable properties_;
-
   // A counter for the number of stops for which the stop may have been caused
   // by PTRACE_INTERRUPT. See description in do_waitpid
   int expecting_ptrace_interrupt_stop;
 
   bool was_reaped;
+  // Let this Task object be destroyed with no consequences.
+  bool forgotten;
 
   Task(Task&) = delete;
   Task operator=(Task&) = delete;
