@@ -408,6 +408,7 @@ void TraceWriter::write_frame(RecordTask* t, const Event& ev,
   frame.setTid(t->tid);
   frame.setTicks(t->tick_count());
   frame.setMonotonicSec(monotonic_now_sec());
+  frame.setInSyscallbufSyscallHook(0);
   auto mem_writes = frame.initMemWrites(raw_recs.size());
   for (size_t i = 0; i < raw_recs.size(); ++i) {
     auto w = mem_writes[i];
@@ -439,7 +440,9 @@ void TraceWriter::write_frame(RecordTask* t, const Event& ev,
       event.setInstructionTrap(Void());
       break;
     case EV_PATCH_SYSCALL:
-      if (ev.PatchSyscall().patch_vsyscall) {
+      if (ev.PatchSyscall().patch_trapping_instruction) {
+        event.setPatchTrappingInstruction(Void());
+      } else if (ev.PatchSyscall().patch_vsyscall) {
         event.setPatchVsyscall(Void());
       } else if (ev.PatchSyscall().patch_after_syscall) {
         event.setPatchAfterSyscall(Void());
@@ -454,6 +457,7 @@ void TraceWriter::write_frame(RecordTask* t, const Event& ev,
       event.setSyscallbufReset(Void());
       break;
     case EV_SCHED:
+      frame.setInSyscallbufSyscallHook(ev.Sched().in_syscallbuf_syscall_hook.register_value());
       event.setSched(Void());
       break;
     case EV_GROW_MAP:
@@ -601,6 +605,10 @@ TraceFrame TraceReader::read_frame() {
     case trace::Frame::Event::PATCH_SYSCALL:
       ret.ev = Event::patch_syscall();
       break;
+    case trace::Frame::Event::PATCH_TRAPPING_INSTRUCTION:
+      ret.ev = Event::patch_syscall();
+      ret.ev.PatchSyscall().patch_trapping_instruction = true;
+      break;
     case trace::Frame::Event::PATCH_VSYSCALL:
       ret.ev = Event::patch_syscall();
       ret.ev.PatchSyscall().patch_vsyscall = true;
@@ -617,6 +625,7 @@ TraceFrame TraceReader::read_frame() {
       break;
     case trace::Frame::Event::SCHED:
       ret.ev = Event::sched();
+      ret.ev.Sched().in_syscallbuf_syscall_hook = frame.getInSyscallbufSyscallHook();
       break;
     case trace::Frame::Event::GROW_MAP:
       ret.ev = Event::grow_map();
@@ -733,6 +742,8 @@ void TraceWriter::write_task_event(const TraceTaskEvent& event) {
         cmd_line.set(i, str_to_data(event_cmd_line[i]));
       }
       exec.setExeBase(event.exe_base().as_int());
+      exec.setInterpBase(event.interp_base().as_int());
+      exec.setInterpName(str_to_data(event.interp_name()));;
       break;
     }
     case TraceTaskEvent::EXIT:
@@ -793,6 +804,8 @@ TraceTaskEvent TraceReader::read_task_event(FrameTime* time) {
         r.cmd_line_[i] = data_to_str(cmd_line[i]);
       }
       r.exe_base_ = exec.getExeBase();
+      r.interp_base_ = exec.getInterpBase();
+      r.interp_name_ = data_to_str(exec.getInterpName());
       break;
     }
     case trace::TaskEvent::Which::EXIT:
@@ -1273,8 +1286,9 @@ static string make_trace_dir(const string& exe_path, const string& output_trace_
       return output_trace_dir;
     }
     if (EEXIST == errno) {
-      // directory already exists
-      FATAL() << "Directory `" << output_trace_dir << "' already exists.";
+      CLEAN_FATAL() << "Trace directory `" << output_trace_dir << "' already exists.";
+    } else if (EACCES == errno) {
+      CLEAN_FATAL() << "Permission denied to create trace directory `" << output_trace_dir << "'";
     } else {
       FATAL() << "Unable to create trace directory `" << output_trace_dir << "'";
     }
@@ -1415,6 +1429,8 @@ void TraceWriter::close(CloseStatus status, const TraceUuid* uuid) {
     auto quirks = header.initQuirks();
     quirks.setExplicitProcMem(false);
     quirks.setSpecialLibrrpage(false);
+    quirks.setPkeyAllocRecordedExtraRegs(true);
+    quirks.setBufferedSyscallForcedTick(true);
   }
   // Add a random UUID to the trace metadata. This lets tools identify a trace
   // easily.
@@ -1427,9 +1443,11 @@ void TraceWriter::close(CloseStatus status, const TraceUuid* uuid) {
   }
   header.setOk(status == CLOSE_OK);
   header.setChaosMode(chaos_mode ? trace::ChaosMode::KNOWN_TRUE : trace::ChaosMode::KNOWN_FALSE);
-  MemoryRange exclusion_range = AddressSpace::get_global_exclusion_range();
+  MemoryRange exclusion_range = AddressSpace::get_global_exclusion_range(nullptr);
   header.setExclusionRangeStart(exclusion_range.start().as_int());
   header.setExclusionRangeEnd(exclusion_range.end().as_int());
+  header.setRuntimePageSize(page_size());
+  header.setPreloadLibraryPageSize(PRELOAD_LIBRARY_PAGE_SIZE);
 
   try {
     writePackedMessageToFd(version_fd, header_msg);
@@ -1558,6 +1576,20 @@ TraceReader::TraceReader(const string& dir)
   PackedFdMessageReader header_msg(version_fd);
 
   trace::Header::Reader header = header_msg.getRoot<trace::Header>();
+  uint16_t syscallbuf_protocol_version = header.getSyscallbufProtocolVersion();
+  if (syscallbuf_protocol_version > SYSCALLBUF_PROTOCOL_VERSION) {
+    fprintf(stderr, "\n"
+                    "rr: error: Recorded trace `%s' has an incompatible "
+                    "syscallbuf protocol version %d; expected\n"
+                    "           %d.  Did you record `%s' with an older version "
+                    "of rr?  If so,\n"
+                    "           you'll need to replay `%s' with that older "
+                    "version.  Otherwise,\n"
+                    "           your trace is likely corrupted.\n"
+                    "\n",
+            path.c_str(), syscallbuf_protocol_version, SYSCALLBUF_PROTOCOL_VERSION, path.c_str(), path.c_str());
+    exit(EX_DATAERR);
+  }
   bind_to_cpu = header.getBindToCpu();
   preload_thread_locals_recorded_ = header.getPreloadThreadLocalsRecorded();
   ticks_semantics_ = from_trace_ticks_semantics(header.getTicksSemantics());
@@ -1572,6 +1604,12 @@ TraceReader::TraceReader(const string& dir)
     if (quirks.getSpecialLibrrpage()) {
       quirks_ |= SpecialLibRRpage;
     }
+    if (quirks.getPkeyAllocRecordedExtraRegs()) {
+      quirks_ |= PkeyAllocRecordedExtraRegs;
+    }
+    if (quirks.getBufferedSyscallForcedTick()) {
+      quirks_ |= BufferedSyscallForcedTick;
+    }
   }
   Data::Reader uuid = header.getUuid();
   uuid_ = unique_ptr<TraceUuid>(new TraceUuid());
