@@ -8,6 +8,7 @@
 #include <vector>
 
 #include "core.h"
+#include "log.h"
 #include "remote_ptr.h"
 
 namespace rr {
@@ -31,6 +32,8 @@ inline bool is_x86ish(SupportedArch arch_) {
   return arch_ == x86 || arch_ == x86_64;
 }
 
+int to_audit_arch(SupportedArch arch);
+
 template <SupportedArch a, typename system_type, typename rr_type>
 struct Verifier {
   // Optimistically say we are the same size.
@@ -64,6 +67,69 @@ template <typename T> struct Verifier<RR_NATIVE_ARCH, T, T> {
 #endif
 
 struct KernelConstants {
+  // These are the same across all architectures. The kernel defines them for
+  // all architectures in the uapi headers, but the libc's headers may not.
+  // Further, the libc headers may conflict with the kernel headers, so for
+  // simplicitly, we just define everything here:
+  static const int PTRACE_TRACEME = 0;
+  static const int PTRACE_PEEKTEXT = 1;
+  static const int PTRACE_PEEKDATA = 2;
+  static const int PTRACE_PEEKUSR = 3;
+  // PEEKUSER is a libc alias.
+  static const int PTRACE_PEEKUSER = PTRACE_PEEKUSR;
+  static const int PTRACE_POKETEXT = 4;
+  static const int PTRACE_POKEDATA = 5;
+  static const int PTRACE_POKEUSR = 6;
+  // POKEUSER is a libc alias.
+  static const int PTRACE_POKEUSER = PTRACE_POKEUSR;
+  static const int PTRACE_CONT = 7;
+  static const int PTRACE_KILL = 8;
+  static const int PTRACE_SINGLESTEP = 9;
+  // PTRACE_GETREGS actually does vary but it's 12 on every arch we care about.
+  static const int PTRACE_GETREGS = 12;
+  static const int PTRACE_GETFPREGS = 14;
+  static const int PTRACE_SETFPREGS = 15;
+  static const int PTRACE_ATTACH = 16;
+  static const int PTRACE_DETACH = 17;
+  static const int PTRACE_SYSCALL = 24;
+  static const int PTRACE_SETOPTIONS = 0x4200;
+  static const int PTRACE_GETEVENTMSG = 0x4201;
+  static const int PTRACE_GETSIGINFO = 0x4202;
+  static const int PTRACE_SETSIGINFO = 0x4203;
+  static const int PTRACE_GETREGSET = 0x4204;
+  static const int PTRACE_SETREGSET = 0x4205;
+  static const int PTRACE_SEIZE = 0x4206;
+  static const int PTRACE_INTERRUPT = 0x4207;
+  static const int PTRACE_LISTEN = 0x4208;
+  static const int PTRACE_GETSIGMASK = 0x420a;
+  static const int PTRACE_SETSIGMASK = 0x420b;
+  static const int PTRACE_GET_SYSCALL_INFO = 0x420e;
+
+  static const int PTRACE_EVENT_NONE = 0;
+  static const int PTRACE_EVENT_FORK = 1;
+  static const int PTRACE_EVENT_VFORK = 2;
+  static const int PTRACE_EVENT_CLONE = 3;
+  static const int PTRACE_EVENT_EXEC = 4;
+  static const int PTRACE_EVENT_VFORK_DONE = 5;
+  static const int PTRACE_EVENT_EXIT = 6;
+  static const int PTRACE_EVENT_SECCOMP = 7; // ubuntu 12.10 and future kernels
+  static const int PTRACE_EVENT_SECCOMP_OBSOLETE = 8; // ubuntu 12.04
+  static const int PTRACE_EVENT_STOP = 128;
+
+  static const int PTRACE_O_TRACESYSGOOD = 1;
+  static const int PTRACE_O_TRACEFORK = 1 << PTRACE_EVENT_FORK;
+  static const int PTRACE_O_TRACEVFORK = 1 << PTRACE_EVENT_VFORK;
+  static const int PTRACE_O_TRACECLONE = 1 << PTRACE_EVENT_CLONE;
+  static const int PTRACE_O_TRACEEXEC = 1 << PTRACE_EVENT_EXEC;
+  static const int PTRACE_O_TRACEEXIT = 1 << PTRACE_EVENT_EXIT;
+  static const int PTRACE_O_TRACESECCOMP = 1 << PTRACE_EVENT_SECCOMP;
+  static const int PTRACE_O_EXITKILL = 1 << 20;
+
+  static const int PTRACE_SYSCALL_INFO_NONE = 0;
+  static const int PTRACE_SYSCALL_INFO_ENTRY = 1;
+  static const int PTRACE_SYSCALL_INFO_EXIT = 2;
+  static const int PTRACE_SYSCALL_INFO_SECCOMP = 3;
+
   static const ::size_t SIGINFO_MAX_SIZE = 128;
 
   // These types are the same size everywhere.
@@ -175,6 +241,17 @@ struct WordSize32Defs {
     uint16_t e_shstrndx;
   } ElfEhdr;
   RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Ehdr, ElfEhdr);
+  typedef struct {
+    uint32_t p_type;
+    uint32_t p_offset;
+    uint32_t p_vaddr;
+    uint32_t p_paddr;
+    uint32_t p_filesz;
+    uint32_t p_memsz;
+    uint32_t p_flags;
+    uint32_t p_align;
+  } ElfPhdr;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Phdr, ElfPhdr);
   typedef struct {
     uint32_t sh_name;
     uint32_t sh_type;
@@ -188,6 +265,11 @@ struct WordSize32Defs {
     uint32_t sh_entsize;
   } ElfShdr;
   RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf32_Shdr, ElfShdr);
+  typedef struct {
+    uint32_t ch_type;
+    uint32_t ch_size;
+    uint32_t ch_addralign;
+  } ElfChdr;
   typedef struct {
     uint32_t st_name;
     uint32_t st_value;
@@ -254,6 +336,17 @@ struct WordSize64Defs {
     uint16_t e_shstrndx;
   } ElfEhdr;
   RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Ehdr, ElfEhdr);
+  typedef struct {
+    uint32_t p_type;
+    uint32_t p_flags;
+    uint64_t p_offset;
+    uint64_t p_vaddr;
+    uint64_t p_paddr;
+    uint64_t p_filesz;
+    uint64_t p_memsz;
+    uint64_t p_align;
+  } ElfPhdr;
+  RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Phdr, ElfPhdr);
   typedef struct {
     uint32_t sh_name;
     uint32_t sh_type;
@@ -267,6 +360,12 @@ struct WordSize64Defs {
     uint64_t sh_entsize;
   } ElfShdr;
   RR_VERIFY_TYPE_ARCH(RR_NATIVE_ARCH, ::Elf64_Shdr, ElfShdr);
+  typedef struct {
+    uint32_t ch_type;
+    uint32_t ch_reserved;
+    uint64_t ch_size;
+    uint64_t ch_addralign;
+  } ElfChdr;
   typedef struct {
     uint32_t st_name;
     uint8_t st_info;
@@ -349,21 +448,6 @@ struct BaseArch : public wordsize,
   typedef uint64_t __u64;
   typedef __u64 aligned_u64 __attribute((aligned(8)));
 
-  // These are the same across all architectures. The kernel defines them for
-  // all architectures in the uapi headers, but the libc's headers may not.
-  // Further, the libc headers may conflict with the kernel headers, so for
-  // simplicitly, we just define everything here:
-  static const int PTRACE_TRACEME = 0;
-  static const int PTRACE_PEEKTEXT = 1;
-  static const int PTRACE_PEEKDATA = 2;
-  static const int PTRACE_PEEKUSR = 3;
-  static const int PTRACE_POKETEXT = 4;
-  static const int PTRACE_POKEDATA = 5;
-  static const int PTRACE_POKEUSR = 6;
-  static const int PTRACE_CONT = 7;
-  static const int PTRACE_KILL = 8;
-  static const int PTRACE_SINGLESTEP = 9;
-
   // If they are defined in the header, undef them now.
   // In rr, we always refer to them as these constants.
 #undef PTRACE_GETREGS
@@ -611,12 +695,12 @@ struct BaseArch : public wordsize,
     tcflag_t c_cflag;
     tcflag_t c_lflag;
     cc_t c_line;
-    cc_t c_cc[32];
-    char _padding[3];
-    speed_t c_ispeed;
-    speed_t c_ospeed;
+    cc_t c_cc[19];
   };
-  RR_VERIFY_TYPE(termios);
+  /* We don't verify termios because the kernel and glibc don't agree on its
+   * layout and ensuring that we only have the kernel termios visible here is
+   * a pain.
+   */
 
   struct termio {
     unsigned_short c_iflag;
@@ -1155,18 +1239,12 @@ struct BaseArch : public wordsize,
   };
   RR_VERIFY_TYPE(__sysctl_args);
 
+  // libc reserves some space in the user facing structures for future
+  // extensibility, so we are careful to use the kernel definition here.
   typedef struct {
     unsigned_long __val[64 / (8 * sizeof(unsigned_long))];
   } kernel_sigset_t;
 
-  // libc reserves some space in the user facing structures for future
-  // extensibility.
-  typedef struct {
-    unsigned_long __val[1024 / (8 * sizeof(unsigned_long))];
-  } __sigset_t;
-  typedef __sigset_t sigset_t;
-  RR_VERIFY_TYPE(sigset_t);
-
   typedef struct {
     ptr<const kernel_sigset_t> ss;
     size_t ss_len;
@@ -1531,6 +1609,21 @@ struct BaseArch : public wordsize,
     struct hci_dev_stats stat;
   };
 
+  struct rseq_t {
+    uint32_t cpu_id_start;
+    uint32_t cpu_id;
+    uint64_t rseq_cs;
+    uint32_t flags;
+  } __attribute__((aligned(32)));
+
+  struct rseq_cs {
+    uint32_t version;
+    uint32_t flags;
+    uint64_t start_ip;
+    uint64_t post_commit_offset;
+    uint64_t abort_ip;
+  } __attribute__((aligned(32)));
+
   typedef struct ifbond {
     int32_t bond_mode;
     int32_t num_slaves;
@@ -1831,6 +1924,28 @@ struct BaseArch : public wordsize,
     uint8_t cdte_datamode;
   };
   RR_VERIFY_TYPE(cdrom_tocentry);
+
+  struct ptrace_syscall_info {
+    uint8_t op;
+    uint32_t arch;
+    uint64_t instruction_pointer;
+    uint64_t stack_pointer;
+    union {
+        struct {
+            uint64_t nr;
+            uint64_t args[6];
+        } entry;
+        struct {
+            int64_t rval;
+            uint8_t is_error;
+        } exit;
+        struct {
+            uint64_t nr;
+            uint64_t args[6];
+            uint32_t ret_data;
+        } seccomp;
+    };
+  };
 };
 
 struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
@@ -1899,8 +2014,8 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
     uint64_t fs;
     uint64_t gs;
   };
-  RR_VERIFY_TYPE_ARCH(SupportedArch::x86_64, ::user_regs_struct,
-                      user_regs_struct);
+  RR_VERIFY_TYPE_X86_ARCH(SupportedArch::x86_64, ::user_regs_struct,
+                          user_regs_struct);
 
   struct sigcontext {
     uint64_t r8;
@@ -1972,8 +2087,11 @@ struct X64Arch : public BaseArch<SupportedArch::x86_64, WordSize64Defs> {
     uint64_t magic;
     char u_comm[32];
     uint64_t u_debugreg[8];
+    uint64_t error_code;
+    uint64_t fault_address;
   };
-  RR_VERIFY_TYPE_X86_ARCH(SupportedArch::x86_64, ::user, user);
+  // Can't verify this one because glibc leaves out the last two members and the
+  // kernel header isn't available to userspace.
 
   struct stat {
     dev_t st_dev;
@@ -2351,7 +2469,7 @@ struct ARM64Arch : public GenericArch<SupportedArch::aarch64, WordSize64Defs> {
     struct hw_bp dbg_regs[16];
   };
 
-  struct sigcontext {
+  struct __attribute((aligned(16))) sigcontext {
     __u64 fault_addr;
     user_pt_regs regs;
     // ISA extension state follows here
@@ -2361,8 +2479,8 @@ struct ARM64Arch : public GenericArch<SupportedArch::aarch64, WordSize64Defs> {
     unsigned long	uc_flags;
     ptr<ucontext> uc_link;
     stack_t		  uc_stack;
-    sigset_t	  uc_sigmask;
-    uint8_t __unused[1024 / 8 - sizeof(sigset_t)];
+    kernel_sigset_t	  uc_sigmask;
+    uint8_t __unused1[1024 / 8 - sizeof(kernel_sigset_t)];
     struct sigcontext uc_mcontext;
   };
 
