@@ -14,7 +14,6 @@
 #include <signal.h>
 #include <stdint.h>
 #include <sys/ioctl.h>
-#include <sys/ptrace.h>
 #include <sys/wait.h>
 
 namespace rr {
@@ -25,47 +24,74 @@ namespace rr {
  * across architectures; those definitions belong in kernel_abi.h.
  */
 
-#ifndef PTRACE_EVENT_NONE
-#define PTRACE_EVENT_NONE 0
-#endif
-#ifndef PTRACE_EVENT_STOP
-#define PTRACE_EVENT_STOP 128
-#endif
+#define KERNEL_CONSTANT(constant) \
+  constant = KernelConstants::constant
 
-#ifndef PTRACE_GETREGSET
-#define PTRACE_GETREGSET 0x4204
-#endif
-#ifndef PTRACE_SETREGSET
-#define PTRACE_SETREGSET 0x4205
-#endif
+enum _ptrace_request {
+  KERNEL_CONSTANT(PTRACE_TRACEME),
+  KERNEL_CONSTANT(PTRACE_PEEKTEXT),
+  KERNEL_CONSTANT(PTRACE_PEEKDATA),
+  KERNEL_CONSTANT(PTRACE_PEEKUSR),
+  KERNEL_CONSTANT(PTRACE_PEEKUSER),
+  KERNEL_CONSTANT(PTRACE_POKETEXT),
+  KERNEL_CONSTANT(PTRACE_POKEDATA),
+  KERNEL_CONSTANT(PTRACE_POKEUSR),
+  KERNEL_CONSTANT(PTRACE_POKEUSER),
+  KERNEL_CONSTANT(PTRACE_CONT),
+  KERNEL_CONSTANT(PTRACE_KILL),
+  KERNEL_CONSTANT(PTRACE_SINGLESTEP),
+  KERNEL_CONSTANT(PTRACE_GETREGS),
+  KERNEL_CONSTANT(PTRACE_GETFPREGS),
+  KERNEL_CONSTANT(PTRACE_SETFPREGS),
+  KERNEL_CONSTANT(PTRACE_ATTACH),
+  KERNEL_CONSTANT(PTRACE_DETACH),
+  KERNEL_CONSTANT(PTRACE_SYSCALL),
+  KERNEL_CONSTANT(PTRACE_SETOPTIONS),
+  KERNEL_CONSTANT(PTRACE_GETEVENTMSG),
+  KERNEL_CONSTANT(PTRACE_GETSIGINFO),
+  KERNEL_CONSTANT(PTRACE_SETSIGINFO),
+  KERNEL_CONSTANT(PTRACE_GETREGSET),
+  KERNEL_CONSTANT(PTRACE_SETREGSET),
+  KERNEL_CONSTANT(PTRACE_SEIZE),
+  KERNEL_CONSTANT(PTRACE_INTERRUPT),
+  KERNEL_CONSTANT(PTRACE_LISTEN),
+  KERNEL_CONSTANT(PTRACE_GETSIGMASK),
+  KERNEL_CONSTANT(PTRACE_SETSIGMASK),
+  KERNEL_CONSTANT(PTRACE_GET_SYSCALL_INFO),
+};
 
-#ifndef PTRACE_SEIZE
-#define PTRACE_SEIZE 0x4206
-#endif
-#ifndef PTRACE_INTERRUPT
-#define PTRACE_INTERRUPT 0x4207
-#endif
-#ifndef PTRACE_LISTEN
-#define PTRACE_LISTEN	0x4208
-#endif
+enum _ptrace_eventcodes {
+  KERNEL_CONSTANT(PTRACE_EVENT_NONE),
+  KERNEL_CONSTANT(PTRACE_EVENT_FORK),
+  KERNEL_CONSTANT(PTRACE_EVENT_VFORK),
+  KERNEL_CONSTANT(PTRACE_EVENT_CLONE),
+  KERNEL_CONSTANT(PTRACE_EVENT_EXEC),
+  KERNEL_CONSTANT(PTRACE_EVENT_VFORK_DONE),
+  KERNEL_CONSTANT(PTRACE_EVENT_EXIT),
+  KERNEL_CONSTANT(PTRACE_EVENT_SECCOMP),
+  KERNEL_CONSTANT(PTRACE_EVENT_SECCOMP_OBSOLETE),
+  KERNEL_CONSTANT(PTRACE_EVENT_STOP),
+};
 
-#ifndef PTRACE_GETSIGMASK
-#define PTRACE_GETSIGMASK 0x420a
-#endif
+enum _ptrace_options {
+  KERNEL_CONSTANT(PTRACE_O_TRACESYSGOOD),
+  KERNEL_CONSTANT(PTRACE_O_TRACEFORK),
+  KERNEL_CONSTANT(PTRACE_O_TRACEVFORK),
+  KERNEL_CONSTANT(PTRACE_O_TRACECLONE),
+  KERNEL_CONSTANT(PTRACE_O_TRACEEXEC),
+  KERNEL_CONSTANT(PTRACE_O_TRACEEXIT),
+  KERNEL_CONSTANT(PTRACE_O_TRACESECCOMP),
+  KERNEL_CONSTANT(PTRACE_O_EXITKILL),
+};
 
-#ifndef PTRACE_SETSIGMASK
-#define PTRACE_SETSIGMASK 0x420b
-#endif
+enum _ptrace_get_syscall_info_op {
+  KERNEL_CONSTANT(PTRACE_SYSCALL_INFO_NONE),
+  KERNEL_CONSTANT(PTRACE_SYSCALL_INFO_ENTRY),
+  KERNEL_CONSTANT(PTRACE_SYSCALL_INFO_EXIT),
+  KERNEL_CONSTANT(PTRACE_SYSCALL_INFO_SECCOMP),
+};
 
-#ifndef PTRACE_O_TRACESECCOMP
-#define PTRACE_O_TRACESECCOMP 0x00000080
-#define PTRACE_EVENT_SECCOMP_OBSOLETE 8 // ubuntu 12.04
-#define PTRACE_EVENT_SECCOMP 7          // ubuntu 12.10 and future kernels
-#endif
-
-#ifndef PTRACE_O_EXITKILL
-#define PTRACE_O_EXITKILL (1 << 20)
-#endif
+#undef KERNEL_CONSTANT
 
 #ifndef SECCOMP_SET_MODE_STRICT
 #define SECCOMP_SET_MODE_STRICT 0
@@ -325,6 +351,14 @@ struct rr_input_mask {
 #define PR_CAP_AMBIENT_CLEAR_ALL 4
 #endif
 
+// New in the 3.17 kernel.
+#ifndef VIDIOC_QUERY_EXT_CTRL
+/* This definition omits the size because in prepare_ioctl
+   it is masked away anyway. And the real size is taken from
+   the real request by _IOC_SIZE(request). */
+#define VIDIOC_QUERY_EXT_CTRL _IOWR('V', 103, 0)
+#endif
+
 // New in the 4.6 kernel.
 #ifndef CLONE_NEWCGROUP
 #define CLONE_NEWCGROUP 0x02000000
@@ -425,6 +459,39 @@ enum {
 #define SEGV_PKUERR 4
 #endif
 
+#define RR_RSEQ_FLAG_UNREGISTER (1 << 0)
+#define RR_RSEQ_CS_FLAG_NO_RESTART_ON_PREEMPT_BIT 0
+#define RR_RSEQ_CS_FLAG_NO_RESTART_ON_SIGNAL_BIT 1
+#define RR_RSEQ_CS_FLAG_NO_RESTART_ON_MIGRATE_BIT 2
+#define RR_RSEQ_CPU_ID_UNINITIALIZED -1
+
+// New in the 5.15 kernel
+#ifndef BLKGETDISKSEQ
+#define BLKGETDISKSEQ _IOR(0x12,128,__u64)
+#endif
+
+// New in the 5.17 kernel
+#ifndef PR_SET_VMA
+#define PR_SET_VMA 0x53564d41
+#endif
+#ifndef PR_SET_VMA_ANON_NAME
+#define PR_SET_VMA_ANON_NAME 0
+#endif
+
+// Technically not "kernel" constants, exactly, since these are defined
+// in libc, but required for compat with older libcs like the rest of
+// this file.
+#ifndef SHF_COMPRESSED
+#define SHF_COMPRESSED (1 << 11)
+#endif
+
+#ifndef ELFCOMPRESS_ZLIB
+#define ELFCOMPRESS_ZLIB 1
+#endif
+
 } // namespace rr
 
+// We can't include libc's ptrace.h, so declare this here.
+extern "C" long int ptrace (enum rr::_ptrace_request _request, ...);
+
 #endif /* RR_KERNEL_SUPPLEMENT_H_ */
