@@ -3,6 +3,7 @@
 #include "log.h"
 
 #include <stdlib.h>
+#include <string.h>
 
 #include <deque>
 #include <fstream>
@@ -17,11 +18,29 @@
 #include "RecordSession.h"
 #include "core.h"
 #include "ftrace.h"
+#include "kernel_abi.h"
 #include "kernel_metadata.h"
 #include "util.h"
 
 using namespace std;
 
+ostream& operator<<(ostream& stream, const siginfo_t& siginfo) {
+  stream << "{signo:" << rr::signal_name(siginfo.si_signo)
+         << ",errno:" << rr::errno_name(siginfo.si_errno)
+         << ",code:" << rr::sicode_name(siginfo.si_code, siginfo.si_signo);
+  switch (siginfo.si_signo) {
+    case SIGILL:
+    case SIGFPE:
+    case SIGSEGV:
+    case SIGBUS:
+    case SIGTRAP:
+      stream << ",addr:" << siginfo.si_addr;
+      break;
+  }
+  stream << "}";
+  return stream;
+}
+
 namespace rr {
 
 struct LogModule {
@@ -45,7 +64,7 @@ static LogLevel to_log_level(const string& str) {
   if (str == "fatal") {
     return LOG_fatal;
   }
-  fprintf(stderr, "Log level %s in RR_LOG is not valid, assuing 'fatal'\n",
+  fprintf(stderr, "Log level %s in RR_LOG is not valid, assuming 'fatal'\n",
           str.c_str());
   return LOG_fatal;
 }
@@ -102,6 +121,58 @@ size_t log_buffer_size;
 
 static void flush_log_file() { log_file->flush(); }
 
+static void init_log_globals();
+
+void apply_log_spec(const char *spec) {
+  init_log_globals();
+  char *env = strdup(spec);
+  DEBUG_ASSERT(env);
+  for (int i = 0; env[i]; ++i) {
+    env[i] = simple_to_lower(env[i]);
+  }
+  char* p = env;
+  while (*p) {
+    char* end = strchrnul(p, ',');
+    char* sep = strchrnul(p, ':');
+    string n;
+    LogLevel level;
+    if (sep >= end) {
+      n = string(p, end - p);
+      level = LOG_debug;
+    } else {
+      n = string(p, sep - p);
+      if (sep + 1 == end) {
+        level = LOG_fatal;
+      } else {
+        level = to_log_level(string(sep + 1, end - (sep + 1)));
+      }
+    }
+    if (n == "" || n == "all") {
+      level_map->clear();
+      default_level = level;
+    } else {
+      (*level_map)[n] = level;
+    }
+    if (*end) {
+      p = end + 1;
+    } else {
+      p = end;
+    }
+  }
+  free(env);
+}
+
+void apply_log_spec_from_env() {
+  const char* log_env = "RR_LOG";
+  if (running_under_rr()) {
+    log_env = "RR_UNDER_RR_LOG";
+  }
+  char* env = getenv(log_env);
+  if (env) {
+    apply_log_spec(env);
+  }
+}
+
 static void init_log_globals() {
   if (log_globals_initialized) {
     return;
@@ -141,48 +212,7 @@ static void init_log_globals() {
     log_file = &cerr;
   }
 
-  const char* log_env = "RR_LOG";
-  if (running_under_rr()) {
-    log_env = "RR_UNDER_RR_LOG";
-  }
-  char* env = getenv(log_env);
-  if (env) {
-    env = strdup(env);
-    DEBUG_ASSERT(env);
-    for (int i = 0; env[i]; ++i) {
-      env[i] = simple_to_lower(env[i]);
-    }
-    char* p = env;
-    while (*p) {
-      char* end = strchrnul(p, ',');
-      char* sep = strchrnul(p, ':');
-      string n;
-      LogLevel level;
-      if (sep >= end) {
-        n = string(p, end - p);
-        level = LOG_debug;
-      } else {
-        n = string(p, sep - p);
-        if (sep + 1 == end) {
-          level = LOG_fatal;
-        } else {
-          level = to_log_level(string(sep + 1, end - (sep + 1)));
-        }
-      }
-      if (n == "" || n == "all") {
-        level_map->clear();
-        default_level = level;
-      } else {
-        (*level_map)[n] = level;
-      }
-      if (*end) {
-        p = end + 1;
-      } else {
-        p = end;
-      }
-    }
-    free(env);
-  }
+  apply_log_spec_from_env();
 }
 
 static LogLevel get_log_level(const string& name) {
@@ -281,17 +311,21 @@ static void flush_log_stream() {
   logging_stream->str(string());
 }
 
-void flush_log_buffer() {
-  if (log_buffer) {
-    for (char c : *log_buffer) {
+void flush_log_buffer(unique_ptr<deque<char>> &this_log_buffer) {
+  if (this_log_buffer) {
+    for (char c : *this_log_buffer) {
       // We could accumulate in a string to speed things up, but this could get
       // called in low-memory situations so be safe.
       *log_file << c;
     }
-    log_buffer->clear();
+    this_log_buffer->clear();
   }
 }
 
+void flush_log_buffer() {
+  flush_log_buffer(log_buffer);
+}
+
 template <typename T>
 static void write_prefix(T& stream, LogLevel level, const char* file, int line,
                          const char* function) {
@@ -391,11 +425,19 @@ static void emergency_debug(Task* t) {
   if (record_session) {
     record_session->close_trace_writer(TraceWriter::CLOSE_ERROR);
   }
+
+  // Capture the log buffer now to prevent the log messages from the trace
+  // stream read below from overwriting any data from the actual failure.
+  flush_log_stream();
+  std::unique_ptr<deque<char>> captured_log_buffer = std::move(log_buffer);
+
   TraceStream* trace_stream = t->session().trace_stream();
   if (trace_stream) {
     dump_last_events(*trace_stream);
   }
 
+  flush_log_buffer(captured_log_buffer);
+
   if (probably_not_interactive() && !Flags::get().force_things &&
       !getenv("RUNNING_UNDER_TEST_MONITOR")) {
     errno = 0;
@@ -403,8 +445,6 @@ static void emergency_debug(Task* t) {
         << "(session doesn't look interactive, aborting emergency debugging)";
   }
 
-  flush_log_buffer();
-
   GdbServer::emergency_debug(t);
   FATAL() << "Can't resume execution from invalid state";
 }
