@@ -26,13 +26,26 @@
 #endif
 #endif
 
+#ifndef __BIONIC__
+
+static int (*real_pthread_mutex_init)(void* mutex, const void* attr);
+static int (*real_pthread_mutex_lock)(void* mutex);
+static int (*real_pthread_mutex_trylock)(void* mutex);
+static int (*real_pthread_mutex_timedlock)(void* mutex,
+                                           const struct timespec* abstime);
+
+static void __attribute__((constructor)) init_override(void) {
+  real_pthread_mutex_init = dlsym(RTLD_NEXT, "pthread_mutex_init");
+  real_pthread_mutex_lock = dlsym(RTLD_NEXT, "pthread_mutex_lock");
+  real_pthread_mutex_trylock = dlsym(RTLD_NEXT, "pthread_mutex_trylock");
+  real_pthread_mutex_timedlock = dlsym(RTLD_NEXT, "pthread_mutex_timedlock");
+}
+
 static void fix_mutex_kind(pthread_mutex_t* mutex) {
   /* Disable priority inheritance. */
   mutex->__data.__kind &= ~PTHREAD_MUTEX_PRIO_INHERIT_NP;
 }
 
-extern int __pthread_mutex_init(pthread_mutex_t* mutex,
-                                const pthread_mutexattr_t* attr);
 #ifdef DOUBLE_UNDERSCORE_PTHREAD_LOCK_AVAILABLE
 /*
  * We need to able to call directly to __pthread_mutex_lock and
@@ -42,6 +55,8 @@ extern int __pthread_mutex_init(pthread_mutex_t* mutex,
  * to use a pthreads-based implementation). So before our pointers are set
  * up, call these.
  */
+extern int __pthread_mutex_init(pthread_mutex_t* mutex,
+                                const pthread_mutexattr_t* attr);
 extern int __pthread_mutex_lock(pthread_mutex_t* mutex);
 extern int __pthread_mutex_trylock(pthread_mutex_t* mutex);
 #endif
@@ -51,25 +66,28 @@ int pthread_mutex_init(pthread_mutex_t* mutex,
   int ret;
   pthread_mutexattr_t realattr;
 
-  if (!attr) {
-    return __pthread_mutex_init(mutex, NULL);
+  if (attr) {
+    /* We wish to enforce the use of plain (no PI) mutex to avoid
+     * needing to handle PI futex() operations.
+     * We also wish to ensure that pthread_mutexattr_getprotocol()
+     * still returns the requested protocol.
+     * So we copy the attribute and force PTHREAD_PRIO_NONE.
+     */
+    memcpy(&realattr, attr, sizeof(realattr));
+    ret = pthread_mutexattr_setprotocol(&realattr, PTHREAD_PRIO_NONE);
+    if (ret) {
+      return ret;
+    }
+    attr = &realattr;
   }
-
-  /* We wish to enforce the use of plain (no PI) mutex to avoid
-   * needing to handle PI futex() operations.
-   * We also wish to ensure that pthread_mutexattr_getprotocol()
-   * still returns the requested protocol.
-   * So we copy the attribute and force PTHREAD_PRIO_NONE.
-   */
-  memcpy(&realattr, attr, sizeof(realattr));
-  ret = pthread_mutexattr_setprotocol(&realattr, PTHREAD_PRIO_NONE);
-  if (ret) {
-    return ret;
+  if (!real_pthread_mutex_init) {
+#ifdef DOUBLE_UNDERSCORE_PTHREAD_LOCK_AVAILABLE
+    return __pthread_mutex_init(mutex, attr);
+#else
+    real_pthread_mutex_init = dlsym(RTLD_NEXT, "pthread_mutex_init");
+#endif
   }
-  if (real_pthread_mutex_init) {
-    return real_pthread_mutex_init(mutex, &realattr);
-  }
-  return __pthread_mutex_init(mutex, &realattr);
+  return real_pthread_mutex_init(mutex, attr);
 }
 
 /* Prevent use of lock elision; Haswell's TSX/RTM features used by
@@ -111,29 +129,7 @@ int pthread_mutex_trylock(pthread_mutex_t* mutex) {
   return real_pthread_mutex_trylock(mutex);
 }
 
-/**
- * Exported glibc synonym for |sysconf()|.  We can't use |dlsym()| to
- * resolve the next "sysconf" symbol, because
- *  - dlysym usually calls malloc()
- *  - custom allocators like jemalloc may use sysconf()
- *  - if our sysconf wrapper is re-entered during initialization, it
- *    has nothing to fall back on to get the conf name, and chaos will
- *    likely ensue if we return something random.
- */
-long __sysconf(int name);
-
-/**
- *  Pretend that only 1 processor is configured/online, because rr
- *  binds all tracees to one logical CPU.
- */
-long sysconf(int name) {
-  switch (name) {
-    case _SC_NPROCESSORS_ONLN:
-    case _SC_NPROCESSORS_CONF:
-      return globals.pretend_num_cores ? globals.pretend_num_cores : 1;
-  }
-  return __sysconf(name);
-}
+#endif
 
 typedef void* Dlopen(const char* filename, int flags);
 
@@ -162,6 +158,9 @@ void* XShmCreateImage(__attribute__((unused)) register void* dpy,
   return 0;
 }
 
+RR_HIDDEN char impose_syscall_delay;
+RR_HIDDEN char impose_spurious_desched;
+
 /**
  * This is for testing purposes only.
  */
@@ -173,6 +172,16 @@ void delayed_syscall(struct syscall_info* info) {
   impose_syscall_delay = 0;
 }
 
+/**
+ * This is for testing purposes only.
+ * Note that this must be defined outside of the syscallbuf code.
+ * Otherwise, the signal recording code may expect exit from this function
+ * to trigger the syscallbuf exit breakpoint.
+ */
+void* syscallbuf_ptr(void) {
+  return ((struct preload_thread_locals*)PRELOAD_THREAD_LOCALS_ADDR)->buffer;
+}
+
 /**
  * This is for testing purposes only.
  */
@@ -184,10 +193,46 @@ void spurious_desched_syscall(struct syscall_info* info) {
   impose_spurious_desched = 0;
 }
 
+/**
+ * clang's LeakSanitizer has regular threads call sched_yield() in a loop while
+ * a helper thread ptrace-attaches to them. If we let sched_yield() enter the
+ * syscallbuf, the helper thread sees that the regular thread SP register
+ * is pointing to the syscallbuf alt-stack, outside the stack region it
+ * expects, which causes it to freak out.
+ * So, override sched_yield() to perform the syscall in a way that can't
+ * be syscall-buffered.
+ */
+int sched_yield(void) {
+#ifdef __i386__
+  // We have no syscall hook for `syscall` followed by `inc %ecx`
+  int trash;
+  asm volatile ("int $0x80; inc %0" : "=c"(trash) : "a"(SYS_sched_yield));
+#elif defined(__x86_64__)
+  // We have no syscall hook for `syscall` followed by `inc %ecx`
+  int trash;
+  asm volatile ("syscall; inc %0" : "=c"(trash) : "a"(SYS_sched_yield));
+#elif defined(__aarch64__)
+  register long x8 __asm__("x8") = SYS_sched_yield;
+  // We explicitly blacklisted syscall that follows `mov x8, 0xdc`
+  // to avoid patching clone. Abuse that to prevent this from being patched.
+  __asm__ __volatile__("b 1f\n\t"
+                       "mov x8, 0xdc\n"
+                       "1:\n\t"
+                       "svc 0\n"
+                       :: "r"(x8) : "x0", "x30"); // x30 = lr
+#else
+#error "Unknown architecture"
+#endif
+  return 0;
+}
+
+#ifndef __aarch64__
+
 /**
  * glibc geteuid() can be compiled to instructions ending in "syscall; ret"
  * which sometimes can't be hooked. So override it here with something that
  * can be hooked.
+ * This is not an issue on aarch64 since we only need to patch a single instruction.
  */
 uid_t geteuid(void) {
 #ifdef __i386__
@@ -197,26 +242,11 @@ uid_t geteuid(void) {
 #endif
 }
 
-/**
- * clang's LeakSanitizer has regular threads call sched_yield() in a loop while
- * a helper thread ptrace-attaches to them. If we let sched_yield() enter the
- * syscallbuf, the helper thread sees that the regular thread SP register
- * is pointing to the syscallbuf alt-stack, outside the stack region it
- * expects, which causes it to freak out.
- * So, override sched_yield() to perform the syscall in a way that can't
- * be syscall-buffered. (We have no syscall hook for `syscall` followed by
- * `inc %ecx`).
- */
-int sched_yield(void) {
-  int trash;
-#ifdef __i386__
-  asm volatile ("int $0x80; inc %0" : "=c"(trash) : "a"(SYS_sched_yield));
-#elif defined(__x86_64__)
-  asm volatile ("syscall; inc %0" : "=c"(trash) : "a"(SYS_sched_yield));
-#else
-#error "Unknown architecture"
-#endif
-  return 0;
+static void libstdcpp_not_found(void) {
+  const char msg[] = "[rr] Interposition for libstdc++ called but symbol lookups into libstdc++ failed.\n"
+    "Was libstdc++ loaded with RTLD_LOCAL? Try recording with `-v LD_PRELOAD=libstdc++.so.6`.\n"
+    "About to crash! ";
+  syscall(SYS_write, STDERR_FILENO, msg, sizeof(msg));
 }
 
 /**
@@ -229,10 +259,16 @@ void _ZNSt13random_device7_M_initERKNSt7__cxx1112basic_stringIcSt11char_traitsIc
   if (!assign_string) {
     assign_string = (void (*)(void *, char*))dlsym(RTLD_NEXT,
       "_ZNSt7__cxx1112basic_stringIcSt11char_traitsIcESaIcEE6assignEPKc");
+    if (!assign_string) {
+      libstdcpp_not_found();
+    }
   }
   assign_string(token, "/dev/urandom");
   if (!random_init) {
     random_init = (void (*)(void *, void*))dlsym(RTLD_NEXT, __func__);
+    if (!random_init) {
+      libstdcpp_not_found();
+    }
   }
   random_init(this, token);
 }
@@ -247,10 +283,18 @@ void _ZNSt13random_device7_M_initERKSs(void* this,
   if (!assign_string) {
     assign_string = (void (*)(void *, char*))dlsym(RTLD_NEXT,
       "_ZNSs6assignEPKc");
+    if (!assign_string) {
+      libstdcpp_not_found();
+    }
   }
   assign_string(token, "/dev/urandom");
   if (!random_init) {
     random_init = (void (*)(void *, void*))dlsym(RTLD_NEXT, __func__);
+    if (!random_init) {
+      libstdcpp_not_found();
+    }
   }
   random_init(this, token);
 }
+
+#endif
