@@ -51,6 +51,7 @@ static inline size_t rrstrlen(const char* s) { return strlen(s); }
 #endif
 
 #include <stdint.h>
+#include <stddef.h>
 
 static inline int strprefix(const char* s1, const char* s2) {
   while (1) {
@@ -123,7 +124,11 @@ static inline const char* extract_file_name(const char* s) {
 
 /* Must match generate_rr_page.py */
 #define RR_PAGE_ADDR 0x70000000
-#define RR_PAGE_SIZE 4096
+#ifdef __aarch64__
+#define PRELOAD_LIBRARY_PAGE_SIZE 65536
+#else
+#define PRELOAD_LIBRARY_PAGE_SIZE 4096
+#endif
 #define RR_PAGE_SYSCALL_ADDR(index)                                            \
   ((void*)(RR_PAGE_ADDR + RR_PAGE_SYSCALL_STUB_SIZE * (index)))
 #define RR_PAGE_SYSCALL_TRACED RR_PAGE_SYSCALL_ADDR(0)
@@ -140,8 +145,13 @@ static inline const char* extract_file_name(const char* s) {
 
 /* PRELOAD_THREAD_LOCALS_ADDR should not change.
  * Tools depend on this address. */
-#define PRELOAD_THREAD_LOCALS_ADDR (RR_PAGE_ADDR + RR_PAGE_SIZE)
-#define PRELOAD_THREAD_LOCALS_SIZE 104
+#define PRELOAD_THREAD_LOCALS_ADDR (RR_PAGE_ADDR + PRELOAD_LIBRARY_PAGE_SIZE)
+#ifdef __aarch64__
+#define PRELOAD_THREAD_LOCAL_SCRATCH2_SIZE (1024 + 8 * 2)
+#else
+#define PRELOAD_THREAD_LOCAL_SCRATCH2_SIZE 0
+#endif
+#define PRELOAD_THREAD_LOCALS_SIZE (104 + PRELOAD_THREAD_LOCAL_SCRATCH2_SIZE)
 
 #include "rrcalls.h"
 
@@ -235,9 +245,8 @@ struct preload_globals {
   unsigned char in_chaos;
   /* The signal to use for desched events */
   unsigned char desched_sig;
-  /* Number of cores to pretend we have. 0 means 1. rr sets this when
-   * the preload library is initialized. */
-  int pretend_num_cores;
+  /* RESERVED */
+  int reserved;
   /**
    * Set by rr.
    * For each fd, indicate a class that is valid for all fds with the given
@@ -303,11 +312,9 @@ struct preload_thread_locals {
   /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
    * rr depends on.
    */
-  int alt_stack_nesting_level;
-  /**
-   * We could use this later.
-   */
-  int unused_padding;
+  int32_t alt_stack_nesting_level;
+  /* Syscall hook saved flags (bottom 16 bits only) */
+  int32_t saved_flags;
   /* The offset of this field MUST NOT CHANGE, it is part of the preload ABI
    * rr depends on. It contains the parameters to the patched syscall, or
    * zero if we're not processing a buffered syscall. Do not depend on this
@@ -367,7 +374,15 @@ struct preload_thread_locals {
   size_t usable_scratch_size;
 
   PTR(struct msghdr) notify_control_msg;
+
+  uint8_t stub_scratch_2[PRELOAD_THREAD_LOCAL_SCRATCH2_SIZE];
 };
+#if defined(__aarch64__) && (defined(RR_IMPLEMENT_PRELOAD) || \
+                             defined(RR_IMPLEMENT_AUDIT))
+// On aarch64, we the stub_scratch_2 offset is hardcoded in the syscallbuf code
+_Static_assert(offsetof(struct preload_thread_locals, stub_scratch_2) == 8 * 13,
+               "stub_scratch_2 offset mismatch");
+#endif
 
 // The set of flags that can be set for each fd in syscallbuf_fds_disabled.
 enum syscallbuf_fd_classes {
@@ -390,6 +405,8 @@ enum syscallbuf_fd_classes {
 /**
  * Packs up the parameters passed to |SYS_rrcall_init_preload|.
  * We use this struct because it's a little cleaner.
+ * When evolving this struct, add new fields at the end and don't
+ * depend on them during replay.
  */
 TEMPLATE_ARCH
 struct rrcall_init_preload_params {
@@ -400,7 +417,7 @@ struct rrcall_init_preload_params {
   int syscallbuf_enabled;
   int syscall_patch_hook_count;
   PTR(struct syscall_patch_hook) syscall_patch_hooks;
-  PTR(void) syscallhook_vsyscall_entry;
+  PTR(void) unused;
   PTR(void) syscallbuf_code_start;
   PTR(void) syscallbuf_code_end;
   PTR(void) get_pc_thunks_start;
@@ -425,6 +442,7 @@ struct rrcall_init_preload_params {
       int breakpoint_mode_sentinel;
     };
   };
+  PTR(void) syscallbuf_syscall_hook;
 };
 
 /**
