@@ -15,7 +15,9 @@ const int SIGCHLD_SYNTHETIC = 0xbeadf00d;
 
 void disarm_desched_event(RecordTask* t);
 void arm_desched_event(RecordTask* t);
+bool handle_syscallbuf_breakpoint(RecordTask* t);
 
+enum SignalBlocked { SIG_UNBLOCKED = 0, SIG_BLOCKED = 1 };
 enum SignalHandled { SIGNAL_HANDLED, SIGNAL_PTRACE_STOP, DEFER_SIGNAL };
 /**
  * Handle the given signal for |t|.
@@ -27,7 +29,8 @@ enum SignalHandled { SIGNAL_HANDLED, SIGNAL_PTRACE_STOP, DEFER_SIGNAL };
  * desched + syscall-interruption events, or no-op.
  */
 SignalHandled handle_signal(RecordTask* t, siginfo_t* si,
-                            SignalDeterministic deterministic);
+                            SignalDeterministic deterministic,
+                            SignalBlocked signal_was_blocked);
 
 } // namespace rr
 
