@@ -31,6 +31,8 @@
 #include "AutoRemoteSyscalls.h"
 #include "EmuFs.h"
 #include "MmappedFileMonitor.h"
+#include "NonvirtualPerfCounterMonitor.h"
+#include "ODirectFileMonitor.h"
 #include "ProcFdDirMonitor.h"
 #include "ProcMemMonitor.h"
 #include "ProcStatMonitor.h"
@@ -85,8 +87,8 @@ static void init_scratch_memory(ReplayTask* t, const KernelMapping& km,
 
   {
     AutoRemoteSyscalls remote(t);
-    remote.infallible_mmap_syscall(t->scratch_ptr, sz, km.prot(),
-                                   km.flags() | MAP_FIXED, -1, 0);
+    remote.infallible_mmap_syscall_if_alive(t->scratch_ptr, sz, km.prot(),
+                                            km.flags() | MAP_FIXED, -1, 0);
     t->vm()->map(t, t->scratch_ptr, sz, km.prot(), km.flags(), 0, string(),
                  KernelMapping::NO_DEVICE, KernelMapping::NO_INODE, nullptr,
                  &km);
@@ -137,7 +139,7 @@ static bool syscall_shares_vm(Registers r)
 {
   switch (r.original_syscallno()) {
     case Arch::clone:
-      return (CLONE_VM & r.arg1());
+      return (CLONE_VM & r.orig_arg1());
     case Arch::vfork:
       return true;
     case Arch::fork:
@@ -331,7 +333,7 @@ static void restore_mapped_region(ReplayTask* t, AutoRemoteSyscalls& remote,
       // Private mapping, so O_RDONLY is always OK.
       remote.finish_direct_mmap(km.start(), km.size(), km.prot(),
                          km.flags(), data.file_name, O_RDONLY,
-                         data.data_offset_bytes / page_size(), real_file,
+                         data.data_offset_bytes, real_file,
                          real_file_name);
       device = real_file.st_dev;
       inode = real_file.st_ino;
@@ -340,9 +342,9 @@ static void restore_mapped_region(ReplayTask* t, AutoRemoteSyscalls& remote,
     case TraceReader::SOURCE_TRACE:
     case TraceReader::SOURCE_ZERO:
       flags |= MAP_ANONYMOUS;
-      remote.infallible_mmap_syscall(km.start(), km.size(), km.prot(),
-                                     (flags & ~MAP_GROWSDOWN) | MAP_FIXED, -1,
-                                     0);
+      remote.infallible_mmap_syscall_if_alive(km.start(), km.size(), km.prot(),
+                                              (flags & ~MAP_GROWSDOWN) | MAP_FIXED, -1,
+                                              0);
       // The data, if any, will be written back by
       // ReplayTask::apply_all_data_records_from_trace
       break;
@@ -445,7 +447,9 @@ static void process_execve(ReplayTask* t, const TraceFrame& trace_frame,
   const string& exe_name = datas[exe_km].file_name.empty()
                                ? kms[exe_km].fsname()
                                : datas[exe_km].file_name;
-  t->post_exec_syscall(exe_name);
+  t->post_exec_syscall(exe_name, kms[exe_km].fsname());
+  t->vm()->set_interp_base(tte.interp_base());
+  t->vm()->set_interp_name(tte.interp_name());
 
   t->fd_table()->close_after_exec(
       t, t->current_trace_frame().event().Syscall().exec_fds_to_close);
@@ -511,9 +515,9 @@ static void process_brk(ReplayTask* t) {
   if (km.flags()) {
     AutoRemoteSyscalls remote(t);
     ASSERT(t, data.source == TraceReader::SOURCE_ZERO);
-    remote.infallible_mmap_syscall(km.start(), km.size(), km.prot(),
-                                   MAP_ANONYMOUS | MAP_FIXED | km.flags(), -1,
-                                   0);
+    remote.infallible_mmap_syscall_if_alive(km.start(), km.size(), km.prot(),
+                                            MAP_ANONYMOUS | MAP_FIXED | km.flags(), -1,
+                                            0);
     t->vm()->map(t, km.start(), km.size(), km.prot(),
                  MAP_ANONYMOUS | km.flags(), 0, "[heap]",
                  KernelMapping::NO_DEVICE, KernelMapping::NO_INODE, nullptr,
@@ -536,10 +540,10 @@ static void finish_anonymous_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
   KernelMapping recorded_km = t->trace_reader().read_mapped_region(&data);
   EmuFile::shr_ptr emu_file;
   if (!(flags & MAP_SHARED)) {
-    remote.infallible_mmap_syscall(rec_addr, length, prot,
-                                   // Tell the kernel to take |rec_addr|
-                                   // seriously.
-                                   (flags & ~MAP_GROWSDOWN) | MAP_FIXED, -1, 0);
+    remote.infallible_mmap_syscall_if_alive(rec_addr, length, prot,
+                                            // Tell the kernel to take |rec_addr|
+                                            // seriously.
+                                            (flags & ~MAP_GROWSDOWN) | MAP_FIXED, -1, 0);
   } else {
     ASSERT(remote.task(), data.source == TraceReader::SOURCE_ZERO);
     emu_file = t->session().emufs().get_or_create(recorded_km);
@@ -629,12 +633,12 @@ static void write_mapped_data(ReplayTask* t,
 
 static void finish_private_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
                                 remote_ptr<void> rec_addr, size_t length,
-                                int prot, int flags, off64_t offset_pages,
+                                int prot, int flags, off64_t offset_bytes,
                                 const KernelMapping& km,
                                 TraceReader::MappedData& data) {
   LOG(debug) << "  finishing private mmap of " << km.fsname();
 
-  remote.infallible_mmap_syscall(
+  remote.infallible_mmap_syscall_if_alive(
       rec_addr, length, prot,
       // Tell the kernel to take |rec_addr| seriously.
       (flags & ~MAP_GROWSDOWN) | MAP_FIXED | MAP_ANONYMOUS, -1, 0);
@@ -643,7 +647,7 @@ static void finish_private_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
   // kernel-bug-workarounds when writing to tracee memory see the up-to-date
   // virtual map.
   t->vm()->map(t, rec_addr, length, prot, flags | MAP_ANONYMOUS,
-               page_size() * offset_pages, string(), KernelMapping::NO_DEVICE,
+               offset_bytes, string(), KernelMapping::NO_DEVICE,
                KernelMapping::NO_INODE, nullptr, &km);
 
   /* Restore the map region we copied. */
@@ -653,7 +657,7 @@ static void finish_private_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
 static void finish_shared_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
                                remote_ptr<void> rec_addr, size_t length,
                                int prot, int flags, const vector<TraceRemoteFd>& fds,
-                               off64_t offset_pages,
+                               off64_t offset_bytes,
                                const KernelMapping& km,
                                TraceReader::MappedData& data) {
   // Ensure there's a virtual file for the file that was mapped
@@ -669,7 +673,7 @@ static void finish_shared_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
   // Emufs file, so open it read-write in case we want to write to it through
   // the task's mem fd.
   remote.finish_direct_mmap(rec_addr, km.size(), prot, flags,
-                     emufile->proc_path(), O_RDWR, offset_pages, real_file,
+                     emufile->proc_path(), O_RDWR, offset_bytes, real_file,
                      real_file_name);
   // Write back the snapshot of the segment that we recorded.
   //
@@ -680,7 +684,6 @@ static void finish_shared_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
   // Update AddressSpace before loading data from the trace. This ensures our
   // kernel-bug-workarounds when writing to tracee memory see the up-to-date
   // virtual map.
-  uint64_t offset_bytes = page_size() * offset_pages;
   t->vm()->map(t, rec_addr, km.size(), prot, flags, offset_bytes,
                real_file_name, real_file.st_dev, real_file.st_ino, nullptr, &km,
                emufile);
@@ -694,10 +697,16 @@ static void finish_shared_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
     auto rt = t->session().find_task(fd.tid);
     ASSERT(t, rt) << "Can't find task " << fd.tid;
     if (rt->fd_table()->is_monitoring(fd.fd)) {
-      ASSERT(rt,
-             rt->fd_table()->get_monitor(fd.fd)->type() ==
-                 FileMonitor::Type::Mmapped);
-      ((MmappedFileMonitor*)rt->fd_table()->get_monitor(fd.fd))->revive();
+      auto type = rt->fd_table()->get_monitor(fd.fd)->type();
+      if (type == FileMonitor::Type::Mmapped) {
+        ((MmappedFileMonitor*)rt->fd_table()->get_monitor(fd.fd))->revive();
+      } else if (type == FileMonitor::Type::ODirect) {
+        rt->fd_table()->replace_monitor(rt, fd.fd, new MmappedFileMonitor(rt, emufile));
+      } else {
+        ASSERT(rt, false)
+            << "Expected monitor type Mmapped | ODirect for fd " << fd.fd << ", got monitor type "
+            << type;
+      }
     } else {
       rt->fd_table()->add_monitor(rt, fd.fd, new MmappedFileMonitor(rt, emufile));
     }
@@ -706,7 +715,7 @@ static void finish_shared_mmap(ReplayTask* t, AutoRemoteSyscalls& remote,
 
 static void process_mmap(ReplayTask* t, const TraceFrame& trace_frame,
                          size_t length, int prot, int flags, int fd,
-                         off64_t offset_pages, ReplayTraceStep* step) {
+                         off64_t offset_bytes, ReplayTraceStep* step) {
   step->action = TSTEP_RETIRE;
 
   {
@@ -730,12 +739,13 @@ static void process_mmap(ReplayTask* t, const TraceFrame& trace_frame,
       if (t->session().has_trace_quirk(TraceReader::SpecialLibRRpage)) {
         FileMonitor *fd_monitor = t->fd_table()->get_monitor(fd);
         if (fd_monitor && fd_monitor->type() == FileMonitor::RRPage) {
-          if (offset_pages == 0 && !(flags & MAP_FIXED) &&
-              length <= 2*page_size() && addr == (RR_PAGE_ADDR - page_size())) {
+          if (offset_bytes == 0 && !(flags & MAP_FIXED) &&
+              length <= 2 * PRELOAD_LIBRARY_PAGE_SIZE &&
+              addr == (RR_PAGE_ADDR - PRELOAD_LIBRARY_PAGE_SIZE)) {
             // We only mapped the first page during record. Do the same here
-            length = page_size();
+            length = PRELOAD_LIBRARY_PAGE_SIZE;
           }
-          if (offset_pages == 1 && length == page_size() &&
+          if (offset_bytes == PRELOAD_LIBRARY_PAGE_SIZE && length == PRELOAD_LIBRARY_PAGE_SIZE &&
               addr == RR_PAGE_ADDR && t->vm()->has_rr_page()) {
             // We skipped this during recording. Setting length to zero here
             // will have the same effect.
@@ -751,15 +761,15 @@ static void process_mmap(ReplayTask* t, const TraceFrame& trace_frame,
         uint64_t map_bytes = min(ceil_page_size(data.file_size_bytes) - data.data_offset_bytes, length);
         remote.finish_direct_mmap(addr, map_bytes, prot, flags,
                            data.file_name, O_RDONLY,
-                           data.data_offset_bytes / page_size(), real_file,
+                           data.data_offset_bytes, real_file,
                            real_file_name);
         KernelMapping km_sub = km.subrange(km.start(), km.start() + ceil_page_size(map_bytes));
         t->vm()->map(t, km.start(), map_bytes, prot, flags,
-                     page_size() * offset_pages, real_file_name,
+                     offset_bytes, real_file_name,
                      real_file.st_dev, real_file.st_ino, nullptr, &km_sub);
         addr += map_bytes;
         length -= map_bytes;
-        offset_pages += ceil_page_size(map_bytes) / page_size();
+        offset_bytes += ceil_page_size(map_bytes);
         data.source = TraceReader::SOURCE_ZERO;
         km = km.subrange(km_sub.end(), km.end());
       }
@@ -769,11 +779,11 @@ static void process_mmap(ReplayTask* t, const TraceFrame& trace_frame,
             extra_fds.push_back({ t->rec_tid, fd });
           }
           finish_shared_mmap(t, remote, addr, length, prot, flags, extra_fds,
-                              offset_pages, km, data);
+                              offset_bytes, km, data);
         } else {
           ASSERT(t, extra_fds.empty());
           finish_private_mmap(t, remote, addr, length, prot, flags,
-                              offset_pages, km, data);
+                              offset_bytes, km, data);
         }
       }
     }
@@ -827,15 +837,15 @@ static void process_mremap(ReplayTask* t, const TraceFrame& trace_frame,
     if (new_addr == old_addr) {
       // Non-moving mremap. Don't pass MREMAP_FIXED or MREMAP_MAYMOVE
       // since that triggers EINVAL when the new map overlaps the old map.
-      remote.infallible_syscall_ptr(trace_regs.original_syscallno(), new_addr,
-                                    old_size, new_size, 0);
+      remote.infallible_syscall_if_alive(trace_regs.original_syscallno(), new_addr,
+                                         old_size, new_size, 0);
     } else {
       // Force the mremap to use the destination address from recording.
       // XXX could the new mapping overlap the old, with different start
       // addresses? Hopefully the kernel doesn't do that to us!!!
-      remote.infallible_syscall_ptr(trace_regs.original_syscallno(), old_addr,
-                                    old_size, new_size,
-                                    MREMAP_MAYMOVE | MREMAP_FIXED, new_addr);
+      remote.infallible_syscall_if_alive(trace_regs.original_syscallno(), old_addr,
+                                         old_size, new_size,
+                                         MREMAP_MAYMOVE | MREMAP_FIXED, new_addr);
     }
 
     remote.regs().set_syscall_result(new_addr);
@@ -865,8 +875,8 @@ static void process_mremap(ReplayTask* t, const TraceFrame& trace_frame,
       // Shared non-EmuFs mappings must be of immutable files so it's OK to
       // just copy the file data into a private mapping here.
       int map_flags = MAP_ANONYMOUS | MAP_PRIVATE | MAP_FIXED;
-      remote.infallible_mmap_syscall(new_addr, new_size, mapping.map.prot(),
-                                     map_flags, -1, 0);
+      remote.infallible_mmap_syscall_if_alive(new_addr, new_size, mapping.map.prot(),
+                                              map_flags, -1, 0);
       t->vm()->unmap(t, new_addr, new_size);
       t->vm()->map(t, new_addr, new_size, mapping.map.prot(), map_flags,
                    mapping.map.file_offset_bytes(), string(),
@@ -1074,7 +1084,7 @@ static void handle_opened_files(ReplayTask* t, int flags) {
     } else if (is_rr_page_lib(o.path.c_str())) {
       file_monitor = new RRPageMonitor();
     } else if (flags & O_DIRECT) {
-      file_monitor = new FileMonitor();
+      file_monitor = new ODirectFileMonitor();
     } else {
       ASSERT(t, false) << "Why did we write filename " << o.path;
     }
@@ -1149,6 +1159,7 @@ static void rep_process_syscall_arch(ReplayTask* t, ReplayTraceStep* step,
    */
   switch (non_negative_syscall(sys)) {
     case Arch::execve:
+    case Arch::execveat:
       return process_execve(t, trace_frame, step);
 
     case Arch::brk:
@@ -1160,20 +1171,20 @@ static void rep_process_syscall_arch(ReplayTask* t, ReplayTraceStep* step,
           auto args = t->read_mem(
               remote_ptr<typename Arch::mmap_args>(trace_regs.orig_arg1()));
           return process_mmap(t, trace_frame, args.len, args.prot, args.flags,
-                              args.fd, args.offset / page_size(), step);
+                              args.fd, args.offset, step);
         }
         case Arch::RegisterArguments:
           return process_mmap(t, trace_frame, trace_regs.arg2(),
                               trace_regs.arg3(), trace_regs.arg4(),
                               trace_regs.arg5(),
-                              trace_regs.arg6() / page_size(), step);
+                              trace_regs.arg6(), step);
       }
       break;
     }
     case Arch::mmap2:
       return process_mmap(t, trace_frame, trace_regs.arg2(), trace_regs.arg3(),
                           trace_regs.arg4(), trace_regs.arg5(),
-                          trace_regs.arg6(), step);
+                          trace_regs.arg6() * 4096, step);
 
     case Arch::shmat:
       return process_shmat(t, trace_frame, trace_regs.arg3(), step);
@@ -1265,20 +1276,47 @@ static void rep_process_syscall_arch(ReplayTask* t, ReplayTraceStep* step,
       step->action = TSTEP_RETIRE;
       return;
 
+    case Arch::pkey_alloc:
+      // Older versions of rr (incorrectly) did not record the extra regs here.
+      if (t->session().has_trace_quirk(TraceReader::PkeyAllocRecordedExtraRegs)) {
+        t->set_extra_regs(trace_frame.extra_regs());
+      }
+      return;
+
+    case Arch::rseq: {
+      Registers r = t->regs();
+      auto rseq = remote_ptr<typename Arch::rseq_t>(r.arg1());
+      int flags = r.arg3();
+      uint32_t sig = r.arg4();
+      if (flags & RR_RSEQ_FLAG_UNREGISTER) {
+        t->rseq_state = nullptr;
+      } else {
+        t->rseq_state = make_unique<RseqState>(rseq, sig);
+      }
+      return;
+    }
+
     case Arch::perf_event_open: {
       Task* target = t->session().find_task((pid_t)trace_regs.arg2_signed());
       int cpu = trace_regs.arg3_signed();
       unsigned long flags = trace_regs.arg5();
       int fd = trace_regs.syscall_result_signed();
       int allowed_perf_flags = PERF_FLAG_FD_CLOEXEC;
+      bool virtualize = false;
       if (target && cpu == -1 && !(flags & ~allowed_perf_flags)) {
         auto attr =
             t->read_mem(remote_ptr<struct perf_event_attr>(trace_regs.orig_arg1()));
-        if (VirtualPerfCounterMonitor::should_virtualize(attr)) {
+        virtualize = VirtualPerfCounterMonitor::should_virtualize(attr);
+        if (virtualize) {
           t->fd_table()->add_monitor(t,
               fd, new VirtualPerfCounterMonitor(t, target, attr));
         }
       }
+      // Ignoring perf_event_open from syscallbuf; we'll attach a PreserveFileMonitor to it if it stays open
+      if (!virtualize &&
+          t->ip() != t->vm()->privileged_traced_syscall_ip().increment_by_syscall_insn_length(t->arch())) {
+        t->fd_table()->add_monitor(t, fd, new NonvirtualPerfCounterMonitor());
+      }
     }
       RR_FALLTHROUGH;
 
