@@ -121,6 +121,14 @@ open = IrregularEmulatedSyscall(x86=5, x64=2)
 # the lock).
 close = IrregularEmulatedSyscall(x86=6, x64=3, generic=57)
 
+#  int close_range(unsigned int first, unsigned int last, unsigned int flags);
+#
+# The close_range() system call closes all open file descriptors
+# from first to last (included).
+#
+# Errors closing a given file descriptor are currently ignored.
+close_range = IrregularEmulatedSyscall(x86=436, x64=436, generic=436)
+
 #  pid_t waitpid(pid_t pid, int *status, int options);
 #
 # The waitpid() system call suspends execution of the calling process
@@ -1304,7 +1312,7 @@ timer_settime = EmulatedSyscall(x86=260, x64=223, generic=110, arg4="typename Ar
 timer_gettime = EmulatedSyscall(x86=261, x64=224, generic=108, arg2="typename Arch::itimerspec")
 timer_getoverrun = EmulatedSyscall(x86=262, x64=225, generic=109)
 timer_delete = EmulatedSyscall(x86=263, x64=226, generic=111)
-clock_settime = UnsupportedSyscall(x86=264, x64=227, generic=112)
+clock_settime = EmulatedSyscall(x86=264, x64=227, generic=112, arg2="typename Arch::timespec")
 
 #  int clock_gettime(clockid_t clk_id, struct timespec *tp);
 #
@@ -1670,7 +1678,7 @@ memfd_create = IrregularEmulatedSyscall(x86=356, x64=319, generic=279)
 arch_prctl = IrregularEmulatedSyscall(x86=384, x64=158)
 
 bpf = IrregularEmulatedSyscall(x86=357, x64=321, generic=280)
-execveat = UnsupportedSyscall(x86=358, x64=322, generic=281)
+execveat = IrregularEmulatedSyscall(x86=358, x64=322, generic=281)
 userfaultfd = IrregularEmulatedSyscall(x86=374, x64=323, generic=282)
 membarrier = EmulatedSyscall(x86=375, x64=324, generic=283)
 mlock2 = UnsupportedSyscall(x86=376, x64=325, generic=284)
@@ -1685,7 +1693,7 @@ io_pgetevents = UnsupportedSyscall(x86=385, x64=333, generic=292)
 rseq = IrregularEmulatedSyscall(x86=386, x64=334, generic=293)
 
 clock_gettime64 = EmulatedSyscall(x86=403, arg2="typename Arch::Arch64::timespec")
-clock_settime64 = UnsupportedSyscall(x86=404)
+clock_settime64 = EmulatedSyscall(x86=404, arg2="typename Arch::Arch64::timespec")
 clock_adjtime64 = EmulatedSyscall(x86=405, arg2="typename Arch::Arch64::timex")
 clock_getres_time64 = EmulatedSyscall(x86=406, arg2="typename Arch::Arch64::timespec")
 clock_nanosleep_time64 = IrregularEmulatedSyscall(x86=407)
@@ -1706,7 +1714,7 @@ futex_time64 = IrregularEmulatedSyscall(x86=422)
 sched_rr_get_interval_time64 = UnsupportedSyscall(x86=423)
 
 # x86-64 decided to skip ahead here to catchup
-pidfd_send_signal = UnsupportedSyscall(x86=424, x64=424, generic=424)
+pidfd_send_signal = EmulatedSyscall(x86=424, x64=424, generic=424)
 io_uring_setup = IrregularEmulatedSyscall(x86=425, x64=425, generic=425)
 io_uring_enter = UnsupportedSyscall(x86=426, x64=426, generic=426)
 io_uring_register = UnsupportedSyscall(x86=427, x64=427, generic=427)
@@ -1728,6 +1736,9 @@ landlock_create_ruleset = UnsupportedSyscall(x86=444, x64=444, generic=444)
 landlock_add_rule = UnsupportedSyscall(x86=445, x64=445, generic=445)
 landlock_restrict_self = UnsupportedSyscall(x86=446, x64=446, generic=446)
 memfd_secret = UnsupportedSyscall(x86=447, x64=447, generic=447)
+process_mrelease = UnsupportedSyscall(x86=448, x64=448, generic=448)
+futex_waitv = UnsupportedSyscall(x86=449, x64=449, generic=449)
+set_mempolicy_home_node = UnsupportedSyscall(x86=450, x64=450, generic=450)
 
 # restart_syscall is a little special.
 restart_syscall = RestartSyscall(x86=0, x64=219, generic=128)
@@ -1744,6 +1755,9 @@ rrcall_notify_stap_semaphore_added = IrregularEmulatedSyscall(x86=1006, x64=1006
 rrcall_notify_stap_semaphore_removed = IrregularEmulatedSyscall(x86=1007, x64=1007, generic=1007)
 rrcall_check_presence = IrregularEmulatedSyscall(x86=1008, x64=1008, generic=1008)
 rrcall_detach_teleport = IrregularEmulatedSyscall(x86=1009, x64=1009, generic=1009)
+rrcall_arm_time_slice = IrregularEmulatedSyscall(x86=1010, x64=1010, generic=1010)
+rrcall_freeze_tid = IrregularEmulatedSyscall(x86=1011, x64=1011, generic=1011)
+rrcall_rdtsc = IrregularEmulatedSyscall(x86=1012, x64=1012, generic=1012)
 
 # These syscalls also appear under `socketcall` on x86.
 socket = EmulatedSyscall(x86=359, x64=41, generic=198)
