@@ -9,7 +9,7 @@ send_gdb('b breakpoint_thread')
 expect_gdb('Breakpoint 2')
 
 send_gdb('set scheduler-locking off')
-for i in xrange(50):
+for i in range(50):
   send_gdb('reverse-step')
 
 expect_gdb('Breakpoint 2')
