@@ -52,7 +52,9 @@
 #include <linux/videodev2.h>
 #include <linux/vt.h>
 #include <linux/wireless.h>
+#ifdef MQUEUE_H
 #include <mqueue.h>
+#endif
 #include <poll.h>
 #include <pthread.h>
 #include <pty.h>
@@ -69,7 +71,9 @@
 #include <sys/auxv.h>
 #include <sys/epoll.h>
 #include <sys/eventfd.h>
+#ifdef FANOTIFY_H
 #include <sys/fanotify.h>
+#endif
 #include <sys/file.h>
 #include <sys/fsuid.h>
 #include <sys/inotify.h>
@@ -112,7 +116,6 @@
 // X86 specific headers
 #if defined(__i386__) || defined(__x86_64__)
 #include <asm/prctl.h>
-#include <sys/io.h>
 #include <x86intrin.h>
 #endif
 
@@ -178,28 +181,31 @@ inline static int atomic_puts(const char* str) {
 #define printf(...) USE_atomic_printf_INSTEAD
 #define puts(...) USE_atomic_puts_INSTEAD
 
-inline static int check_cond(int cond) {
+inline static int atomic_assert(int cond, const char *str,
+				const char *file, const int line) {
   if (!cond) {
-    atomic_printf("FAILED: errno=%d (%s)\n", errno, strerror(errno));
-  }
-  return cond;
-}
-
-inline static int atomic_assert(int cond, const char *str) {
-  if (!check_cond(cond)) {
-    atomic_printf("FAILED: !%s\n", str);
+    atomic_printf("FAILED at %s:%d: !(%s) errno:%d (%s)\n", file, line, str,
+		  errno, strerror(errno));
     raise(SIGABRT);
   }
   return 1;
 }
 
-#define test_assert(cond) atomic_assert(cond, #cond)
+#define test_assert(cond) atomic_assert(cond, #cond, __FILE__, __LINE__)
 
 /**
  * Return the calling task's id.
  */
 inline static pid_t sys_gettid(void) { return syscall(SYS_gettid); }
 
+/**
+ * get the current cpu/node
+ */
+inline static int sys_getcpu(unsigned int* cpu, unsigned int* node) {
+  /* The 'tcache' parameter is unused in all kernels rr works on */
+  return syscall(SYS_getcpu, cpu, node, (void*)0);
+}
+
 /**
  * Ensure that |len| bytes of |buf| are the same across recording and
  * replay.
@@ -382,6 +388,9 @@ inline static SyscallWrapper get_spurious_desched_syscall(void) {
       (SyscallWrapper)dlsym(RTLD_DEFAULT, "spurious_desched_syscall");
   return ret ? ret : default_syscall_wrapper;
 }
+/* Old systems don't have these functions, re-define using the syscall */
+#define tgkill(tgid, tid, sig) \
+  syscall(SYS_tgkill, (int)(tgid), (int)(tid), (int)(sig))
 
 #define ALLOCATE_GUARD(p, v) p = allocate_guard(sizeof(*p), v)
 #define VERIFY_GUARD(p) verify_guard(sizeof(*p), p)
@@ -472,5 +481,15 @@ inline static SyscallWrapper get_spurious_desched_syscall(void) {
 #ifndef PR_SPEC_FORCE_DISABLE
 #define PR_SPEC_FORCE_DISABLE 8
 #endif
+#ifndef PR_SET_VMA
+#define PR_SET_VMA 0x53564d41
+#endif
+#ifndef PR_SET_VMA_ANON_NAME
+#define PR_SET_VMA_ANON_NAME 0
+#endif
+
+#ifndef BLKGETDISKSEQ
+#define BLKGETDISKSEQ _IOR(0x12,128,__u64)
+#endif
 
 #endif /* RRUTIL_H */
