@@ -20,6 +20,12 @@
 # scheduler, don't use it for anything else.
 #
 
+if ! which pidof &> /dev/null; then
+    pidof() {
+        ps ax -o 'pid= exe=' | grep ' \(\|.*/\)'"$1"'$' | awk '{print $1}'
+    }
+fi
+
 #  delay_kill <sig> <delay_secs> <proc>
 #
 # Deliver the signal |sig|, after waiting |delay_secs| seconds, to the
@@ -296,7 +302,7 @@ function do_ps { psflags=$1
 function debug { expectscript=$1; replayargs=$2
     _RR_TRACE_DIR="$workdir" test-monitor $TIMEOUT debug.err \
         python3 $TESTDIR/$expectscript.py \
-        $RR_EXE $GLOBAL_OPTIONS replay -o-n -x $TESTDIR/test_setup.gdb $replayargs
+        $RR_EXE $GLOBAL_OPTIONS replay -o-n -o-ix -o$TESTDIR/test_setup.gdb $replayargs
     if [[ $? == 0 ]]; then
         passed
     else
@@ -492,3 +498,11 @@ function checkpoint_test { exe=$1; min=$2; max=$3;
         fi
     done
 }
+
+function wait_for_complete {
+    local record_dir=${1:-"${workdir}/latest-trace"}
+    for ((i = 0; i < TIMEOUT * 10; i++)); do
+        [[ -f "$record_dir/incomplete" ]] || break
+        sleep 0.1
+    done
+}
