@@ -5,7 +5,9 @@
 #include <arpa/inet.h>
 #include <dirent.h>
 #include <elf.h>
+#ifdef EXECINFO_H
 #include <execinfo.h>
+#endif
 #include <fcntl.h>
 #include <inttypes.h>
 #include <limits.h>
@@ -29,6 +31,7 @@
 #include <string>
 #include <numeric>
 #include <random>
+#include <sstream>
 
 #include "preload/preload_interface.h"
 
@@ -121,10 +124,22 @@ remote_ptr<void> read_interpreter_base(std::vector<uint8_t> auxv) {
 }
 
 std::string read_ld_path(Task* t, remote_ptr<void> interpreter_base) {
-  if (!interpreter_base || !t->vm()->has_mapping(interpreter_base)) {
+  if (!interpreter_base) {
     return {};
   }
-  return t->vm()->mapping_of(interpreter_base).map.fsname();
+  // Find first executable mapping after the interpreter_base and use its fsname().
+  // Non-executable maps may not have a file name.
+  remote_ptr<void> addr = interpreter_base;
+  while (true) {
+    if (!t->vm()->has_mapping(addr)) {
+      return {};
+    }
+    const KernelMapping& map = t->vm()->mapping_of(addr).map;
+    if (map.prot() & PROT_EXEC) {
+      return map.fsname();
+    }
+    addr = map.end();
+  }
 }
 
 template <typename Arch> void patch_auxv_vdso_arch(RecordTask* t, uintptr_t search, uintptr_t new_entry_native) {
@@ -181,6 +196,9 @@ static bool is_start_of_scratch_region(Task* t, remote_ptr<void> start_addr) {
 }
 
 bool probably_not_interactive(int fd) {
+  if (Flags::get().non_interactive) {
+    return true;
+  }
   /* Eminently tunable heuristic, but this is guaranteed to be
    * true during rr unit tests, where we care most about this
    * check (to a first degree).  A failing test shouldn't
@@ -738,6 +756,13 @@ bool should_copy_mmap_region(const KernelMapping& mapping,
     LOG(debug) << "  copying " << mapping.fsname();
     return true;
   }
+  if (mapping.fsname().rfind("/etc/passwd", 0) == 0 ||
+      mapping.fsname().rfind("/etc/group", 0) == 0) {
+    // These files (and suffixes such as .cache etc) change very frequently in
+    // some environments.
+    LOG(debug) << "  copying " << mapping.fsname();
+    return true;
+  }
   if (private_mapping && (prot & PROT_EXEC)) {
     /* Be optimistic about private executable mappings */
     LOG(debug) << "  (no copy for +x private mapping " << mapping.fsname() << ")";
@@ -1049,7 +1074,7 @@ vector<CPUIDRecord> all_cpuid_records() {
 #ifdef SYS_arch_prctl
 #define RR_ARCH_PRCTL(a, b) syscall(SYS_arch_prctl, a, b)
 #else
-#define RR_ARCH_PRCTL(a, b) -1
+#define RR_ARCH_PRCTL(a, b) ((void)a, (void)b, -1)
 #endif
 
 bool cpuid_faulting_works() {
@@ -1177,7 +1202,7 @@ static CloneParameters extract_clone_parameters_arch(const Registers& regs) {
       result.ctid = regs.arg4();
       break;
   }
-  int flags = (int)regs.arg1();
+  int flags = (int)regs.orig_arg1();
   // If these flags aren't set, the corresponding clone parameters may be
   // invalid pointers, so make sure they're ignored.
   if (!(flags & (CLONE_PARENT_SETTID | CLONE_PIDFD))) {
@@ -1312,9 +1337,14 @@ string find_helper_library(const char *basepath)
 vector<string> read_proc_status_fields(pid_t tid, const char* name,
                                        const char* name2, const char* name3) {
   vector<string> result;
+  FILE *f;
   char buf[1000];
-  sprintf(buf, "/proc/%d/status", tid);
-  FILE* f = fopen(buf, "r");
+  if (tid == 0) {
+    f = fopen("/proc/self/status", "r");
+  } else {
+    sprintf(buf, "/proc/%d/status", tid);
+    f = fopen(buf, "r");
+  }
   if (!f) {
     return result;
   }
@@ -1509,7 +1539,7 @@ bool read_proc_net_socket_addresses(Task* t, int fd,
         LOG(warn) << "Remote address not in expected format";
         break;
       }
-      uint16_t rport = strtoul(remote_addr_str + 33, NULL, 16);
+      unsigned long rport = strtoul(remote_addr_str + 33, NULL, 16);
       if (rport > USHRT_MAX) {
         LOG(warn) << "Remote address not in expected format";
         break;
@@ -1660,27 +1690,39 @@ ScopedFd open_socket(const char* address, unsigned short* port,
   do {
     addr.sin_port = htons(*port);
     ret = ::bind(listen_fd, (struct sockaddr*)&addr, sizeof(addr));
-    if (ret && (EADDRINUSE == errno || EACCES == errno || EINVAL == errno)) {
+    if (ret && probe == PROBE_PORT && (EADDRINUSE == errno || EACCES == errno || EINVAL == errno)) {
+      *port = 0;
       continue;
     }
     if (ret) {
-      FATAL() << "Couldn't bind to port " << *port;
+      CLEAN_FATAL() << "Couldn't bind to port " << *port;
     }
 
     ret = listen(listen_fd, 1 /*backlogged connection*/);
-    if (ret && EADDRINUSE == errno) {
+    if (ret && probe == PROBE_PORT && EADDRINUSE == errno) {
+      *port = 0;
       continue;
     }
     if (ret) {
       FATAL() << "Couldn't listen on port " << *port;
     }
+    if (*port == 0) {
+      socklen_t sa_size = sizeof(addr);
+      ret = getsockname(listen_fd, (struct sockaddr*)&addr, &sa_size);
+      if (ret) {
+        FATAL() << "Could not get socket port";
+      }
+
+      *port = ntohs(addr.sin_port);
+    }
     break;
-  } while (++(*port), probe == PROBE_PORT);
+  } while (probe == PROBE_PORT);
   return listen_fd;
 }
 
 void notifying_abort() {
   flush_log_buffer();
+  dump_rr_stack();
 
   char* test_monitor_pid = getenv("RUNNING_UNDER_TEST_MONITOR");
   if (test_monitor_pid) {
@@ -1689,8 +1731,6 @@ void notifying_abort() {
     // do so.
     kill(pid, SIGURG);
     sleep(10000);
-  } else {
-    dump_rr_stack();
   }
 
   abort();
@@ -1699,9 +1739,14 @@ void notifying_abort() {
 void dump_rr_stack() {
   static const char msg[] = "=== Start rr backtrace:\n";
   write_all(STDERR_FILENO, msg, sizeof(msg) - 1);
+#if EXECINFO_H
   void* buffer[1024];
   int count = backtrace(buffer, 1024);
   backtrace_symbols_fd(buffer, count, STDERR_FILENO);
+#else
+  static const char msg_fallback[] = "<rr backtraces not available on this system>\n";
+  write_all(STDERR_FILENO, msg_fallback, sizeof(msg_fallback) - 1);
+#endif
   static const char msg2[] = "=== End rr backtrace\n";
   write_all(STDERR_FILENO, msg2, sizeof(msg2) - 1);
 }
@@ -1778,7 +1823,7 @@ TempFile create_temporary_file(const char* pattern) {
   snprintf(buf, sizeof(buf) - 1, "%s/%s", tmp_dir(), pattern);
   buf[sizeof(buf) - 1] = 0;
   TempFile result;
-  result.fd = mkstemp(buf);
+  result.fd = ScopedFd(mkstemp(buf));
   result.name = buf;
   return result;
 }
@@ -1803,7 +1848,7 @@ static ScopedFd create_tmpfs_file(const string &real_name) {
   name = name.substr(0, 255);
 
   ScopedFd fd =
-      open(name.c_str(), O_CREAT | O_EXCL | O_RDWR | O_CLOEXEC, 0700);
+      ScopedFd(name.c_str(), O_CREAT | O_EXCL | O_RDWR | O_CLOEXEC, 0700);
   /* Remove the fs name so that we don't have to worry about
    * cleaning up this segment in error conditions. */
   unlink(name.c_str());
@@ -1847,11 +1892,16 @@ vector<string> current_env() {
 }
 
 int get_num_cpus() {
-  int cpus = (int)sysconf(_SC_NPROCESSORS_ONLN);
-  return cpus > 0 ? cpus : 1;
+  cpu_set_t affinity_mask;
+  int ret = sched_getaffinity(0, sizeof(affinity_mask), &affinity_mask);
+  if (ret < 0) {
+    FATAL() << "sched_getaffinity failed";
+  }
+  return CPU_COUNT(&affinity_mask);
 }
 
-static const uint8_t rdtsc_insn[] = { 0x0f, 0x31 };
+const uint8_t rdtsc_insn[2] = { 0x0f, 0x31 };
+
 static const uint8_t rdtscp_insn[] = { 0x0f, 0x01, 0xf9 };
 static const uint8_t cpuid_insn[] = { 0x0f, 0xa2 };
 static const uint8_t int3_insn[] = { 0xcc };
@@ -1936,58 +1986,6 @@ bool is_advanced_pc_and_signaled_instruction(Task* t, remote_code_ptr ip) {
   return false;
 }
 
-/**
- * Read and parse the available CPU list then select a random CPU from the list.
- */
-static vector<int> get_cgroup_cpus() {
-  vector<int> cpus;
-  ifstream self_cpuset("/proc/self/cpuset");
-  if (!self_cpuset.is_open()) {
-    return cpus;
-  }
-  string cpuset_path;
-  getline(self_cpuset, cpuset_path);
-  self_cpuset.close();
-  if (cpuset_path.empty()) {
-    return cpus;
-  }
-  ifstream cpuset("/sys/fs/cgroup/cpuset" + cpuset_path + "/cpuset.cpus");
-  if (!cpuset.good()) {
-    return cpus;
-  }
-  while (true) {
-    int cpu1;
-    cpuset >> cpu1;
-    if (cpuset.fail()) {
-      return std::vector<int>{};
-    }
-    cpus.push_back(cpu1);
-    char c = cpuset.get();
-    if (cpuset.eof() || c == '\n') {
-      break;
-    } else if (c == ',') {
-      continue;
-    } else if (c != '-') {
-      return std::vector<int>{};;
-    }
-    int cpu2;
-    cpuset >> cpu2;
-    if (cpuset.fail()) {
-      return std::vector<int>{};
-    }
-    for (int cpu = cpu1 + 1; cpu <= cpu2; cpu++) {
-      cpus.push_back(cpu);
-    }
-    c = cpuset.get();
-    if (cpuset.eof() || c == '\n') {
-      break;
-    } else if (c != ',') {
-      return std::vector<int>{};
-    }
-  }
-  return cpus;
-}
-
 static string get_cpu_lock_file() {
   const char* lock_file = getenv("_RR_CPU_LOCK_FILE");
   return lock_file ? lock_file : trace_save_dir() + "/cpu_lock";
@@ -2002,11 +2000,23 @@ int choose_cpu(BindCPU bind_cpu, ScopedFd &cpu_lock_fd_out) {
     return -1;
   }
 
-  // Find out which CPUs we're allowed to run on at all
-  std::vector<int> cpus = get_cgroup_cpus();
+  // Find out which CPUs we're allowed to run on at all.
+  // sched_getaffinity intersects the task's `cpu_mask`
+  // (/proc/.../status Cpus_allowed_list) with `cpu_active_mask`
+  // which is almost the same as /sys/devices/system/cpu/online
+  cpu_set_t affinity_mask;
+  int ret = sched_getaffinity(0, sizeof(affinity_mask), &affinity_mask);
+  if (ret < 0) {
+    FATAL() << "sched_getaffinity failed";
+  }
+  std::vector<int> cpus;
+  for (int i = 0; i < CPU_SETSIZE; ++i) {
+    if (CPU_ISSET(i, &affinity_mask) && PerfCounters::support_cpu(i)) {
+      cpus.push_back(i);
+    }
+  }
   if (cpus.empty()) {
-    cpus.resize(get_num_cpus());
-    std::iota(cpus.begin(), cpus.end(), 0);
+    FATAL() << "Can't find a valid CPU to run on";
   }
 
   // When many copies of rr are running on the same machine, it's easy for them
@@ -2022,8 +2032,9 @@ int choose_cpu(BindCPU bind_cpu, ScopedFd &cpu_lock_fd_out) {
     struct stat stat;
     int err = fstat(cpu_lock_fd_out, &stat);
     DEBUG_ASSERT(err == 0);
-    if (stat.st_size < get_num_cpus()) {
-      if (ftruncate(cpu_lock_fd_out, get_num_cpus())) {
+    int configured_cpus = (int)sysconf(_SC_NPROCESSORS_CONF);
+    if (stat.st_size < configured_cpus) {
+      if (ftruncate(cpu_lock_fd_out, configured_cpus)) {
         FATAL() << "Failed to resize locks file";
       }
     }
@@ -2235,7 +2246,7 @@ size_t word_size(SupportedArch arch) {
 string json_escape(const string& str, size_t pos) {
   string out;
   for (size_t i = pos; i < str.size(); ++i) {
-    char c = str[i];
+    unsigned char c = str[i];
     if (c < 32) {
       char buf[8];
       sprintf(buf, "\\u%04x", c);
@@ -2286,6 +2297,11 @@ void normalize_file_name(string& s)
       }
       if (is_component(s.c_str() + i + 1, "..")) {
         // Peel off '/..'
+        if (out == 0) {
+          // If there's nothing to peel off, just discard this.
+          i += 2;
+          continue;
+        }
         size_t p = s.rfind('/', out - 1);
         if (p != string::npos) {
           out = p;
@@ -2353,9 +2369,85 @@ void SAFE_FATAL(int err, const char *msg)
     {.iov_base = (char*)msg, .iov_len=strlen(msg)},
     {.iov_base = nl, .iov_len=sizeof(nl)}
   };
-  (void)::writev(STDERR_FILENO, out, sizeof(out)/sizeof(struct iovec));
+  __attribute__((unused)) ssize_t ret = ::writev(STDERR_FILENO, out, sizeof(out)/sizeof(struct iovec));
   abort();
 }
 
+static int child_SIGSEGV(__attribute__((unused)) void* arg) {
+  kill(getpid(), SIGSEGV);
+  return 0;
+}
 
+bool coredumping_signal_takes_down_entire_vm() {
+  // The kernel behavior here changed in 5.16. Prior to that,
+  // a coredumping signal would bring down the entire vm.
+  // Starting with 5.16 it only brings down the thread group.
+  // -1 here indicates uninitialized, 0 the new behavior, and
+  // 1 the old behavior.
+  static int coredumping_signal_vm_behavior = -1;
+  if (coredumping_signal_vm_behavior < 0) {
+    pid_t child = -1;
+    int status = -1;
+    LOG(debug) << "Testing coredumping behavior in the presence of CLONE_VM";
+    if ((child = fork()) == 0) {
+      // Remove any handlers and make sure we get the default signal behavior.
+      signal(SIGSEGV, SIG_DFL);
+      // Don't litter the system with core dumps.
+      prctl(PR_SET_DUMPABLE, 0);
+      // Allocate a stack for the child.
+      const size_t stack_size = 1 << 20;
+      void* stack = mmap(NULL, stack_size, PROT_READ | PROT_WRITE,
+                         MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
+
+      pid_t tid = clone(child_SIGSEGV, (void*)((uintptr_t)stack + stack_size),
+                      CLONE_VM, NULL, NULL, NULL, NULL);
+      DEBUG_ASSERT(tid > 0);
+
+      pid_t ret = waitpid(tid, &status, __WALL);
+      DEBUG_ASSERT(ret == tid);
+      DEBUG_ASSERT(WIFSIGNALED(status));
+      exit(0);
+    }
+
+    DEBUG_ASSERT(child > 0);
+    pid_t ret = waitpid(child, &status, __WALL);
+    DEBUG_ASSERT(ret == child);
+    coredumping_signal_vm_behavior = WIFSIGNALED(status);
+  }
+
+  return coredumping_signal_vm_behavior > 0;
+}
+
+int parse_tid_from_proc_path(const std::string& pathname,
+                             const std::string& property) {
+  // XXX When rr becomes c++17 - use string view instead. Has better API for
+  // this stuff.
+  const auto proc_length = 6;
+  const auto prop_should_begin =
+      pathname.size() - property.size() - (pathname.back() == '/');
+  if (pathname.substr(prop_should_begin, property.size()) == property) {
+    const auto task = "task/"s;
+    const auto pos = pathname.find(task);
+    if (pos == std::string::npos) {
+      auto s = pathname.substr(proc_length);
+      if (pathname.back() == '/')
+        s.pop_back();
+      char* end;
+      const int tid = strtol(s.c_str(), &end, 10);
+      if (end == property) {
+        return tid;
+      }
+    } else {
+      auto s = pathname.substr(pos + task.size());
+      if (pathname.back() == '/')
+        s.pop_back();
+      char* end;
+      const int tid = strtol(s.c_str(), &end, 10);
+      if (end == property) {
+        return tid;
+      }
+    }
+  }
+  return -1;
+}
 } // namespace rr
